'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
import bpy
import pickle
import os
from mathutils import noise, Vector

# DONE: read from blender preset, then mix using vert group
# DONE: Implement cloth presets as: Main material + secondary material

GTOOL_GLOBAL_PRESETS_FOLDER = 'gtool_cloth_presets'

cloth_materials = {
    'cotton': {'stiffness': 15, 'shear': 15, 'bend': 0.5, 'shrink': 0.0},
    'denim': {'stiffness': 40, 'shear': 40, 'bend': 10, 'shrink': 0.0},
    'leather': {'stiffness': 80, 'shear': 80, 'bend': 120, 'shrink': 0.0},
    'cuffs': {'stiffness': 40, 'shear': 40, 'bend': 10, 'shrink': 0.3},
}
full_cloth_props = [
    # 'quality' : 5,
    'mass',
    'air_damping',
    'sewing_force',
    # 'bending_model', # alwa use 'angular'  (linear is legacy)

    'tension_stiffness',
    'compression_stiffness',
    'shear_stiffness',
    'bending_stiffness',

    'tension_damping',
    'compression_damping',
    'shear_damping',
    'bending_damping',

    # no shrink by default in blender cloth presets
    'shrink_min',
]
denim = {
    # 'quality' : 5,
    'mass' : 0.15000000596046448,
    'air_damping' : 1.0,
    'bending_model' : 'ANGULAR',

    'tension_stiffness' : 5.0,
    'compression_stiffness' : 5.0,
    'shear_stiffness' : 5.0,
    'bending_stiffness' : 0.05000000074505806,

    'tension_damping' : 0.0,
    'compression_damping' : 0.0,
    'shear_damping' : 0.0,
    'bending_damping' : 1.0,
    'shrink_min' : 0.0, # written to shrink_max
    # no shrink...
}

def get_presets_path():
    all_presets = os.path.join(bpy.utils.user_resource('SCRIPTS'), 'presets')
    # all_presets = bpy.utils.user_resource('SCRIPTS')
    view_presets = os.path.join(all_presets, GTOOL_GLOBAL_PRESETS_FOLDER)
    if not os.path.exists(view_presets):
        # os.mkdir(view_presets) # use mkdirs instead to create all dirs (eg 'presets' and 'gtool_cloth_presets') at same time (else it may error out)
        os.makedirs(view_presets)
    return view_presets


def get_attribs(read_props):
    out_dict = {}
    for attr in full_cloth_props:
        out_dict[attr] = getattr(read_props, attr)
    return out_dict
    # for prop in read_props.bl_rna.properties.keys():
    #     if prop in {'name', 'rna_type'}:
    #         continue
    #     if not read_props.bl_rna.properties[prop].is_readonly:
    #         value = getattr(read_props, prop)
    #         #* if pickle, then cant save Matrix,Color to it.
    #         if value.__class__ == mathutils.Color or value.__class__ == mathutils.Vector or value.__class__ == mathutils.Quaternion:
    #             out_dict[prop] = value[:] #passes the copy, rather than ref
    #         elif str(type(value)) == "<class 'bpy_prop_array'>":
    #             out_dict[prop] = value[:]  # passes the copy, rather than ref
    #         else:

def set_settings(target_props, props_dict):
    for attr in full_cloth_props:
        val = props_dict.get(attr, None)
        if val is not None:
            setattr(target_props, attr, val)
        # setattr(target_props, attr, props_dict[attr])
    # for prop in target_props.bl_rna.properties.keys():
    #     if prop in {'name', 'rna_type',  'use_local_camera'}:  # crash on use_local_camera
    #         continue
    #     if not target_props.bl_rna.properties[prop].is_readonly and prop in props_dict:
    #         # print(f'set: {prop} from {getattr(target_props,prop)} to {props_dict[prop]}')
    #         try:
    #             setattr(target_props, prop, props_dict[prop])
    #         except TypeError as te:
    #             print(str(te))
    #             continue



class GT_OT_SaveClothPreset(bpy.types.Operator):
    bl_idname = "cloth.save_global_cloth_preset"
    bl_label = "Save Global Preset"
    bl_description = "Store/Override global cloth preset with active scene cloth settings"
    bl_options = {'REGISTER'}

    preset_name: bpy.props.StringProperty(name='Name', description='', default='')

    def invoke(self, context, event):
        gprops = context.scene.garment_props
        active_local_preset = gprops.local_cloth_presets[gprops.active_lpreset_index]
        self.preset_name = active_local_preset.name
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        gprops = context.scene.garment_props
        active_local_preset = gprops.local_cloth_presets[gprops.active_lpreset_index]
        cloth_props = get_attribs(active_local_preset)
        pickle_file = os.path.join(get_presets_path(), self.preset_name + '.dat')
        # check if  file exists - if so label it as override
        if os.path.exists(pickle_file):
            layout.label(text=f'Override preset: {self.preset_name}?', icon='ERROR')
        else:
            layout.label(text=f'Store new preset:', icon='QUESTION')

        layout.prop(self, 'preset_name')


    def execute(self, context):
        gprops = context.scene.garment_props
        active_local_preset = gprops.local_cloth_presets[gprops.active_lpreset_index]
        cloth_props = get_attribs(active_local_preset)
        pickle_file = os.path.join(get_presets_path(), self.preset_name + '.dat')
        # check if  file exists
        with open(pickle_file, "wb") as f:
            pickle.dump(cloth_props, f)

        return {'FINISHED'}


class GT_OT_OverrideClothPreset(bpy.types.Operator):
    bl_idname = "cloth.override_cloth_preset"
    bl_label = "Override preset"
    bl_description = "Override stored preset, with current cloth settings"
    bl_options = {"REGISTER"}

    override_name: bpy.props.StringProperty(name='Name', description='', default='')

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.label(text=f'Override: {self.override_name[:-4]}?')
        layout.prop(self, 'store_eevee')

    def execute(self, context):
        bpy.ops.cloth.save_global_cloth_preset(preset_name=self.override_name[:-4])
        self.report({'INFO'}, f'Override succesful')

        return {"FINISHED"}


class GT_OT_DeleteClothPreset(bpy.types.Operator):
    bl_idname = "cloth.delete_global_cloth_preset"
    bl_label = "Delete preset"
    bl_description = "Delete cloth preset"
    bl_options = {'REGISTER'}

    preset: bpy.props.StringProperty(name='Name', description='Stored preset name', default='default')

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.label(text=f"Remove global preset: '{self.preset}'?")

    def execute(self, context):
        pickle_file = os.path.join(get_presets_path(), self.preset+'.dat')
        try:
            os.remove(pickle_file)
        except FileNotFoundError as Err:
            self.report({'ERROR'}, Err.__str__())
            return {'CANCELLED'}
        self.report({'INFO'}, f'File {pickle_file} removed successfully')
        return {'FINISHED'}


# def reset global presets from __file__ + 'defaults_presets' folderj. With clear - option - remove existing presets
class GT_OT_ResetClothPresets(bpy.types.Operator):
    bl_idname = "cloth.reset_global_cloth_presets"
    bl_label = "Reset Global Presets"
    bl_description = "Reset global cloth presets to default, by replacing them with build-in GarmentTool presets."
    bl_options = {'REGISTER'}

    clear: bpy.props.BoolProperty(name='Clear', description='Remove existing presets', default=False)

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.label(text=f"Reset global presets to default?")
        layout.prop(self, 'clear')

    def execute(self, context):
        user_presets_path = get_presets_path()
        default_presets_path = os.path.join(os.path.dirname(__file__), GTOOL_GLOBAL_PRESETS_FOLDER)
        if self.clear:
            for f in os.listdir(user_presets_path):
                os.remove(os.path.join(user_presets_path, f))
        import shutil # standard lib
        for f in os.listdir(default_presets_path):
            shutil.copy(os.path.join(default_presets_path, f), os.path.join(user_presets_path, f))

        self.report({'INFO'}, f'Presets reset to default')
        return {'FINISHED'}

PRESET_FILES_LIST = []
def cloth_file_presets(self, context):
    files = [f[:-4] for f in os.listdir(get_presets_path()) if f.endswith('.dat')]
    # sort by name
    sortedf = [(f, f, f) for f in sorted(files)]
    # sorted = [(f[:-4], f[:-4], f[:-4]) for f in os.listdir(get_presets_path()) if f.endswith('.dat')]
    global PRESET_FILES_LIST  # or else enum will be messed up
    if PRESET_FILES_LIST != sortedf:
        PRESET_FILES_LIST = sortedf
    return PRESET_FILES_LIST

LocalPresetsList = []
def cloth_local_default_presets(self, context):
    '''Used only in garment.default_cloth_preset'''
    gprops = context.scene.garment_props
    loc_presets = [(lpreset.name, lpreset.name, lpreset.name) for lpreset in gprops.local_cloth_presets]
    loc_presets.insert(0, ('None', 'None', 'Do not modify cloth settings'))
    global LocalPresetsList  # or else enum will be messed up
    if LocalPresetsList != loc_presets:
        LocalPresetsList = loc_presets
    return LocalPresetsList


SecLocalPresetsList = []
def cloth_local_pattern_presets(self, context):
    gprops = context.scene.garment_props
    # active_local_preset = gprops.local_cloth_presets[gprops.active_lpreset_index]
    files = [(lpreset.name, lpreset.name, lpreset.name) for lpreset in gprops.local_cloth_presets]

    files.insert(0, ('Default', 'Default', 'Default'))
    global SecLocalPresetsList  # or else enum will be messed up
    if SecLocalPresetsList != files:
        SecLocalPresetsList = files
    return SecLocalPresetsList


# def init_cloth_presets():
#     prefs = get_addon_preferences()
#     file_presets = [f[:-4] for f in os.listdir(get_presets_path()) if f.endswith('.dat')]
#     np.random.seed(0)
#     existing_blender_presets = [preset.name for preset in prefs.cloth_presets]
#
#     # NOTE: remove blender presets that are not in file_presets
#     for i, b_preset in reversed(list(enumerate(prefs.cloth_presets))):
#         if b_preset.name not in file_presets:
#             prefs.cloth_presets.remove(i)
#
#     # NOTE: add file presets that are not in blender presets
#     np.random.seed(0)
#     for f in file_presets:
#         if f not in existing_blender_presets:
#             global_preset = prefs.cloth_presets.add()
#             global_preset.name = f
#             global_preset.color = np.random.random_sample(3).tolist()#(0.8, 0.8, 0.8)


def read_preset(preset):
    pickle_file = os.path.join(get_presets_path(), preset+'.dat')
    cloth_preset = {}
    with open(pickle_file, "rb") as f: #*oreder is important
        cloth_preset = pickle.load(f)
    return cloth_preset

def get_set_preset_mat(preset_name):
    preset_mat = bpy.data.materials.get(preset_name)
    if not preset_mat:
        preset_mat = bpy.data.materials.new(preset_name)
        preset_mat.use_fake_user = True
        noise.seed_set(hash(preset_name) % 100000)
        col = noise.random_unit_vector(size=3)
        col = col * 0.5 + Vector((0.5, 0.5, 0.5))
        preset_mat.diffuse_color = col.normalized().to_tuple() + (1,) # (1, 1, 1)
    return preset_mat

# operator that will generate cloth material usinf get_set_preset_mat
class GT_OT_GenerateClothMaterial(bpy.types.Operator):
    bl_idname = "cloth.generate_cloth_material"
    bl_label = "Generate material"
    bl_description = "Generate cloth preset material (if missing)"
    bl_options = {'REGISTER'}

    preset: bpy.props.StringProperty(name='Preset', description='Preset to generate')

    def execute(self, context):
        get_set_preset_mat(self.preset)
        return {'FINISHED'}


class GT_OT_LoadClothPreset(bpy.types.Operator):
    bl_idname = "cloth.load_global_cloth_preset"
    bl_label = "Load cloth preset"
    bl_description = "Load/Reload cloth preset from file into scene local cloth presets.\nShift+Click to rename"
    bl_options = {'REGISTER', 'UNDO'}

    preset: bpy.props.StringProperty(name="Preset", description='Preset to load')
    rename_name: bpy.props.StringProperty(name='New name', description='', default='')

    shift_clicked = False #if clicked shift - rename
    def invoke(self, context, event):
        if event.shift:
            self.shift_clicked = True
            self.rename_name = self.preset
            return context.window_manager.invoke_props_dialog(self)

        # check if self.preset already exist in local presets
        self.override = False
        for local_preset in context.scene.garment_props.local_cloth_presets:
            if local_preset.name == self.preset:
                self.override = True
                self.rename_name = self.preset
                return context.window_manager.invoke_props_dialog(self)

        return self.execute(context)

    def draw(self, context):
        layout = self.layout
        if self.shift_clicked:
            layout.label(text=f'Renaming {self.preset}:')
            row = layout.row(align=True)
            row.prop(self, 'rename_name')
        elif self.override:
            layout.label(text=f'Override local preset: {self.preset}?')

    def execute(self, context):
        if self.shift_clicked:
            pickle_file = os.path.join(get_presets_path(), self.preset)
            new_fname = os.path.join(get_presets_path(), self.rename_name)
            try:
                os.rename(pickle_file+'.dat', new_fname+'.dat')
            except Exception as e:
                self.report({'ERROR'}, str(e))
                print(e)
                return {'FINISHED'}

            self.report({'INFO'}, f'File renamed')
            return {'FINISHED'}


        imported_preset = read_preset(self.preset)
        gt_scn_props = context.scene.garment_props

        # check if preset exists
        if self.override:
            for local_preset in gt_scn_props.local_cloth_presets:
                if local_preset.name == self.preset:
                    set_settings(local_preset, imported_preset)
                    # force update (by running prop update fun)
                    local_preset.mass = imported_preset['mass'] # ?
        else: # create just new local preset
            local_preset = gt_scn_props.local_cloth_presets.add()
            local_preset.name = self.preset
            get_set_preset_mat(self.preset)
            set_settings(local_preset, imported_preset)

        self.report({'INFO'}, f'Imported {self.preset} preset')
        return {'FINISHED'}


