'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from pathlib import Path
import bpy
from .utils.helper_functions import addon_name_lowercase, fix_garment_vgroups, get_addon_preferences, get_create_pin_group, fix_garment_colliding_names, get_ancestors_collections, find_view_coll
from .garments_library.garments_lib import enum_previews_from_directory_items
from .garments_library import garments_lib
from .bind_mesh import ProjectObjectProps
from .utils.curve_helpers import get_predicted_vert_count
from .edit_cloth_preset import assign_cloth_preset_weights_to_all_patterns
from .cloth_presets_lib import cloth_local_default_presets, cloth_local_pattern_presets, get_set_preset_mat
from .spline_tools.sewings_overlay import check_enable_draw_sewings, update_pattern_segs
from .spline_tools import sewings_overlay
from .mesh_cloth_functions import update_keyframe_prop
import os
import time


class GtLibDirectories(bpy.types.PropertyGroup):
    def set_abs_path(self, context):
        print(f'Library directory is being changed to absolute. Current: {self.directory}')
        abs_p = os.path.abspath(self.directory)
        if os.path.exists(abs_p):
            self['directory'] = os.path.abspath(self.directory)
        else:
            self['directory'] = str(Path.home())
        print(f"Updated to: {self['directory']}")

    name: bpy.props.StringProperty(name='Library Name', default='Library')
    directory: bpy.props.StringProperty(name="Library directory", default=str(Path.home()), description="Library directory", subtype="DIR_PATH", update=set_abs_path)

class GTLIB_UL_LibDirs(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        path = item
        if path:
            layout.prop(path, 'name', text='', emboss=False)
        else:
            layout.label(text="", translate=False, icon_value=icon)

# operator 'gt_preferences.lib_directory_add'  that will add new item to get_addon_preferences().lib_paths
class GTOOL_OT_LibDirectoryAdd (bpy.types.Operator):
    bl_idname = "gt_preferences.lib_directory_add"
    bl_label = "Add Library Directory"
    bl_description = "Add new library directory"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        prefs = get_addon_preferences()
        prefs.lib_paths.add()
        prefs.lib_active_index = len(prefs.lib_paths) - 1
        self.report({'INFO'}, f"Library directory added")
        return {'FINISHED'}


class GTOOL_OT_LibDirectoryRemove (bpy.types.Operator):
    bl_idname = "gt_preferences.lib_directory_remove"
    bl_label = "Remove Library Directory"
    bl_description = "Remove library directory"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        prefs = get_addon_preferences()
        prefs.lib_paths.remove(prefs.lib_active_index)
        prefs.lib_active_index -= 1
        self.report({'INFO'}, f"Library directory removed")
        return {'FINISHED'}



class GarmentToolPreferences(bpy.types.AddonPreferences):
    bl_idname = 'garment_tool'

    def chage_size(self, context):
        for garment in context.scene.cloth_garment_data:
            for pin in garment.pins:
                if pin.source_obj and pin.source_obj.type == 'EMPTY':
                    pin.source_obj.empty_display_size = self.pin_empties_size
                if pin.target_obj and pin.target_obj.type == 'EMPTY':
                    pin.target_obj.empty_display_size = self.pin_empties_size


    tabs: bpy.props.EnumProperty(name="Tabs", items=[("UPDATE", "Update", ""),
                                                     ("PROPS", "Properties", ""),
                                                     ("LIBRARY", "Library", "")], default="UPDATE")

    #library tab
    lib_paths: bpy.props.CollectionProperty(type=GtLibDirectories)
    lib_active_index: bpy.props.IntProperty(name="Active Index", default=0)

    rerender: bpy.props.BoolProperty(default=False)
    debug_view: bpy.props.BoolProperty(name='Debug UI',default=False, description='Show mode detailed lists of Pockets and Pins')

    #Properties tab
    use_absolute_dimensions: bpy.props.BoolProperty(name="Switch to absolute segment dimensions", description="Enable centimeters when displaying segments length",  default=False)
    pin_empties_size : bpy.props.FloatProperty(name='Pin Empties Size ', description='', default= 0.01, min=0.001, soft_max=10, update=chage_size)

    #autoupdate tab
    update_exist: bpy.props.BoolProperty(name="Update Exist", description="There is new GroupPro update",  default=False)
    update_text: bpy.props.StringProperty(name="Update text",  default='')


    def draw(self, context):
        layout = self.layout
        row = layout.row(align=True)
        row.prop(self, "tabs", expand=True)
        box = layout.box()
        if self.tabs == "UPDATE":
            col = box.column()
            sub_row = col.row(align=True)
            sub_row.operator(addon_name_lowercase()+".check_for_update")
            split_lines_text = self.update_text.splitlines()
            for line in split_lines_text:
                sub_row = col.row(align=True)
                sub_row.label(text=line)
            sub_row.separator()
            sub_row = col.row(align=True)
            if self.update_exist:
                sub_row.operator(addon_name_lowercase()+".update_addon", text='Install latest version').reinstall = False
            else:
                sub_row.operator(addon_name_lowercase()+".update_addon", text='Reinstall current version').reinstall = True
            sub_row.operator(addon_name_lowercase()+".rollback_addon")

        elif self.tabs == "PROPS":
            col = box.column()
            sub_col = col.column()
            sub_col.prop(self, "use_absolute_dimensions")
            sub_col.prop(self, "pin_empties_size")
            sub_col.prop(self, "debug_view")

        elif self.tabs == "LIBRARY":
            col = box.column()
            sub_col = col.column()

            lib_paths = self.lib_paths

            if len(lib_paths) == 0:
                layout.operator("gt_preferences.lib_directory_add", text="Add", icon='ADD')
                return

            row = box.row()
            row.template_list("GTLIB_UL_LibDirs", "", self, "lib_paths", self, "lib_active_index", rows=3)
            col = row.column(align=True)
            col.operator("gt_preferences.lib_directory_add", icon="ADD", text='')
            col.operator("gt_preferences.lib_directory_remove", icon="REMOVE", text='')

            path = lib_paths[self.lib_active_index]
            row = box.row()
            row.prop(path, 'directory', text='')

            sub_row = box.row(align=True)
            sub_row.operator("garment_lib.render_previews", text="Render missing previews")
            sub_row.prop(self, 'rerender', text='Re-render ALL previews')



class SegmentIndexCollection(bpy.types.PropertyGroup):
    def test_obj_is_curve2(self, context):
        if self.source_obj and self.source_obj.type != 'CURVE':
            self['source_obj'] = None
        if self.target_obj and self.target_obj.type != 'CURVE':
            self['target_obj'] = None

    source_obj: bpy.props.PointerProperty(name='Stitch source',  description='Stitch source', type=bpy.types.Object,  update=test_obj_is_curve2)
    target_obj: bpy.props.PointerProperty(name='Stitch Target',  description='Stitch Target', type=bpy.types.Object, update=test_obj_is_curve2)
    from_segment_idx: bpy.props.IntProperty(name='Segment index', default=0, min=0, soft_max=32,  description='Seweing source segment index')
    from_spline_idx: bpy.props.IntProperty(name='Spline index', default=0, min=0, soft_max=8,  description='Spline from index')
    to_segment_idx: bpy.props.IntProperty(name='Segment index', default=0, min=0, soft_max=32,  description='Seweing target segment index')
    to_spline_idx: bpy.props.IntProperty(name='Spline index', default=0, min=0, soft_max=8,  description='Spline to index')
    flip: bpy.props.BoolProperty(name='Flip', default=False, description='Flip sweing')


def shared_update_preset(self, context):
    # precalculate all data required for assign_cloth_preset_to_all_patterns()
    # garment = self.id_data.path_resolve(self.path_from_id().split('.')[0])
    garment_idx = context.scene.garment_index
    garment = context.scene.cloth_garment_data[garment_idx]
    assign_cloth_preset_weights_to_all_patterns(garment) # update all, since we need to normalize through other patterns cloth presets...


# register_class(GTOOL_PG_SegmentIndexCollection)
ITEMS_PATTERNS = [] #for keeping reference in python to items - fix enum bug in blender
ITEMS_STITCHES = [] #for keeping reference in python to items - fix enum bug in blender
TIME = None
class PatternPropGroup(bpy.types.PropertyGroup):
    def test_obj_is_curve (self, context):
        garment = self.id_data.path_resolve(self.path_from_id().split('.')[0])  # 'cloth_garment_data[0].sewing_patterns[1]'

        current_garment_mesh = self.id_data.path_resolve(self.path_from_id()) # we skipp checking self with itself
        if self['pattern_obj']:
            if self['pattern_obj'] and self['pattern_obj'].type != 'CURVE':
                self['pattern_obj'] = None
                print('Object is not curve')

            for pattern in garment.sewing_patterns:
                if pattern != current_garment_mesh and self.pattern_obj == pattern.pattern_obj:
                    self['pattern_obj'] = None
                    print('Pattern is already assigned to garment')
                    break
        total_point_count = 0
        for pattern in garment.sewing_patterns:
            if pattern.pattern_obj and pattern.is_enabled:
                area = pattern.pattern_obj.dimensions.x * pattern.pattern_obj.dimensions.y
                total_point_count += area * garment.cloth_res * garment.cloth_res
        garment.predicted_out_vert_count = int(total_point_count)

        # NOTE: deal with changing vgroups masks, on pattern deletion
        # XXX: is this still needed, after we added cloth presets ..
        global TIME #?
        if not TIME or (time.time() - TIME)>0.1:
            if self['pattern_obj']:
                new_element_id = [i for i, p in enumerate(garment.sewing_patterns) if p.pattern_obj and p.pattern_obj == self['pattern_obj']][0]
                fix_garment_vgroups(garment, new_element_id=new_element_id) #when addin patter, fix garment.vgroups
                if self['pattern_obj'].data.dimensions == '3D':  # add 3d curves to special 'PIN' vgroup
                    vg_2d_pin = get_create_pin_group(garment) #? mayeb noo nee since we alway add to end?
                    vg_2d_pin.vgroups_patterns = vg_2d_pin.vgroups_patterns.union({self['pattern_obj'].name+'.'+str(len(garment.sewing_patterns)-1)})
            else:
                pass #do nothing when only cleaning obj name. Removing Sewing handled in sewing_pattern.remove() operator
                # removed_id = [i for i, p in enumerate(garment.sewing_patterns) if p.pattern_obj is None][0]
                # g_id = int(self.path_from_id().split('.')[0][-2])
                # bpy.ops.garment.pattern_remove(garment_index=g_id, pattern_idx=removed_id)
                # fix_garment_vgroups(garment, removed_id=removed_id)  # we use pattern_obj == none to detect removed element

        TIME = time.time()
        assign_cloth_preset_weights_to_all_patterns(garment) # auto add material and cloth presets, when adding new pattern

    def clear_in_overlay_cache(self, context):
        #* fix overlays_drawing when disabling pattern
        if not self.is_enabled:
            if self.pattern_obj and self.pattern_obj.name in sewings_overlay.PATTER_SEGMENTS_CACHE.keys():
                del sewings_overlay.PATTER_SEGMENTS_CACHE[self.pattern_obj.name]
        else:
            if self.pattern_obj:
                update_pattern_segs(self.pattern_obj)

    def update_preset_prop(self, context):
        shared_update_preset(self, context)

    pattern_obj: bpy.props.PointerProperty(name="Pattern", type=bpy.types.Object, update=test_obj_is_curve)
    is_enabled:  bpy.props.BoolProperty(name='Enable', default=True, description='Enable pattern', update=clear_in_overlay_cache)
    add_lattice: bpy.props.BoolProperty(name="Generate Lattice", default=False, description="Generate Lattice")
    bend_factor: bpy.props.FloatProperty(name="Bend Factor", default=1, description="Bend Factor")
    cloth_preset: bpy.props.EnumProperty(name="Main Preset", description="Main Cloth Preset applied to whole garment", items=cloth_local_pattern_presets, update=update_preset_prop)


class PocketCurve(bpy.types.PropertyGroup):
    def pocket_sewing(self, context):
        global ITEMS_STITCHES
        if self.pattern_obj is not None and len(self.pattern_obj.data.splines[0].bezier_points) > 0:
            ITEMS_STITCHES.clear()
            for i in range(len(self.pattern_obj.data.splines[0].bezier_points)):
                ITEMS_STITCHES.append((str(i), str(i), ""))
        return ITEMS_STITCHES

    def update_preset_prop(self, context):
        shared_update_preset(self, context)

    pattern_obj: bpy.props.PointerProperty(name="Pocket pattern", type=bpy.types.Object)
    target_pattern: bpy.props.PointerProperty(name="Target pattern", type=bpy.types.Object)
    pocket_sewing: bpy.props.EnumProperty( name="pocket sewing", description="pocket sewing", items=pocket_sewing, options={"ENUM_FLAG"})
    cloth_preset: bpy.props.EnumProperty(name="Main Preset", description="Main Cloth Preset applied to whole garment", items=cloth_local_pattern_presets, update=update_preset_prop)


class GarmentPin(bpy.types.PropertyGroup):
    source_obj: bpy.props.PointerProperty(name='Pin source',  description='Pin source', type=bpy.types.Object)
    target_obj: bpy.props.PointerProperty(name='Pin Target',  description='Pin Target', type=bpy.types.Object)
    source_co: bpy.props.FloatVectorProperty(name="", description="", default=(0.0, 0.0, 0.0))
    target_co: bpy.props.FloatVectorProperty(name="", description="", default=(0.0, 0.0, 0.0))


class PatternVertGroups(bpy.types.PropertyGroup):
    def pattern_items(self, context):
        # parent_garment_idx = int(self.path_from_id()[19]) #gives cloth_garment_data[0].generated_vgroups[1]
        # garment = self.id_data.path_resolve(self.path_from_id().split('.')[0])
        garment_idx = context.scene.garment_index
        garment = context.scene.cloth_garment_data[garment_idx]

        global ITEMS_PATTERNS
        it_patt = []
        for i,pattern in enumerate(garment.sewing_patterns):
            if pattern.pattern_obj is not None:
               it_patt.append((f"{pattern.pattern_obj.name}.{i}", pattern.pattern_obj.name, ""))
            else:
               it_patt.append((f"None.{i}", "", ""))  # placeholders hack to kepp the order in vgroups_patterns set, when clearing pattern name.
        if ITEMS_PATTERNS != it_patt:
            ITEMS_PATTERNS = it_patt
        return ITEMS_PATTERNS

    name: bpy.props.StringProperty(name="Group", description="New vertex group name", default='vg_name')
    vgroups_patterns: bpy.props.EnumProperty(name="Patterns", description="Choose patterns that will receive this vertex group", items=pattern_items, options={"ENUM_FLAG"})
    # add_bend_deform: bpy.props.BoolProperty(name="Generate Bend Deform", default=False, description="Generate Bend Deform")
    add_lattice: bpy.props.BoolProperty(name="Generate Lattice", default=False, description="Generate Lattice")
    # add_bend_deform: bpy.props.BoolProperty(name="Add Bend Deform", default=False, description="Add Bend Deform")


ITEMS_GARMENTS = []  # for keeping reference in python to items - fix enum bug in blender


def garment_items(self, context):
    # parent_garment_idx = int(self.path_from_id()[19]) #gives cloth_garment_data[0].generated_vgroups[1]
    global ITEMS_GARMENTS
    garments = context.scene.cloth_garment_data
    #? how to prevent random values like *3() in some items
    new_it = [(str(i), garment.name, "") for i, garment in enumerate(garments)]
    if new_it != ITEMS_GARMENTS:  # for  now this seems to fix random changes....
        ITEMS_GARMENTS = new_it
    #! IF this works, propaget it to other dynamic enums...
    return ITEMS_GARMENTS


FOLDER_ITEMS = {}  # for keeping reference in python to items - fix enum bug in blender

def folder_items(self, context):
    global FOLDER_ITEMS
    pref = get_addon_preferences()
    lib_path_len = len(pref.lib_paths[pref.lib_active_index].directory)
    curr_sub_dir = context.scene.ht_props.lib_relative_sub_path
    lib_extended_path = os.path.join(pref.lib_paths[pref.lib_active_index].directory, curr_sub_dir)
    if not garments_lib.FORCE_UPDATE_FOLDERS and lib_extended_path in FOLDER_ITEMS.keys():
        return FOLDER_ITEMS[lib_extended_path]

    # print(f'Updating. Force update was {FORCE_UPDATE_FOLDERS}')
    items = []
    if os.path.exists(lib_extended_path):
        for fn in os.listdir(lib_extended_path):
            file_path = os.path.join(lib_extended_path, fn)
            if os.path.isdir(file_path):
                relative_sub_dir = file_path[lib_path_len:]
                items.append((relative_sub_dir, '+', fn))
        FOLDER_ITEMS[lib_extended_path] = items
        garments_lib.FORCE_UPDATE_FOLDERS = False
        return FOLDER_ITEMS[lib_extended_path]
    else:
        return items


class LocalMainClothPreset(bpy.types.PropertyGroup):
    def update_preset_prop(self, context):
        # write  cloth presets into patterns
        # precalculate all data required for assign_cloth_preset_to_all_patterns()
        for garment in context.scene.cloth_garment_data:
            # check if any garment pattern uses current local preset
            is_used = False
            for pattern in garment.sewing_patterns[:] + garment.pockets[:]:
                if pattern.cloth_preset == self.name or (pattern.cloth_preset == 'Default' and garment.default_cloth_preset == self.name):
                    is_used = True
                    break
            if is_used:
                assign_cloth_preset_weights_to_all_patterns(garment)

    def update_shared_prop(self, context):
        '''Update garments base preset props (mass, air_damping, etc) if theirs default cloth preset is changed'''
        for garment in context.scene.cloth_garment_data:
            sim_time = garment.sim_time #use for gravity ? or as bake cache time?
            sewing_fac = garment.warmup_factor
            if garment.default_cloth_preset == self.name:
                target_cloth_obj = garment.target_cloth_object
                if target_cloth_obj:
                    for modifier in target_cloth_obj.modifiers:
                        if modifier.type == 'CLOTH':
                            settings = modifier.settings
                            # default garment preset props (shared props)
                            settings.mass = self.mass
                            settings.air_damping = self.air_damping

                            settings.use_sewing_springs = True
                            update_keyframe_prop(target_cloth_obj, 'modifiers["Cloth"].settings.sewing_force_max',
                                                 frame=int(sim_time*sewing_fac),
                                                 max_val=self.sewing_force, min_val=0.4*self.sewing_force)
                            settings.sewing_force_max = self.sewing_force # not needed, since keyframe above?


                            settings.tension_damping = self.tension_damping
                            settings.compression_damping = self.compression_damping
                            settings.shear_damping = self.shear_damping
                            settings.bending_damping = self.bending_damping


    def get_name(self):
        return self.get('helper_name', 'Cloth')

    def set_name(self, value):
        '''Update garments base preset props if theirs default cloth preset is changed'''
        # search for material that has same name as preset and update it too
        if self.get('helper_name') == None:
            self['helper_name'] = 'Cloth'
        old_mat = bpy.data.materials.get(self['helper_name'])
        if old_mat:
            old_mat.name = value
        else:
            get_set_preset_mat(value)
        self['helper_name'] = value


    name: bpy.props.StringProperty(name="Name", description="Preset name", default='Cloth', set=set_name,  get=get_name)
    tension_stiffness: bpy.props.FloatProperty(name="Tension stiffness", description="Tension stiffness", default=15.0, min=0.0, max=1000.0, step = 3, update=update_preset_prop)
    compression_stiffness: bpy.props.FloatProperty(name="Compression stiffness", description="Compression stiffness", default=15.0, min=0.0, max=1000.0, step = 3, update=update_preset_prop)
    shear_stiffness: bpy.props.FloatProperty(name="Shear stiffness", description="Shear stiffness", default=5.0, min=0.0, max=1000.0, step = 3, update=update_preset_prop)
    bending_stiffness: bpy.props.FloatProperty(name="Bending stiffness", description="Bending stiffness", default=0.5, min=0.0, max=1000.0, step = 3, update=update_preset_prop)
    shrink_min: bpy.props.FloatProperty(name="Shrink min", description="Shrink min", default=0.0, min=0.0, max=1.0, update=update_preset_prop)

    # props that are shared (cannot be masked by vert group)
    mass: bpy.props.FloatProperty(name="Mass", description="Mass", default=1.0, min=0.0, max=100.0, step = 3, update=update_shared_prop)
    air_damping: bpy.props.FloatProperty(name="Air damping", description="Air damping", default=1.0, min=0.0, max=10.0, step = 3, update=update_shared_prop)
    sewing_force: bpy.props.FloatProperty(name="Sewing Force", description="Sewing Force", default=2.0, min=0.0, max=10.0, step = 3, update=update_shared_prop)

    tension_damping: bpy.props.FloatProperty(name="Tension damping", description="Tension damping", default=5.0, min=0.0, max=50.0, step = 3, update=update_shared_prop)
    compression_damping: bpy.props.FloatProperty(name="Compression damping", description="Compression damping", default=5.0, min=0.0, max=50.0, step = 3, update=update_shared_prop)
    shear_damping: bpy.props.FloatProperty(name="Shear damping", description="Shear damping", default=5.0, min=0.0, max=50.0, step = 3, update=update_shared_prop)
    bending_damping: bpy.props.FloatProperty(name="Bending damping", description="Bending damping", default=0.5, min=0.0, max=1000.0, step = 3, update=update_shared_prop)



class SceneGarmentProps(bpy.types.PropertyGroup):
    lib_relative_sub_path: bpy.props.StringProperty(name="Folder Path", subtype='DIR_PATH', description='Holds library subfolder', default="")
    lib_previews: bpy.props.EnumProperty(items=enum_previews_from_directory_items, description='Holds previews')
    export_merged_name: bpy.props.StringProperty(name="Merged name", description="When saving multiple garments, define output name", default="Merged")
    save_with_screen_grab: bpy.props.BoolProperty(name="Save with Screen Grab", description="Save with Screen Grab", default=True)
    export_garments_li: bpy.props.EnumProperty(name="Export Garment", description="Choose garment to export", items=garment_items, options={"ENUM_FLAG"})
    bool_folders: bpy.props.EnumProperty(name='Folders select', description='Checkbox for folders to perform operation on',
                                         items=folder_items,
                                         options={'ENUM_FLAG'})
    shape_project: bpy.props.PointerProperty(type=ProjectObjectProps)
    # continuous_sim: bpy.props.BoolProperty(name="Continuous simulation", description="Simulate cloth without looping", default=False)
    pull_force: bpy.props.IntProperty(name="Pull force", description="Pull force", default=50, min=0, soft_max=100)
    pull_falloff: bpy.props.FloatProperty(name="Pull falloff", description="Pull falloff", default=0.5, min=0.0, max=1.0, step = 2, subtype='FACTOR')
    distance_max: bpy.props.FloatProperty(name="Distance Max", description="Distance Max", default=0.1, min=0.001, soft_max=1.0, step = 3, subtype='DISTANCE')
    use_max_distance: bpy.props.BoolProperty(name="Use max distance", description="Use max distance", default=True)
    animation_method: bpy.props.EnumProperty(name='Animation method', description='Describe do you want to interact with cloth',
        items=[
            ('CONSTANT', 'Permanent', 'Animation is played continuously'),
            ('INTERRUPTED', 'Discontinuous', 'Animation is interrupted when not dragging the cloth')
        ], default='CONSTANT')

    local_cloth_presets: bpy.props.CollectionProperty(name='Local Cloth Presets',type=LocalMainClothPreset)
    active_lpreset_index: bpy.props.IntProperty(name='Active Local Cloth Preset',default=0,min=0,max=10)


class ObjectGarmentProps(bpy.types.PropertyGroup):
    def check_res(self,context):
        self.predicted_out_vert_count = get_predicted_vert_count(context, self)

    def draw_sewings_update(self, context):
        check_enable_draw_sewings(self, context)

    def update_preset_prop(self, context):
        # write  cloth presets into patterns
        # precalculate all data required for assign_cloth_preset_to_all_patterns()
        assign_cloth_preset_weights_to_all_patterns(garment=self)

    def name_update(self, context):
        fix_garment_colliding_names(self, context)
        old_name = self.target_cloth_object.name
        self.target_cloth_object.name = self.name
        # seee if we have collection with old name and rename it too
        old_col = bpy.data.collections.get(old_name)
        if old_col and old_col in self.target_cloth_object.users_collection[:]:
            old_col.name = self.name

    def toggle_patterns_visibility(self, context):
        for pattern in self.sewing_patterns[:]+self.pockets[:]:
            if pattern.pattern_obj:
                pattern.pattern_obj.hide_viewport = not self.show_patterns
                pattern.pattern_obj.hide_set(not self.show_patterns)
                pattern.pattern_obj.hide_render = not self.show_patterns
        if pattern:
            obj_col = pattern.pattern_obj.users_collection[0]
            if obj_col and self.show_patterns: # only for unhide - make sure coll is visible
                for ancestor in get_ancestors_collections(context, obj_col)[:-1]: # skip scene coll
                    ancestor.hide_viewport = False
                    # unhide view layer too
                    vcol = find_view_coll(context.view_layer.layer_collection, ancestor.name)  # find in view_layer cols (! same as bpy.data.colls...)
                    vcol.hide_viewport = False



    name: bpy.props.StringProperty(name="Garment name", description = 'This name will be given to generated simulation mesh', default="Garment", update=name_update)
    target_cloth_object: bpy.props.PointerProperty(name='Output',  description='Target simulated mesh object', type=bpy.types.Object)
    sewings_overlay: bpy.props.BoolProperty(name="Draw sewings overlay", description="Draw sewings overlay with segments lengths. White sewings connect segments of equal length", default=False, update=draw_sewings_update)
    smooth_strength: bpy.props.IntProperty(name="Smooth", description="Smoothing applied to garment, when converting to mesh (default-10 works best)", default=10, min=0, max=10, step=1)
    sim_time: bpy.props.IntProperty(name="Simulation startup time", description="Describes the frame on which gravity will reach 100%. Bigger values, prevets cloth from falling too fast on the floor", default=50, min=10, soft_max=150, step=1)
    warmup_factor: bpy.props.FloatProperty(name="Sewing Factor", description="How fast sewing force (and shrink factor) will reach 100%, as factor of 'Simulation Startup' time", default=0.4, min=0.1, max=1.0, step = 2, subtype='FACTOR')

    predicted_out_vert_count: bpy.props.IntProperty(name="Predicted cloth point count", description='Predicted cloth point count' )
    cloth_res: bpy.props.IntProperty(name="Cloth Resolution", description='Triangulation resolution - N points / meter. \
        \nVery low values often give corrupted triangulation result \
        \nTriangle icons - warns that output polycount is high (converting to mesh may take some time)', default=25, min=2, soft_max=80, update = check_res)
    output_topology: bpy.props.EnumProperty(name="Topology", default="TRIS",
                                            items=(("TRIS", "Tris", "Generated triangle mesh"),
                                                   ("QUAD", "Quads", "Generated quad mesh"),
                                             ))
    merge_patterns: bpy.props.BoolProperty(name='Merge Generated Patterns', description='Merge Generated Patterns and sewing after convertign them to meshes',  default=True)

    sewing_patterns: bpy.props.CollectionProperty(name='Sewing Patterns', type=PatternPropGroup)
    show_patterns: bpy.props.BoolProperty(name='Show Patterns', description='Show Patterns', default=True, update=toggle_patterns_visibility)

    enable_pockets:  bpy.props.BoolProperty(name='Enable Pockets', default=True, description='Enable Pockets')
    pockets: bpy.props.CollectionProperty(type=PocketCurve)
    enable_pins:  bpy.props.BoolProperty(name='Enable pins', default=True, description='Enable pins')
    pins: bpy.props.CollectionProperty(type=GarmentPin)

    generated_vgroups: bpy.props.CollectionProperty(type=PatternVertGroups)
    # add_border_mask:  bpy.props.BoolProperty(name='Generate Border Mask', default=True, description='Generate border vertex group mask')
    # add_non_border_mask:  bpy.props.BoolProperty(name='Generate non Border Mask', default=True, description='Generate non border vertex group Mask')

    garment_sewings: bpy.props.CollectionProperty(type=SegmentIndexCollection)

    default_cloth_preset: bpy.props.EnumProperty(name="Default Preset", description="Default Cloth Preset applied to whole garment", items=cloth_local_default_presets, update=update_preset_prop)



# bpy.utils.register_class(GTOOL_PG_ObjectGarment)
def develte_vert_menu(self, context):
    self.layout.separator()
    self.layout.operator("curve.curve_del_vert_gt")

def subdivide_segment_menu(self, context):
    self.layout.separator()
    self.layout.operator("curve.curve_subdivide_gt")
    self.layout.operator('curve.connect_points')

def update_index(self, context):
    ''' Just select pattern object when changing index'''
    if len(context.scene.cloth_garment_data) == 0:
        return
    garment = context.scene.cloth_garment_data[context.scene.garment_index]
    for o in context.scene.objects:
        o.select_set(False)
    for pattern in garment.sewing_patterns[:] + garment.pockets[:]:
        if pattern.pattern_obj:
            pattern.pattern_obj.select_set(True)
            context.view_layer.objects.active = pattern.pattern_obj

def register_props():
    bpy.types.Scene.cloth_garment_data = bpy.props.CollectionProperty(type=ObjectGarmentProps)
    bpy.types.Scene.garment_index = bpy.props.IntProperty(name="Garment index", default=0, min=0, max=10, update=update_index)
    bpy.types.Scene.garment_props = bpy.props.PointerProperty(type=SceneGarmentProps)
    bpy.types.VIEW3D_MT_edit_curve_delete.append(develte_vert_menu)
    bpy.types.VIEW3D_MT_edit_curve_context_menu.append(subdivide_segment_menu)

def unregister_props():
    bpy.types.VIEW3D_MT_edit_curve_context_menu.remove(subdivide_segment_menu)
    bpy.types.VIEW3D_MT_edit_curve_delete.remove(develte_vert_menu)
    del bpy.types.Scene.cloth_garment_data
    del bpy.types.Scene.garment_props
