'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import gpu
import blf
import bpy
import mathutils
import bmesh
from mathutils.bvhtree import BVHTree
from mathutils import Vector
from bpy_extras import view3d_utils
from gpu_extras.batch import batch_for_shader
from os import path
import logging

from .utils.fsm_oper import ModalCommon, CookieCutter_FSM
from .utils.helper_functions import draw_text_line, raycast_bvh, get_obj_mesh_bvht,\
    raycast_bvh, import_node_group, is_name_r_l, get_mirrored_name

# logging.basicConfig(format='%(lineno)d - %(message)s Dupa', level=logging.DEBUG)  # wtf wont work
logger = logging.getLogger()
logger.disabled = True

bend_deform_node_group = 'BendDeform_GN'
short_hair_mod_name = 'Bend Deform GT'
BEND_DEFORM_GT_FILE = path.dirname(__file__)+"/bend_deform_gt.blend"


class GTOOL_OT_EditBending(bpy.types.Operator, ModalCommon, CookieCutter_FSM):
    """Modal object selection with a ray cast"""
    bl_idname = "cloth.edit_bend"
    bl_label = "Edit Bend"
    bl_description = "Edit Bend"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH' and context.active_object.modifiers.get(short_hair_mod_name)

    @staticmethod
    def draw_text(self, context):
        draw_text_lines = []
        draw_text_lines.append(["[LMB] Drag - Bend", 'p'])
        draw_text_lines.append(["[ENTER] Confirm", 'p'])
        draw_text_lines.append(["[ESC/RMB] Cancel", 'p'])
        draw_text_line(draw_text_lines)

    @staticmethod
    def draw_px(self, context):
        return

    def update_pattern_bend(self, context, fac, pattern_name):
        '''Store bend factor - so that we can regenerate patterns with the restored bend value'''
        garment_index = context.scene.garment_index
        garment = context.scene.cloth_garment_data[garment_index]
        for pattern in garment.sewing_patterns:
            if pattern.pattern_obj and pattern.pattern_obj.name == pattern_name:
                pattern.bend_factor = fac

    def modal(self, context, event):
        obj = context.active_object
        if self.state == 'INIT':
            context.workspace.status_text_set('[LBM] - Bend Pattern       [RMB] - Cancel        [ENTER] - Confirm')
            if event.type == 'LEFTMOUSE' and event.value == 'PRESS':
                self.init_mouse_co_2d = Vector((event.mouse_region_x, event.mouse_region_y))
                face_hit_loc, normal, index, distance = raycast_bvh(self.obj_bvh, context, event)
                if face_hit_loc:
                    face = context.active_object.data.polygons[index]
                    pattern = self.vert_to_pattern_name[face.vertices[0]] # aka vert group
                    bend_attrib = obj.data.attributes.get('GT_Bend_Factor')
                    self.init_factor = bend_attrib.data[face.vertices[0]].value
                    if is_name_r_l(pattern):
                        mirrored_name = get_mirrored_name(pattern)


                    self.pattern_name = pattern
                    logging.debug(f'LMB PRESS from Main')
                    if face_hit_loc:
                        self.state = 'DRAGGING'

        elif self.state == 'DRAGGING':
            if event.type == 'MOUSEMOVE':
                current_mouse_co_2d = Vector((event.mouse_region_x, event.mouse_region_y))
                diff = current_mouse_co_2d.x - self.init_mouse_co_2d.x
                length = max(min(self.init_factor+diff/30, 10), -30)
                if abs(length) < 0.1:
                    length = 0

                bend_factor = obj.data.attributes.get('GT_Bend_Factor')
                if not bend_factor:
                    bend_factor = obj.data.attributes.new('GT_Bend_Factor', 'FLOAT','POINT')
                for v_id in self.pattern_name_to_verts[self.pattern_name]:
                    bend_factor.data[v_id].value = length

                self.update_pattern_bend(context, length, self.pattern_name)

                if is_name_r_l(self.pattern_name):
                    mirrored_name = get_mirrored_name(self.pattern_name)
                    if mirrored_name in self.pattern_name_to_verts.keys():
                        for v_id in self.pattern_name_to_verts[mirrored_name]:
                            bend_factor.data[v_id].value = length
                        self.update_pattern_bend(context, length, mirrored_name)

                context.area.header_text_set(f'Factor: {length:.2f}')

            elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE':
                self.state = 'INIT'
                logging.debug(f'LMB RELEASE from Main')

        if event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        elif event.type in {'RIGHTMOUSE', 'ESC'} and event.value == 'PRESS':
            context.workspace.status_text_set(None)
            context.area.header_text_set(None)
            self.disable_handlers()
            context.area.tag_redraw()
            return {'CANCELLED'}

        elif (event.type == 'RET' or event.type == 'SPACE') and event.value == 'PRESS':
            context.workspace.status_text_set(None)
            context.area.header_text_set(None)
            self.end_commit()
            context.area.tag_redraw()
            return {'FINISHED'}

        return {'RUNNING_MODAL'}


    def before_start(self, context):
        self.state = 'INIT' # 'DRAGGING'
        self.obj_bvh = get_obj_mesh_bvht(context.active_object, context.evaluated_depsgraph_get(), world_space=False)
        garment_index = context.scene.garment_index
        garment = context.scene.cloth_garment_data[garment_index]
        pattern_names = {pattern.pattern_obj.name:True for pattern in garment.sewing_patterns[:]+garment.pockets[:] if pattern.pattern_obj}

        obj = context.active_object
        if obj.name != garment.target_cloth_object.name:
            self.report({'WARNING'}, f'Active object is not output of active garment: {garment.name}. Cancelling')
            return False

        # assign vertice by its vertex group name, to list

        me = obj.data
        bm = bmesh.new()   # create an empty BMesh
        bm.from_mesh(me)   # fill it in from a Mesh

         # Ensure custom data exists.
        bm.verts.ensure_lookup_table()
        bm.verts.layers.deform.verify()

        deform = bm.verts.layers.deform.active

        # for v in bm.verts:
        #     g = v[deform]
        #     # First & second vertex group (by index)
        #     g[0] = 0.5
        #     g[1] = 0.75
        pattern_name_to_verts = {}
        for pname in pattern_names:
            vg = obj.vertex_groups.get(pname)
            if vg:
                g_idx = vg.index
                pattern_name_to_verts[pname] = [v.index for v in bm.verts if v[deform].get(g_idx) and v[deform][g_idx] > 0.9]
        self.pattern_name_to_verts = pattern_name_to_verts

        vert_to_pattern_name = {}
        for vert in bm.verts:
            w_layer = vert[deform]
            for g_idx,w in w_layer.items():
                vg = obj.vertex_groups[g_idx]
                if vg.name in pattern_names and w > 0.9:
                    vert_to_pattern_name[vert.index] = vg.name
                    break

            # for group in w_layer.groups:
            #     gr_name = obj.vertex_groups[group.group].name
            #     if gr_name in pattern_names and group.weight > 0.9:
            #         vert_to_pattern_name[vert.index] = gr_name
                # vg_name = obj.vertex_groups[vert.groups[0].group].name
                # if vg_name in pattern_names:
                #     vert_to_pattern_name[vert.index] = vg_name
        self.vert_to_pattern_name = vert_to_pattern_name
        # bm.to_mesh(me)
        bm.free()  # free and prevent further access
        return True

    def invoke(self, context, event):
        if context.space_data.type == 'VIEW_3D':
            # self.invoke_garment(context, event) # build spatterns outline pts
            if not self.before_start(context):
                return {'CANCELLED'}
            self.enable_draw_handlers(context)
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}


class GTOOL_OT_UpdateBendNodeGroup(bpy.types.Operator):
    bl_idname = "garment.update_bend_node_group"
    bl_label = "Update Bend Node Group"
    bl_description = "Update (or Reset) Bend modifier node group,\nUseful when you want to grab latest version of Bend modifier, after addon update"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        original_node_groups = {node_group.name: node_group for node_group in bpy.data.node_groups}
        seg_guide_ng = import_node_group(bend_deform_node_group, BEND_DEFORM_GT_FILE, force_update=True)
        seg_guide_ng.use_fake_user = True

        # for obj in bpy.data.objects:
        #     if obj.type == 'MESH' and obj.ht_props.short_hair_props.is_short_hair:
        #         short_hair_mod = obj.modifiers.get('ShortHair')
        #         if short_hair_mod:
        #             short_hair_mod['Output_10_attribute_name'] = 'UVMapGN' # Deformer
        #             short_hair_mod['Output_27_attribute_name'] = 'UV_Base' # Deformer
        #             short_hair_mod['Output_14_attribute_name'] = 'ColGN' # Deformer
        return {"FINISHED"}
