'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import bmesh
# from bpy_extras import view3d_utils
# from gpu_extras.batch import batch_for_shader
import logging
from .cloth_presets_lib import get_set_preset_mat, cloth_local_pattern_presets

from .utils.fsm_oper import ModalCommon, CookieCutter_FSM
from .utils.helper_functions import draw_text_line, raycast_bvh, get_obj_mesh_bvht,\
    is_name_r_l, get_mirrored_name, set_weights_uniform, update_keyframe_prop, get_garment


# TODO: implement material assign (and color) when assigning cloth preset
# TODO: implement for curve patterns too?
# TODO: bend for curves too

# logging.basicConfig(format='%(lineno)d - %(message)s Dupa', level=logging.DEBUG)  # wtf wont work
logger = logging.getLogger()
logger.disabled = True

bend_mod_name = 'Bend Deform GT'

def pattern_to_mesh_data(garment):
    pattern_names = {pattern.pattern_obj.name:True for pattern in garment.sewing_patterns[:]+garment.pockets[:] if pattern.pattern_obj}
    obj = garment.target_cloth_object
    me = obj.data
    bm = bmesh.new()   # create an empty BMesh
    bm.from_mesh(me)   # fill it in from a Mesh

    bm.verts.ensure_lookup_table()
    bm.verts.layers.deform.verify()

    deform = bm.verts.layers.deform.active

    pattern_name_to_verts = {}
    for pname in pattern_names:
        pattern_name_to_verts[pname] = []
        vg = obj.vertex_groups.get(pname)
        if vg:
            g_idx = vg.index
            pattern_name_to_verts[pname] = [v.index for v in bm.verts if v[deform].get(g_idx) and v[deform][g_idx] > 0.9]

    pattern_name_to_faces = {}
    for pname in pattern_names:
        pattern_name_to_faces[pname] = []
        vg = obj.vertex_groups.get(pname)
        if vg:
            g_idx = vg.index
            pattern_name_to_faces[pname] = [f.index for f in bm.faces if f.verts[0][deform].get(g_idx) and f.verts[0][deform][g_idx] > 0.9]

    vert_to_pattern_name = {}
    for vert in bm.verts:
        w_layer = vert[deform]
        for g_idx,w in w_layer.items():
            vg = obj.vertex_groups[g_idx]
            if vg.name in pattern_names and w > 0.9:
                vert_to_pattern_name[vert.index] = vg.name
                break

    face_to_pattern_name = {}
    for face in bm.faces:
        w_layer = face.verts[0][deform]
        for g_idx,w in w_layer.items():
            vg = obj.vertex_groups[g_idx]
            if vg.name in pattern_names and w > 0.9:
                face_to_pattern_name[face.index] = vg.name
                break

    # bm.to_mesh(me)
    bm.free()  # free and prevent further access
    return pattern_name_to_verts, pattern_name_to_faces, vert_to_pattern_name, face_to_pattern_name

def remap_to_01(val, min_val, max_val):
    if min_val == max_val:
        return 1
    return (val - min_val) / (max_val - min_val)

def calculate_cloth_preset_vg_weights(garment):
    """

    Args:
        garment ():

    Returns:  preset_normalized_weights (['structural_stiffness']= 0.1), max_cloth_params ( ['structural_stiffness']= 30)

    vgroup(0.0 ; 1.0) -> (tension, tension_max)

    """
    limited_cloth_props = ('structural_stiffness', 'shear_stiffness', 'bending', 'shrink') # can be affected by vert groups

    gprops = bpy.context.scene.garment_props
    # active_local_preset = gprops.local_cloth_presets[gprops.active_lpreset_index]
    def get_local_preset_props(preset_name):
        for local_preset in gprops.local_cloth_presets:
            if local_preset.name == preset_name:
                return local_preset

    # check if any cloth preset is assigned to default_cloth_preset
    default_preset_is_none = garment.default_cloth_preset == 'None'
    patterns_using_default = [pattern.cloth_preset == 'Default' for pattern in garment.sewing_patterns[:] + garment.pockets[:]]
    if default_preset_is_none:
        if all(patterns_using_default): # no cloth preset assigned to any pattern...
            print('No cloth preset assigned to any pattern. Cloth settings wont be chagned')
            return None, None, None # nothing to do...
        else: # some patterns are using default preset, some not..
            # in that case - set default_cloth_preset to first pattern using local preset
            first_local_presset = next((pattern.cloth_preset for pattern in garment.sewing_patterns[:] + garment.pockets[:] if pattern.cloth_preset != 'Default'), None)
            garment.default_cloth_preset = first_local_presset # using [''] - wont work.. (due to using dynamic enum?)
            # hacky way to avoid recursion - just return None, None, None, (this code will be called again anyway)
            return None, None, None
    else: # default preset exists, but does any pattern use it
        patterns_using_default = [(pattern.cloth_preset == 'Default' or pattern.cloth_preset == garment.default_cloth_preset) for pattern in garment.sewing_patterns[:] + garment.pockets[:]]
        if not any(patterns_using_default): # no pattern is using default preset... does not make sense to use default preset then..
            # we can just use average of all used presets, or pick first cloth physics preset  -> and set it as default....
            # lets just pick most used preset, and set it as default
            used_presets = [pattern.cloth_preset for pattern in garment.sewing_patterns[:] + garment.pockets[:]]
            most_used_preset = max(set(used_presets), key=used_presets.count)
            garment.default_cloth_preset = most_used_preset # using [''] - wont work.. (due to using dynamic enum?)
            # hacky way to avoid recursion - just return None, None, None, (this code will be called again anyway)
            return None, None, None

    # pattern -> cloth preset -> preset_props
    cloth_presets = {} # 'preset_name': {'structural_stiffness': 0.5, 'shear_stiffness': 0.5, 'bending': 0.5, 'shrink': 0.5}
    if garment.default_cloth_preset != 'None' and any(patterns_using_default): # add default preset,  if it is not 'None' and any pattern is using it
        cloth_presets[garment.default_cloth_preset] = get_local_preset_props(garment.default_cloth_preset) # default for all clothes in garment

    for sewing_pattern in garment.sewing_patterns[:] + garment.pockets[:]:
        if sewing_pattern.cloth_preset not in cloth_presets and sewing_pattern.cloth_preset != 'Default':
            # XXX: this will re-run assign_cloth_preset_weights_to_all_patterns - but should nt be a problem
            cloth_presets[sewing_pattern.cloth_preset] = get_local_preset_props(sewing_pattern.cloth_preset)

    # TODO: skip this and return None, None ? if len(cloth_presets) == 0
    if len(cloth_presets) == 0:
        return None, None, None

    # NOTE: calc vert groups weights
    tension_stiffness = []
    compression_stiffness = []
    structural_stiffness = [] # for vert group (shared with tension and compression)
    shear_stiffness = []
    bending = []
    shrink = []
    for preset_name, cloth_preset in cloth_presets.items():
        structural_stiffness.append((cloth_preset.tension_stiffness + cloth_preset.compression_stiffness)/2)
        tension_stiffness.append(cloth_preset.tension_stiffness)
        compression_stiffness.append(cloth_preset.compression_stiffness)
        shear_stiffness.append(cloth_preset.shear_stiffness)
        bending.append(cloth_preset.bending_stiffness)
        shrink.append(cloth_preset.shrink_min)

    # sum_structural_stiffness = sum(structural_stiffness)
    max_structural_stiffness = max(structural_stiffness)
    min_structural_stiffness = min(structural_stiffness) # this is how it works in blender
    ss_norm = [remap_to_01(x,min_structural_stiffness, max_structural_stiffness) for x in structural_stiffness]

    max_tension_stiffness = max(tension_stiffness)
    min_tension_stiffness = min(tension_stiffness)

    max_compression_stiffness = max(compression_stiffness)
    min_compression_stiffness = min(compression_stiffness)

    # sum_shear_stiffness = sum(shear_stiffness)
    max_shear_stiffness = max(shear_stiffness)
    min_shear_stiffness = min(shear_stiffness)
    shear_norm = [remap_to_01(x,min_shear_stiffness,max_shear_stiffness) for x in shear_stiffness]

    # sum_bending = sum(bending)
    max_bending = max(bending)
    min_bending = min(bending)
    bending_norm = [remap_to_01(x, min_bending, max_bending) for x in bending]

    # sum_shrink = sum(shrink)
    max_shrink = max(shrink)
    min_shrink = min(shrink)
    shrink_norm = [remap_to_01(x, min_shrink, max_shrink) for x in shrink]
    # if max_shrink > 0.01: # whi it was like this?
    #     shrink_norm = [x/max_shrink for x in shrink]
    # else:
    #     shrink_norm = [0 for x in shrink] # or 0 ?  shrink * 0 -> no shrink

    # NOTE: this * wg.weight = preset value
    preset_weights_01 = {}  # 0 - stifness,  1 - max_stiffness etc.
    for i,(preset_name, cloth_preset) in enumerate(cloth_presets.items()):
        # pattern_name = sewing_pattern.pattern_obj.name
        preset_weights_01[preset_name] = {
            'structural_stiffness': ss_norm[i],
            'shear_stiffness': shear_norm[i],
            'bending': bending_norm[i],
            'shrink': shrink_norm[i]
        }

    # NOTE: max_cloth_params * wg.weight = preset value
    max_cloth_params = {
        'tension_stiffness': max_tension_stiffness,
        'compression_stiffness': max_compression_stiffness,
        'shear_stiffness': max_shear_stiffness,
        'bending': max_bending,
        'shrink': max_shrink
    }

    min_cloth_params = {
        'tension_stiffness': min_tension_stiffness,
        'compression_stiffness': min_compression_stiffness,
        'shear_stiffness': min_shear_stiffness,
        'bending': min_bending,
        'shrink': min_shrink,
    }

    return preset_weights_01, min_cloth_params, max_cloth_params


def get_default_preset(garment):
    if garment.default_cloth_preset == 'None': # can be any local preset + 'None'
        default_preset = None
    else:
        garment_props = bpy.context.scene.garment_props
        for lpreset in garment_props.local_cloth_presets:
            if lpreset.name == garment.default_cloth_preset:
                default_preset = lpreset
    return default_preset



def write_preset_max_props_to_cloth_mod(garment, merged_obj, min_cloth_params, max_cloth_params):
    ''' Also updates keyframes for 'shrink_max' and 'sewing_force_max' properties
        vgroup(0.0 ; 1.0) * (tension, tension_max) =
    '''

    # get default preset, based in local_cloth_presets
    default_preset = get_default_preset(garment)

    # check if any cloth preset is assigned to default_cloth_preset
    default_preset_is_none = garment.default_cloth_preset == 'None'
    patterns_using_default = [pattern.cloth_preset == 'Default' for pattern in garment.sewing_patterns[:] + garment.pockets[:]]


    sim_time = garment.sim_time #use for gravity ? or as bake cache time?
    sewing_fac = garment.warmup_factor
    for modifier in merged_obj.modifiers:
        if modifier.type == 'CLOTH':
            settings = modifier.settings
            settings.tension_stiffness_max = max_cloth_params['tension_stiffness']      # works with VGroup
            settings.tension_stiffness = min_cloth_params['tension_stiffness']
            settings.compression_stiffness_max = max_cloth_params['compression_stiffness']  # works with VGroup
            settings.compression_stiffness = min_cloth_params['compression_stiffness']
            settings.shear_stiffness_max = max_cloth_params['shear_stiffness']
            settings.shear_stiffness = min_cloth_params['shear_stiffness']
            settings.bending_stiffness_max = max_cloth_params['bending']                   # works with VGroup
            settings.bending_stiffness = min_cloth_params['bending']

            update_keyframe_prop(merged_obj, 'modifiers["Cloth"].settings.shrink_max', frame=int(sim_time*sewing_fac*1.6), max_val=max_cloth_params['shrink'])
            settings.shrink_max = max_cloth_params['shrink'] # redundant? due to keyframe...
            settings.shrink_min = min_cloth_params['shrink']

            settings.vertex_group_structural_stiffness = 'structural_stiffness'
            settings.vertex_group_shear_stiffness = 'shear_stiffness'
            settings.vertex_group_bending = 'bending'
            settings.vertex_group_shrink = 'shrink'

            # default garment preset props (shared props)
            if default_preset: # and any(patterns_using_default): # shared props...
                settings.mass = default_preset.mass
                settings.air_damping = default_preset.air_damping
                update_keyframe_prop(merged_obj, 'modifiers["Cloth"].settings.sewing_force_max',
                                     frame=int(sim_time*sewing_fac),
                                     max_val=default_preset.sewing_force, min_val=0.4*default_preset.sewing_force)
                settings.sewing_force_max = default_preset.sewing_force # redundant? due to keyframe...

                settings.tension_damping = default_preset.tension_damping
                settings.compression_damping = default_preset.compression_damping
                settings.shear_damping = default_preset.shear_damping
                settings.bending_damping = default_preset.bending_damping
            #  XXX: no pattern is using default preset... does not make sense to use default preset then..
            # Handled already in calculate_cloth_preset_vg_weights
            # elif not any(patterns_using_default):
            #     gprops = bpy.context.scene.garment_props
            #     def get_local_preset_props(preset_name):
            #         for local_preset in gprops.local_cloth_presets:
            #             if local_preset.name == preset_name:
            #                 return local_preset
            #     # pattern  (cloth preset_name) -> preset_props
            #     cloth_presets = {} # 'preset_name': {'structural_stiffness': 0.5, 'shear_stiffness': 0.5, 'bending': 0.5, 'shrink': 0.5}
            #     for sewing_pattern in garment.sewing_patterns[:] + garment.pockets[:]:
            #         if sewing_pattern.cloth_preset not in cloth_presets and sewing_pattern.cloth_preset != 'Default':
            #             cloth_presets[sewing_pattern.cloth_preset] = get_local_preset_props(sewing_pattern.cloth_preset)
            #
            #     def aver_prop(prop_name):
            #         return sum([preset[prop_name] for preset in cloth_presets.values()])/len(cloth_presets)
            #     # use averga then from used presets to flll remaining props
            #     settings.mass = aver_prop('mass')
            #     settings.air_damping = aver_prop('air_damping')
            #     sewing_force = aver_prop('sewing_force')
            #     update_keyframe_prop(merged_obj, 'modifiers["Cloth"].settings.sewing_force_max',
            #                          frame=int(sim_time*sewing_fac),
            #                          max_val=sewing_force, min_val=0.4*sewing_force)
            #     settings.sewing_force_max = sewing_force # redundant? due to keyframe...
            #
            #     settings.tension_damping = aver_prop('tension_damping')
            #     settings.compression_damping = aver_prop( 'compression_damping')
            #     settings.shear_damping = aver_prop('shear_damping')
            #     settings.bending_damping = aver_prop( 'bending_damping')



def get_preset_material_idx(cloth_obj, preset_name):
    # search preset_mat in obj material slots
    if not cloth_obj:
        return
    preset_mat = get_set_preset_mat(preset_name)
    in_obj_mat = cloth_obj.data.materials.get(preset_name)
    if not in_obj_mat:
        cloth_obj.data.materials.append(preset_mat)
        mat_idx = len(cloth_obj.data.materials) - 1
    else:
        mat_idx = cloth_obj.data.materials.find(preset_name)
    return mat_idx

def get_matched_pattern(garment, pattern_name):
    '''Get matched pattern or pocket'''
    for pattern in garment.sewing_patterns[:]+garment.pockets[:]:
        if pattern.pattern_obj and pattern.pattern_obj.name == pattern_name:
            return pattern

def set_curve_pattern_material(pattern):
    '''Assign material to pattern curve object'''
    garment_idx = bpy.context.scene.garment_index
    garment = bpy.context.scene.cloth_garment_data[garment_idx]
    cloth_preset = garment.default_cloth_preset if pattern.cloth_preset == 'Default' else pattern.cloth_preset
    pattern_obj = pattern.pattern_obj
    if not pattern_obj:
        return
    if cloth_preset == 'None':
        # clear materials
        pattern_obj.data.materials.clear()
    else:
        preset_mat = bpy.data.materials.get(cloth_preset)
        if preset_mat: # assign to object (not to ob.data)
            if len(pattern_obj.data.materials) == 0:
                pattern_obj.data.materials.append(None)
                pattern_obj.material_slots[0].link = 'OBJECT'
            else:
                pattern_obj.material_slots[0].link = 'OBJECT'
            pattern_obj.material_slots[0].material = preset_mat

def assign_cloth_preset_weights_to_all_patterns(garment):
    '''Assign material[preset] to pattern faces, assign preset to garment.pockets or garment.sewing_patterns
      Also - sets animation keyframes for  'shrink_max' and 'sewing_force_max' properties
    '''
    print(f'Assigning materials to all patterns')
    for pattern in garment.sewing_patterns[:]+garment.pockets[:]:
        set_curve_pattern_material(pattern)

    cloth_obj = garment.target_cloth_object
    if not cloth_obj:
        return

    # XXX: calc_preset_vg  may set pattern.cloth_preset -> thus re-execute cloth update
    preset_normalized_weights, min_cloth_params, max_cloth_params = calculate_cloth_preset_vg_weights(garment)

    if preset_normalized_weights is None: # no cloth preset assigned to any pattern..
        return
    pattern_name_to_verts, pattern_name_to_faces, _, _ = pattern_to_mesh_data(garment)

    for pattern in garment.sewing_patterns[:]+garment.pockets[:]:
        pattern_name = pattern.pattern_obj.name
        print(f"Pattern {pattern_name}: "+ pattern.cloth_preset)

        cloth_preset = garment.default_cloth_preset if pattern.cloth_preset == 'Default' else pattern.cloth_preset
        # if cloth_preset == 'None': # if garment.default_cloth_preset  == 'None'
        #     continue

        mat_idx = get_preset_material_idx(cloth_obj, cloth_preset)
        for face_idx in pattern_name_to_faces[pattern_name]:
            cloth_obj.data.polygons[face_idx].material_index = mat_idx

        for vg_name, vg_weight in preset_normalized_weights[cloth_preset].items():
            set_weights_uniform(cloth_obj, vg_name, vg_weight, v_ids = pattern_name_to_verts[pattern_name])

    write_preset_max_props_to_cloth_mod(garment, cloth_obj, min_cloth_params, max_cloth_params) # does the param animation too
    # TODO: handle if no cloth preset is assigned to pattern...
    # XXX: what if pattern does not use any cloth preset? : do nothing, but what is half of patterns are using some preset, half not?


class GTOOL_OT_SwitchPatternPreset(bpy.types.Operator):
    bl_idname = "garment.switch_pattern_preset"
    bl_label = "Switch Pattern Preset"
    bl_description = "Switch Pattern Preset"
    bl_options = {'REGISTER', 'UNDO'}

    pattern_index: bpy.props.IntProperty(default=0)
    preset: bpy.props.EnumProperty(name="Cloth Preset", description="Assign Cloth Preset to selected pattern", items=cloth_local_pattern_presets)

    def execute(self, context):
        garment = get_garment(context)
        pattern = garment.sewing_patterns[self.pattern_index]
        pattern.cloth_preset = self.preset # this should update vgroups, and materials
        # assign_cloth_preset_weights_to_all_patterns(garment)
        return {'FINISHED'}


# TODO: make sure pockets are handled in materials presets
# TODO: update opengl drawing to b 4.0 specs...

class GTOOL_OT_AssignClothPreset(bpy.types.Operator, ModalCommon):
    bl_idname = "garment.assign_cloth_preset"
    bl_label = "Assign Cloth Preset"
    bl_description = "Assign Cloth Preset to selected sewing pattern.\nClick with LMB on pattern, to assign active preset"
    bl_options = {'REGISTER', 'UNDO'}

    preset: bpy.props.EnumProperty(name="Cloth Preset", description="Assign Cloth Preset to selected pattern", items=cloth_local_pattern_presets)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type in {'MESH', 'CURVE'} # and context.active_object.modifiers.get(bend_mod_name)

    @staticmethod
    def draw_text(self, context):
        draw_text_lines = []
        draw_text_lines.append([f"[LMB] - Assign Preset: [{self.preset}] to: ", 'p'])
        draw_text_lines.append([f"[{self.hover_pattern_name}]", 'p'])
        draw_text_lines.append(["[ESC/RMB] Quit", 'p'])
        draw_text_line(draw_text_lines)

    @staticmethod
    def draw_px(self, context):
        return

    def modal(self, context, event):
        if event.type == 'LEFTMOUSE' and event.value == 'PRESS':
            pattern = None
            if context.active_object.type == 'MESH': # for garment.target_cloth_object - aka output mesh
                pattern_name = self.hover_pattern_name
                pattern = get_matched_pattern(self.garment, pattern_name)

            else: # must be curve pattern
                hit_c, loc_w, curve_obj = self.find_curve_under_cursor(event, curves_2d=True, curves_3d=True)
                if hit_c:  # picking pocket obj
                    pattern_id = self.get_pattern_from_curve(curve_obj)
                    if pattern_id >= 0:
                        pattern = self.garment.sewing_patterns[pattern_id]
                        pattern_name = pattern.pattern_obj.name

            if pattern:
                pattern.cloth_preset = self.preset   # this update vgroups, and materials through pattern.cloth_presets - update()
                self.report({'INFO'}, f'Assigned Cloth Preset: {self.preset} to {pattern_name}')

                if is_name_r_l(pattern_name):
                    mirrored_name = get_mirrored_name(pattern_name)
                    mirrored_pattern = get_matched_pattern(self.garment, mirrored_name)
                    if mirrored_pattern:
                        mirrored_pattern.cloth_preset = self.preset   # NOTE: this update vgroups, and materials
            else:
                return {'PASS_THROUGH'}

        elif event.type == 'MOUSEMOVE':
            if context.active_object.type == 'MESH':
                face_hit_loc, normal, index, distance = raycast_bvh(self.obj_bvh, context, event)
                self.hover_pattern_name = self.face_to_pattern_name[index] if face_hit_loc else ''
                if self.prev_pattern_name != self.hover_pattern_name:
                    self.prev_pattern_name = self.hover_pattern_name
                    context.area.tag_redraw()


        if event.type in {'RIGHTMOUSE', 'ESC'} and event.value == 'PRESS':
            context.workspace.status_text_set(None)
            context.area.header_text_set(None)
            self.disable_handlers()
            context.area.tag_redraw()
            return {'CANCELLED'}

        if event.type in ('RET', 'SPACE') and event.value == 'PRESS':
            context.workspace.status_text_set(None)
            context.area.header_text_set(None)
            self.end_commit()
            context.area.tag_redraw()
            return {'FINISHED'}

        if event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}


    def before_start(self, context, event):
        cloth_obj = context.active_object

        # NOTE: disable constructive modifiers first (if any), on active obj - so bvh will be built on non-modified mesh
        constructive_modifiers = ['SUBSURF', 'BEVEL', 'MIRROR', 'MASK', 'EDGE_SPLIT', 'SOLIDIFY']
        for mod in cloth_obj.modifiers:
            if mod.type in constructive_modifiers:
                mod.show_viewport = False

        if context.active_object.type == 'MESH':
            self.obj_bvh = get_obj_mesh_bvht(cloth_obj, context.evaluated_depsgraph_get(), world_space=False)

        self.garment = context.scene.cloth_garment_data[self.garment_index]
        self.hover_pattern_name = ''
        self.prev_pattern_name = ''

        if cloth_obj.type == 'MESH' and cloth_obj != self.garment.target_cloth_object:
            self.report({'WARNING'}, 'Active object seems to not be the garment')
            return False
        if cloth_obj.type == 'MESH':
            self.pattern_name_to_verts, self.pattern_name_to_faces, self.vert_to_pattern_name, self.face_to_pattern_name = pattern_to_mesh_data(self.garment)
        else: # just assume we do it for actiev garment sewing patterns (curves)
            self.calc_pattern_segments_cache(context) # for patter_segments_cache - if we work on curves
        return True

    def invoke(self, context, event):
        self.garment_index = self.garment_index_from_obj(context, context.active_object)
        if self.garment_index < 0: # fallback to active garment
            self.garment_index = context.scene.garment_index
        else:
            context.scene['garment_index'] = self.garment_index # [''], to avoid auto running of 'update prop' method
        if context.space_data.type == 'VIEW_3D':
            # self.invoke_garment(context, event) # build spatterns outline pts
            if not self.before_start(context, event):
                return {'CANCELLED'}
            self.enable_draw_handlers(context)
            context.workspace.status_text_set('[LBM] - Pick Pattern       [RMB] - Cancel        [ENTER] - Confirm')
            context.area.tag_redraw()
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        self.report({'WARNING'}, "Active space must be a View3d")
        return {'CANCELLED'}



class GTOOL_OT_RemoveLocalPreset(bpy.types.Operator):
    bl_idname = "cloth.delete_local_preset"
    bl_label = "Delete preset"
    bl_description = "Delete Local Cloth preset"
    bl_options = {'REGISTER', 'UNDO'}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        scn_gt_props = context.scene.garment_props
        local_preset = scn_gt_props.local_cloth_presets[scn_gt_props.active_lpreset_index]
        self.layout.label(text=f"Do you want to remove '{local_preset.name}' preset?")

    def execute(self, context):
        scn_gt_props = context.scene.garment_props
        active_idx = scn_gt_props.active_lpreset_index
        local_preset = scn_gt_props.local_cloth_presets[active_idx]
        rem_preset_name = local_preset.name
        for garment in context.scene.cloth_garment_data:
            cloth_obj = garment.target_cloth_object
            rem_mat_idx = get_preset_material_idx(cloth_obj, rem_preset_name)

            for pattern in garment.sewing_patterns[:]+garment.pockets[:]:
                if pattern.cloth_preset == rem_preset_name:
                    # use [''] - to avoid auto running of 'update prop' method
                    pattern['cloth_preset'] = 'Default' # DONE: what if default was removed? - points to none

            if garment.default_cloth_preset == rem_preset_name: # ??
                # DONE: maybe add None into default_cloth_preset list?
                garment['default_cloth_preset'] = 'None'

            assign_cloth_preset_weights_to_all_patterns(garment)

            if cloth_obj:
                cloth_obj.data.materials.pop(index=rem_mat_idx)

        preset_mat = bpy.data.materials.get(rem_preset_name)
        if preset_mat:
            bpy.data.materials.remove(preset_mat)

        scn_gt_props.local_cloth_presets.remove(active_idx)
        scn_gt_props.active_lpreset_index = min(active_idx, len(scn_gt_props.local_cloth_presets)-1)
        self.report({'INFO'}, f'Preset {rem_preset_name} removed successfully')
        return {'FINISHED'}


# add new preset - by duplicating existing one
class GTOOL_OT_AddLocalPreset(bpy.types.Operator):
    bl_idname = "cloth.add_local_preset"
    bl_label = "Add preset"
    bl_description = "Add Local Cloth preset"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        scn_gt_props = context.scene.garment_props
        active_idx = scn_gt_props.active_lpreset_index
        if active_idx < len(scn_gt_props.local_cloth_presets) and active_idx >= 0:
            local_preset = scn_gt_props.local_cloth_presets[active_idx]
            new_preset = scn_gt_props.local_cloth_presets.add()
            new_preset.name = local_preset.name + '.001'
            new_preset.mass = local_preset.mass
            new_preset.air_damping = local_preset.air_damping
            new_preset.sewing_force = local_preset.sewing_force
            new_preset.tension_damping = local_preset.tension_damping
            new_preset.compression_damping = local_preset.compression_damping
            new_preset.shear_damping = local_preset.shear_damping
            new_preset.bending_damping = local_preset.bending_damping
            new_preset.tension_stiffness = local_preset.tension_stiffness
            new_preset.compression_stiffness = local_preset.compression_stiffness
            new_preset.shear_stiffness = local_preset.shear_stiffness
            new_preset.bending_stiffness = local_preset.bending_stiffness
            new_preset.shrink_min = local_preset.shrink_min
        else:
            new_preset = scn_gt_props.local_cloth_presets.add()
        get_set_preset_mat(new_preset.name)
        scn_gt_props.active_lpreset_index = len(scn_gt_props.local_cloth_presets) - 1
        return {'FINISHED'}
