'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import gpu
from gpu_extras.batch import batch_for_shader
from .utils.helper_functions import draw_text_line, create_empty, set_parent, link_to_same_collection, clear_parent
from .utils.fsm_oper import ModalCommon, CookieCutter_FSM
from mathutils import Vector
from mathutils.geometry import intersect_point_line, intersect_line_plane
from bpy_extras import view3d_utils

import time
import logging

# logging.basicConfig(format='%(lineno)d - %(message)s Dupa', level=logging.DEBUG)  # wtf wont work
logger = logging.getLogger()
logger.disabled = True

class PropsCache(): #hack to hold data in modal func
    pass

class GTOOL_OT_RayCastEditPins(bpy.types.Operator, ModalCommon, CookieCutter_FSM):
    """Modal object selection with a ray cast"""
    bl_idname = "garment.raycast_edit_pins"
    bl_label = "Edit Pins"
    bl_description = "Similar to sewings, except Pins are connecting patterns with single stitch"
    bl_options = {'REGISTER', 'UNDO'}

    pin_idx: bpy.props.IntProperty(default=0)

    current_pin_pos = None  # if None -> we do not have the pin pos yet

    @staticmethod
    def draw_text(self, context):
        garment = context.scene.cloth_garment_data[self.garment_index]
        draw_text_lines = []
        if self._state == 'main':
            draw_text_lines.append(["[LMB] Connect 2 points on surface", 'p'])
            draw_text_lines.append(["[DEL/R] Remove Pin", 'p'])
            draw_text_lines.append(["[ENTER] Confirm", 'p'])
        elif self._state == 'edit_pin_end':
            draw_text_lines.append(["[LMB] Pick second point", 'H1'])
        draw_text_lines.append(["[ESC/RMB] Cancel", 'p'])
        # draw_text_lines.append(["Editing pin ID: " + str(self.hover_pin_idx+1) + "/" + str(len(garment.pins)), 'p'])
        draw_text_line(draw_text_lines)


    @staticmethod
    def draw_px(self, context):
        shader = self.shader
        garment = context.scene.cloth_garment_data[self.garment_index]
        if len(garment.pins) == 0:
            return

        gpu.state.blend_set("ALPHA")
        gpu.state.point_size_set(10)
        gpu.state.line_width_set(3.0)
        gpu.state.depth_test_set("LESS")

        current_pin = garment.pins[self.hover_pin_idx] if self.hover_pin_idx != -1 else None
        if self._state == 'edit_pin_end':
            pin_pos = []
            if current_pin:
                if self.which_end == 'SOURCE' and current_pin.target_co:  # we edit src so try getting target_co
                    pin_pos.append(current_pin.target_obj.matrix_world @ Vector(current_pin.target_co))
                if self.which_end == 'TARGET' and current_pin.source_co: # we edit target so try getting src_co
                    pin_pos.append(current_pin.source_obj.matrix_world @ Vector(current_pin.source_co))
            if self._prop and self._prop.pin_pos:
                pin_pos.append(self._prop.pin_pos) #is in world space

            if pin_pos:
                batch = batch_for_shader(shader, "POINTS", {"pos": pin_pos})
                shader.bind()
                shader.uniform_float("color", (0.2, 0.9, 0.2, 0.5))
                batch.draw(shader)

                batch = batch_for_shader(shader, 'LINES', {"pos": pin_pos})
                shader.uniform_float("color", (1.0, 0.8, 0.0, 0.5))
                batch.draw(shader)

        points = []
        # drawing all pins
        for i, pin in enumerate(garment.pins):
            #skip drawing if we are editing pin pin_idx
            if i == self.hover_pin_idx and self._state == 'edit_pin_end':
                continue
            if pin.source_obj and pin.target_obj:
                verts_form = pin.source_obj.matrix_world @ Vector(pin.source_co)
                verts_to = pin.target_obj.matrix_world @ Vector(pin.target_co)

            points = [verts_form[:], verts_to[:]]
            shader.bind()
            batch = batch_for_shader(shader, 'LINES', {"pos": points})
            if self.hover_pin_idx == i and self.which_end not in {'SOURCE','TARGET'}:  # draw currently defined pin
                col = (0.2, 0.9, 0.2, 0.5)
            else:
                col = (1.0, 0.8, 0.0, 0.5)

            shader.uniform_float("color", col)
            batch.draw(shader)
            batch = batch_for_shader(shader, 'POINTS', {"pos": points})
            if self.hover_pin_idx == i and self.which_end in {'SOURCE', 'TARGET'}:  # draw currently defined pin
                col = (0.2, 0.9, 0.2, 0.5)
            else:
                col = (1.0, 0.8, 0.0, 0.5)
            shader.uniform_float("color", col)
            batch.draw(shader)
        # batch = shader.new_batch('LINES', {"pos": points})

        #restore opengl defaults
        gpu.state.depth_test_set("NONE")
        gpu.state.line_width_set(1)
        gpu.state.point_size_set(1)
        gpu.state.blend_set("NONE")
        # bgl.glDisable(bgl.GL_LINE_STIPPLE)

    def get_visible_pins(self):
        return [idx for idx, pin in enumerate(self.garment.pins)]  # if pin.source_obj in self.context.visible_objects and (pin.target_obj in self.context.visible_objects)


    def set_pin_end(self, new_pin_obj, co):
        logging.debug(f'setting pin {self.which_end } to obj: {new_pin_obj}')
        pin = self.garment.pins[self.hover_pin_idx]
        old_pin_obj = pin.source_obj if self.which_end == 'SOURCE' or self.which_end == 'BOTH' else pin.target_obj

        #* we are re-pining from empty to curve - so del(old_empty)
        if old_pin_obj and old_pin_obj.type == 'EMPTY' and new_pin_obj and new_pin_obj.type == 'CURVE':
            bpy.data.objects.remove(old_pin_obj)

        if new_pin_obj is None: #we are pining to empty space (no parenting then)
            if old_pin_obj and old_pin_obj.type == 'EMPTY': #use old empty as pin
                new_pin_obj = old_pin_obj # we wont be overriding old obj
                clear_parent(new_pin_obj)
            else:
                empty = create_empty('Pin')
                new_pin_obj = empty

            #DONE: handle collections
            #move to same collections as other end if possibel
            if (self.which_end == 'SOURCE' or self.which_end == 'BOTH') and pin.target_obj:
                link_to_same_collection(pin.target_obj, new_pin_obj)
            elif (self.which_end == 'TARGET') and pin.source_obj:
                link_to_same_collection(pin.source_obj, new_pin_obj)
            else: #no othwer way to predict where to link new empty obj. Maybe scene col?
                if len(self.garment.sewing_patterns):
                    link_to_same_collection(self.garment.sewing_patterns[0].pattern_obj, new_pin_obj)
                else:
                    self.context.scene.collection.objects.link(empty)
                # self.context.collection.objects.link(empty)

            new_pin_obj.matrix_world.translation = co
            co = Vector((0,0,0)) #for empties - pin_co is stored in empty loc

        elif new_pin_obj.type == 'MESH': #we will pin to mesh surface using empty
            if old_pin_obj and old_pin_obj.type == 'EMPTY':  # use old empty as pin
                old_pin_obj.matrix_world.translation = co
                set_parent(old_pin_obj, new_pin_obj)  # re-parent old_empty to mesh (new_pin_obj)
                new_pin_obj = old_pin_obj  # we wont be overriding old obj
            else:
                empty = create_empty('Pin')
                empty.matrix_world.translation = co
                link_to_same_collection(new_pin_obj, empty)  # move to same collection as mesh we are pining too
                set_parent(empty, new_pin_obj)  # re-parent to mesh
                new_pin_obj = empty
            co = Vector((0,0,0)) #for empties - pin_co is stored in empty loc

        else: #curve then
            self.assign_target_obj_to_garment(new_pin_obj)
            new_pin_obj.select_set(True)
            co = new_pin_obj.matrix_world.inverted() @ co
            co.z = 0

        if self.which_end == 'SOURCE' or self.which_end == 'BOTH':
            pin.source_obj = new_pin_obj
            pin.source_co = co
        if self.which_end == 'TARGET':
            pin.target_obj = new_pin_obj
            pin.target_co = co



    def find_closest_pin(self):
        ''' Return index of sewing closest to mouse'''
        region = self.context.region
        rv3d = self.context.region_data
        scene = self.context.scene
        region = self.context.region
        rv3d = self.context.region_data
        mouse_co_2d = Vector((self.event.mouse_region_x, self.event.mouse_region_y))

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, mouse_co_2d)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, mouse_co_2d)

        # cast rays and find the closest object
        visible_pins_ids = self.get_visible_pins()
        distances = []

        mouse_2d = Vector(mouse_co_2d)
        for pin_id in visible_pins_ids:
            pin = self.garment.pins[pin_id]
            # if i == self.hover_pin_idx and self._state == 'edit_pin_end':
            #     continue
            src_co = pin.source_obj.matrix_world @ Vector(pin.source_co)
            from_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, src_co)
            target_co = pin.target_obj.matrix_world @ Vector(pin.target_co)
            to_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, target_co)

            if (from_2d - mouse_co_2d).length < 8:
                return pin_id, 'SOURCE', src_co

            if (to_2d - mouse_co_2d).length < 8:
                return pin_id, 'TARGET', target_co

            closest_pt, dist_factor = intersect_point_line(mouse_co_2d, from_2d, to_2d) #dist-factor - from first 2d line pt.
            if (closest_pt - mouse_2d).length < 6 and  0 < dist_factor < 1:
                return pin_id, None, src_co.lerp(target_co, dist_factor)
        return -1, None, None

    def pick_obj_or_curve(self, depth=None):  #
        ''' returns (hit, hit_co_world, hit_obj)
        If nothing hit, get output location based on region_2d_to_location_3d
        '''
        (hit, hit_co_world, hit_obj) = (None, None, None)
        hit_c, curv_loc_w, curve_obj = self.find_curve_under_cursor(self.event)
        hit_o, o_loc_world, obj = self.find_mesh_obj_under_cursor(self.context, self.event, self.depsgraph)

        def set_return(obj, loc):
            nonlocal hit, hit_obj, hit_co_world
            hit = True
            hit_obj = obj
            hit_co_world = loc

        region = self.context.region
        rv3d = self.context.region_data
        coord = self.event.mouse_region_x, self.event.mouse_region_y
        if hit_c and hit_o:  # select the one closer to view.
            ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
            if (ray_origin-curv_loc_w).length < (ray_origin-o_loc_world).length:
                set_return(curve_obj, curv_loc_w)
            else:
                set_return(obj, o_loc_world)
        elif hit_c:
            set_return(curve_obj, curv_loc_w)
        elif hit_o:
            set_return(obj, o_loc_world)
        else: #nothing was hit.
            pin = self.garment.pins[self.hover_pin_idx]
            pattern_sampling_obj = None #for Normal * line intersect
            if self.which_end == 'SOURCE' and pin.target_obj and pin.target_obj.type == 'CURVE':
                pattern_sampling_obj = pin.target_obj
            elif self.which_end == 'TARGET' and pin.source_obj and pin.source_obj.type == 'CURVE':
                pattern_sampling_obj = pin.source_obj

            if pattern_sampling_obj:  # * project current view origing, to pin.source_obj normal
                patern_plane_no = pattern_sampling_obj.matrix_world @ Vector((0,0,1,0))
                view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
                ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
                line_b = ray_origin + view_vector
                #intersect_line_plane(line_a, line_b, plane_co, plane_no, no_flip=False)
                hit_co_world = intersect_line_plane(ray_origin, line_b, pattern_sampling_obj.matrix_world.to_translation(), patern_plane_no)

            elif depth: # from last hit co
                hit_co_world = view3d_utils.region_2d_to_location_3d(region, rv3d, coord, depth_location=depth)

        return (hit, hit_co_world, hit_obj)


    @CookieCutter_FSM.FSM_State('main')
    def modal_main(self):
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            self.prev_click_time = time.time()
            if self.hover_pin_idx != -1 and self.which_end: #go edit pin_end
                #* setup props for 'edit_pin_end'
                _prop = PropsCache()  # hack to keep variables states saved in new mode
                _prop.pin_obj = None
                _prop.pin_pos = self.ray_cast_3dco
                _prop.drag_state = True #go to hover state automatically, in 'edit_pin_end'
                self._prop = _prop # so that 'edit_pin_end' wont re-init it when None
                logging.debug(f'LMB PRESS from Main({self.which_end})')
                return 'edit_pin_end'

            else: #no hover pin ends picked. So just create new pin
                hit_c, loc_w, hit_obj = self.pick_obj_or_curve()
                # hit_c, loc_w, curve_obj = self.find_curve_under_cursor(self.event)
                if hit_obj:
                    new_pin = self.garment.pins.add()
                    self.hover_pin_idx = len(self.garment.pins) - 1  # used in: self.set_pin_end
                    self.which_end = 'SOURCE'   # used in: self.set_pin_end
                    self.set_pin_end(hit_obj, loc_w)

                    #* setup props for 'edit_pin_end'
                    self.which_end = 'TARGET'
                    _prop = PropsCache()  # hack to keep variables states saved in new mode
                    _prop.pin_obj = None
                    _prop.pin_pos = loc_w #for depth picking mostly...
                    _prop.drag_state = True
                    self._prop = _prop
                    logging.debug(f'LMB PRESS from Main({self.which_end})')
                    return 'edit_pin_end'

                else:
                    self.report({'INFO'}, 'Nothing hit. Try again!')
                    return  # stay in this state
            return

        elif self.event.type == 'A' and self.event.value == 'PRESS':
            self.garment.pins.add()
            self.hover_pin_idx = len(self.garment.pins) - 1
            return #stay in same state

        elif (self.event.type == 'R' or self.event.type == 'DEL') and self.event.value == 'PRESS':
            if self.hover_pin_idx != -1:
                pin = self.garment.pins[self.hover_pin_idx]
                tip_empties = [obj for obj in [pin.source_obj, pin.target_obj] if obj and obj.type=='EMPTY']
                p_opj = self.garment.pins[self.hover_pin_idx].target_obj
                for empty in tip_empties:
                    bpy.data.objects.remove(empty)

                self.garment.pins.remove(self.hover_pin_idx)
                self.hover_pin_idx = -1
                self.report({'INFO'}, f'Pin Removed')
            else:
                self.report({'INFO'}, f'Highlight Pin for removal!')

        #change pin ++ id when not in edith pin mode
        elif self.event.type == 'MOUSEMOVE':
            self.hover_pin_idx, self.which_end, self.ray_cast_3dco = self.find_closest_pin()  # sets self.hover_pin_idx

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.end_cancel()
            return
        elif (self.event.type == 'RET' or self.event.type == 'SPACE') and self.event.value == 'PRESS':
            #sort pattern end first:
            self.end_commit()
            return

    # finish pin_from/pin_to and close
    @CookieCutter_FSM.FSM_State('edit_pin_end')
    def edit_pin_end(self):
        max_click_time = 0.2 # boundary between drag click, and LMB Press click
        if self._prop is None: #init _prop on first pass
            _prop = PropsCache()  # hack to keep variables states saved in new mode
            _prop.pin_obj = None
            _prop.pin_pos = None
            _prop.drag_state = False
            self._prop = _prop
        _prop = self._prop
        pin = self.garment.pins[self.hover_pin_idx]

        def pick_hover_move():
            if ((self.which_end == 'SOURCE' and pin.target_obj and pin.target_obj.type != 'CURVE')
                                         or (self.which_end == 'TARGET' and pin.source_obj and pin.source_obj.type != 'CURVE')):  # we cant have two loose ends - pinned to empties
                hit, _prop.pin_pos, _prop.pin_obj = self.find_curve_under_cursor(self.event, depth=_prop.pin_pos)
            else :#we fine. Pass hits to props
                hit, hit_co, hit_obj = self.pick_obj_or_curve(depth=_prop.pin_obj)
                hit, _prop.pin_pos, _prop.pin_obj = hit, hit_co, hit_obj
            logging.debug(_prop.pin_pos)

        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            logging.debug('LMB PRESS from edit_pin_end()')
            _prop.drag_state = True #if self.which_end == 'BOTH' else False

        elif self.event.type == 'MOUSEMOVE' and _prop.drag_state == True:
            pick_hover_move()
        #* finish creation of pin
        elif (self.event.type == 'LEFTMOUSE' and self.event.value == 'RELEASE') or (_prop.drag_state == True and self.event.value == 'RELEASE'): #fixes: fast click wont go to Finish
            if time.time()-self.prev_click_time < max_click_time:  # treat as single click == dont define second target pin yet
                _prop.drag_state = False #Move out of drag state, so that second LBM click (or drag) can define second point
                logging.debug('LMB Release. Fast')
                return

            pick_hover_move() #last pick on mouse release
            logging.debug('LMB Release drag')
            if _prop.pin_obj:
                self.set_pin_end(_prop.pin_obj, _prop.pin_pos)
            elif _prop.pin_pos: # we clicked not on curve, so generate hover end
                self.set_pin_end(None, _prop.pin_pos) #empty flowating in 3d

            else: #nothing hit - quit
                self.report({'INFO'}, 'Click on pattern!')
                _prop.drag_state == None
                return  # stay in this state

            self._prop = None #reset props cache
            if self.which_end == 'TARGET' or self.which_end == 'SOURCE':  # Finish
                self.hover_pin_idx = -1
                self.deselect_all()
                self.report({'INFO'}, f'Pin End Done')
                self.which_end = None
                return 'main'
            if self.which_end == 'BOTH':  # Finish
                self.which_end = 'TARGET'
                self.report({'INFO'}, f'Pin Start Done')
                # self.prev_click_time = time.time()
                return #go stay in current state

        elif self.event.type == 'MIDDLEMOUSE':
            self.allow_nav = True
        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.cancel_current_pin()
            self._prop = None  # reset props cache
            self.which_end = None
            return 'main'



    def cancel_current_pin(self):  # disable enable sewing, cycle through them
        self.garment.pins[self.hover_pin_idx].target_obj = None
        self.garment.pins[self.hover_pin_idx].source_obj = None
        self.deselect_all()


    def before_start(self, context):
        self._prop = None # will hold local props of modal method
        visible_pins_ids = self.get_visible_pins()
        self.hover_pin_idx = -1
        self.ray_cast_3dco = None # view raycast hit co, helper
        self.which_end = None # TARGET, SOURCE, BOTH
        self.prev_click_time = None
        self.depsgraph = context.evaluated_depsgraph_get()


    def modal(self, context, event):
        self.context = context
        self.event = event
        context.area.tag_redraw()

        if self.done:
            return {self.done}

        self.fsm_update()
        if self.allow_nav:
            self.allow_nav = False
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        self.garment_index = context.scene.garment_index
        if context.space_data.type == 'VIEW_3D':
            self.invoke_garment(context, event)
            self.fsm_init()
            self.before_start(context)
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}


