'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import gpu
import blf
import bpy
import bmesh
from mathutils.bvhtree import BVHTree
from bpy_extras import view3d_utils
from gpu_extras.batch import batch_for_shader

from .utils.fsm_oper import ModalCommon, CookieCutter_FSM
from .utils.helper_functions import draw_text_line, get_addon_preferences


class GTOOL_OT_EditSewings(bpy.types.Operator, ModalCommon, CookieCutter_FSM):
    """Modal object selection with a ray cast"""
    bl_idname = "cloth.raycast_edit_sewing"
    bl_label = "Edit Sewings"
    bl_description = "Connect and stitch two pattern segments\nShift + Click enables viewport sewing Overlay with segments lengths"
    bl_options = {'REGISTER', 'UNDO'}

    sewing_index: bpy.props.IntProperty(default=0)
    closest_segment_id = -1  # if -1 -> we do not have the sewing segment id yet

    @staticmethod
    def draw_text(self, context):
        prefs = get_addon_preferences()
        sewing_count = len(self.garment.garment_sewings)
        draw_text_lines = []
        if self._state == 'main':
            if self.sewing_index == -1:
                draw_text_lines.append(["[LMB] - Add sewing", 'p'])
            else:
                draw_text_lines.append(["[LMB] - Add sewing", 'p'])
                draw_text_lines.append(["[TAB] - Edit sewing", 'p'])
            if len(self.sorted_hover_sewings) > 1:
                draw_text_lines.append(["[MMB Scroll] - Cycle sewings", 'p'])
            draw_text_lines.append(["[DEL/R] - Remove sewing", 'p]'])
            draw_text_lines.append(["[ENTER] - Confirm ", 'p]'])

            if self.sewing_index >= 0:
                font_id = 0
                blf.color(font_id, 1.0, 1.0, 1.0, 1.0)
                if bpy.app.version < (4, 0, 0):
                    blf.size(0, 18, 60)
                else: # for blender 4 4 and up
                    blf.size(0, 18)

                units = 'cm' if prefs.use_absolute_dimensions else ''
                sewing = self.garment.garment_sewings[self.sewing_index] if sewing_count != 0 else None
                if sewing and sewing.source_obj and sewing.target_obj:
                    #draw segment proportions forward to it
                    region = context.region
                    rv3d = context.region_data

                    position1 = view3d_utils.location_3d_to_region_2d(region, rv3d, self.patter_segments_cache[sewing.source_obj.name][sewing.from_spline_idx][sewing.from_segment_idx][3])
                    position2 = view3d_utils.location_3d_to_region_2d(region, rv3d, self.patter_segments_cache[sewing.target_obj.name][sewing.to_spline_idx][sewing.to_segment_idx][3])
                    seg_len1 = self.patter_segments_cache[sewing.source_obj.name+'_len'][sewing.from_spline_idx][sewing.from_segment_idx]
                    seg_len2 = self.patter_segments_cache[sewing.target_obj.name+'_len'][sewing.to_spline_idx][sewing.to_segment_idx]

                    len1 = seg_len1*100 if prefs.use_absolute_dimensions else seg_len1/seg_len2
                    blf.position(0, position1[0], position1[1], 0)
                    blf.draw(0, str(round(len1, 2))+units)

                    len2 = seg_len2*100 if prefs.use_absolute_dimensions else seg_len2/seg_len1
                    blf.position(0, position2[0], position2[1], 0)
                    blf.draw(0, str(round(len2, 2))+units)

        elif self._state == 'pick_source_pattern':
            draw_text_lines.append(["[LMB] - Pick source pattern", 'p'])
            if not self.is_new_sewing:
                draw_text_lines.append(["[MMB Scroll] - flip sewing", 'p'])
                draw_text_lines.append(["[Enter] - Confirm", 'p'])

        elif self._state == 'pick_target_pattern':
            draw_text_lines.append(["[LMB] - Pick target pattern", 'p'])
        elif self._state in ['pick_target_segment', 'pick_source_segment']:
            draw_text_lines.append(["[LMB] - Pick segment...", 'p'])
        if self._state == 'pick_target_segment':
            draw_text_lines.append(["[MMB Scroll] - flip sewing", 'p'])

        # if self._state == 'main':
        #     draw_text_lines.append(["Highlighted sewing: " + str(self.sewing_index+1) + "/" + str(sewing_count), 'p'])
        # else:
        #     draw_text_lines.append(["Editing sewing: " + str(self.sewing_index+1) + "/" + str(sewing_count), 'p'])

        draw_text_line(draw_text_lines)


    @staticmethod
    def draw_px(self, context):
        # * bgl.glDepthFunc(bgl.GL_LEQUAL) #fix 2.91 second layer dept bug
        gpu.state.blend_set("ALPHA")
        gpu.state.point_size_set(8)
        gpu.state.line_width_set(2.0)
        #* bgl.glDepthMask(bgl.GL_TRUE) - enable writing to dept -helps notihig
        gpu.state.depth_test_set("LESS")

        shader = self.shader
        if self.sewing_index >=0: #if we are negative, we are not editing sewing.
            if self._state in ['pick_source_segment', 'pick_target_segment'] and self.sewing_index >= 0:  # hover highlight
                current_sewing = self.garment.garment_sewings[self.sewing_index]
                pattern_obj_name = current_sewing.source_obj.name if self._state == 'pick_source_segment' else current_sewing.target_obj.name
                splines_segments = self.patter_segments_cache[pattern_obj_name]
                if self.closest_spline_id >= 0:
                    segment = splines_segments[self.closest_spline_id][self.closest_segment_id]
                    points = [(point[0], point[1], point[2]) for point in segment]
                    batch = batch_for_shader(shader, "POINTS", {"pos": points})
                    shader.bind()
                    shader.uniform_float("color", (0.2, 0.9, 0.2, 0.5))
                    batch.draw(shader)

            if self._state in ['pick_target_pattern'] and self.sewing_index >= 0:  # higligh source segment, when defining target
                current_sewing = self.garment.garment_sewings[self.sewing_index]
                splines_segments = self.patter_segments_cache[current_sewing.source_obj.name]
                segment = splines_segments[current_sewing.from_spline_idx][current_sewing.from_segment_idx]
                points = [(point[0], point[1], point[2]) for point in segment]
                batch = batch_for_shader(shader, "POINTS", {"pos": points})
                shader.bind()
                shader.uniform_float("color", (0.2, 0.9, 0.2, 0.5))
                batch.draw(shader)

        points = []
        # drawing  sewing
        for i, sewing in enumerate(self.garment.garment_sewings):
            # self.closest_segment_id<0: #skip sewing if we are defining sewing_to
            if self.sewing_index == i and self._state in ['pick_source_segment', 'pick_target_pattern']:
                continue

            if sewing.source_obj is not None and sewing.target_obj is not None:
                if sewing.source_obj.name not in self.patter_segments_cache or sewing.target_obj.name not in self.patter_segments_cache:
                    continue
                vert_stich_01_list = self.patter_segments_cache[sewing.source_obj.name]
                vert_stich_02_list = self.patter_segments_cache[sewing.target_obj.name]

                # if sewing.from_segment_idx >= len(vert_stich_01_list):
                #     sewing.from_segment_idx = 0
                # if sewing.to_segment_idx >= len(vert_stich_02_list):
                #     sewing.to_segment_idx = 0

                verts_from = vert_stich_01_list[sewing.from_spline_idx][sewing.from_segment_idx]
                if self.sewing_index == i and self.closest_segment_id >= 0:
                    verts_to = vert_stich_02_list[self.closest_spline_id][self.closest_segment_id]
                else:
                    verts_to = vert_stich_02_list[sewing.to_spline_idx][sewing.to_segment_idx]
            else:
                continue

            if sewing.flip:
                verts_to = list(reversed(verts_to))

            points = []
            for (point_from, point_to) in zip(verts_from, verts_to):
                points.extend([(point_from.x, point_from.y, point_from.z), (point_to.x, point_to.y, point_to.z)])
            batch = batch_for_shader(shader, 'LINES', {"pos": points})
            shader.bind()
            if self.sewing_index == i:  # draw currently defined sewing
                col = (0.2, 0.9, 0.2, 0.5) if self._state != 'main' else (1.0, 0.8, 0.0, 0.7)
                shader.uniform_float("color", col)
            else:
                shader.uniform_float("color", (0.95, 0.65, 0.15, 0.5))
            batch.draw(shader)
        # batch = shader.new_batch('LINES', {"pos": points})

        #restore opengl defaults
        gpu.state.depth_test_set("NONE")
        gpu.state.line_width_set(1)
        gpu.state.point_size_set(1)
        gpu.state.blend_set("NONE")

    def get_visible_sewing(self):
        return [idx for idx, sewing in enumerate( self.garment.garment_sewings) if sewing.source_obj in self.context.visible_objects and sewing.target_obj in self.context.visible_objects]


    def sewing_idx_update(self):
        if self.sorted_hover_sewings:
            self.sewing_index = self.sorted_hover_sewings[self.sewing_id_offset % len(self.sorted_hover_sewings)]
        else:
            self.sewing_index = -1
            self.sewing_id_offset = 0

    def find_closest_sewing(self):
        ''' Return index of sewing closest to mouse'''
        region = self.context.region
        rv3d = self.context.region_data
        scene = self.context.scene
        region = self.context.region
        rv3d = self.context.region_data
        coord = self.event.mouse_region_x, self.event.mouse_region_y

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

        # cast rays and find the closest object
        visible_sewing_ids = self.get_visible_sewing()
        distances = []
        hovered_sevings = []
        for sewing_id in visible_sewing_ids:
            loc, normal, face_id, dist = self.sewings_bvht[sewing_id].ray_cast(ray_origin, view_vector)
            if loc:
                hovered_sevings.append(sewing_id)  # self.garment.garment_sewings
                distances.append(dist)
        if len(hovered_sevings) != len(self.sorted_hover_sewings): #if seg count under cursor is different...
            self.sewing_id_offset = 0
        return [s_id for _, s_id in sorted(zip(distances, hovered_sevings))]  # sorted_hover_sewings by dist to cam. What about artho


    @CookieCutter_FSM.FSM_State('main')
    def modal_main(self):
        sewing_count = len(self.garment.garment_sewings)
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            hit_c, loc_w, curve_obj = self.find_curve_under_cursor(self.event, curves_2d=True, curves_3d=True)
            if hit_c:
                curve_obj.select_set(True)
                # if self.sewing_index == -1:
                self.garment.garment_sewings.add()
                self.sewing_index = len(self.garment.garment_sewings)-1
                self.sewings_bvht.append(None)  # for now None
                self.garment.garment_sewings[self.sewing_index].source_obj = curve_obj
                self.assign_target_obj_to_garment(curve_obj)
                self.is_new_sewing = True
                self.sewing_id_offset = 0
                self.closest_segment_id = -1
                self.closest_spline_id = -1
                return 'pick_source_segment'
            else:
                self.report({'INFO'}, 'No close pattern detected. Click Again')
                return  # stay in same state

        elif self.event.type == 'TAB' and self.event.value == 'PRESS' and self.sewing_index != -1:
            self.is_new_sewing = False
            self.backup_current_sewing()
            return 'pick_source_pattern'

        elif (self.event.type == 'R' or self.event.type == 'DEL') and self.event.value == 'PRESS':
            if self.sewing_index != -1:  #if we have any sewing selected
                self.garment.garment_sewings.remove(self.sewing_index)
                self.sewing_id_offset = 0
                self.sewing_index = -1
                self.update_sewings_bvht()

        elif self.event.type == 'MOUSEMOVE':
            # current_sewing = self.garment.garment_sewings[self.sewing_index]
            self.sorted_hover_sewings = self.find_closest_sewing()  # sets self.sewing_index
            self.sewing_idx_update()

        #change sewing ++ id when not in edith sewing mode
        elif self.event.type == 'WHEELUPMOUSE':
            if self.sewing_index == -1 or len(self.sorted_hover_sewings)<2:
                self.allow_nav = True
                return
            self.sewing_id_offset += 1
            self.sewing_idx_update()

        #change sewing -- id when not in edith sewing mode
        elif self.event.type == 'WHEELDOWNMOUSE':
            if self.sewing_index == -1 or len(self.sorted_hover_sewings) < 2:
                self.allow_nav = True
                return
            self.sewing_id_offset -= 1
            self.sewing_idx_update()


        elif self.event.type == 'MIDDLEMOUSE':
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.end_cancel()
            return

        elif (self.event.type == 'RET' or self.event.type == 'SPACE') and self.event.value == 'PRESS':
            self.end_commit()
            return


    # finish pin_from/pin_to and close
    @CookieCutter_FSM.FSM_State('pick_source_pattern')
    def pick_source_pattern(self):
        # finish sewing_from/sewing_to and close
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            hit_c, loc_w, curve_obj = self.find_curve_under_cursor(self.event, curves_2d=True, curves_3d=True)
            if hit_c:
                curve_obj.select_set(True)
                if self.sewing_index == -1:
                    self.garment.garment_sewings.add()
                    self.sewing_index = len(self.garment.garment_sewings)-1
                    self.sewings_bvht.append(None) #for now None
                self.garment.garment_sewings[self.sewing_index].source_obj = curve_obj
                self.assign_target_obj_to_garment(curve_obj)
                self.sewing_id_offset = 0
                self.closest_segment_id = -1
                self.closest_spline_id = -1
                return 'pick_source_segment'
            else:
                self.report({'INFO'}, 'No close pattern detected. Click Again')
                return  # stay in same state

        elif self.event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE', 'F'} and self.event.value == 'PRESS' and not self.is_new_sewing:
            current_sewing = self.garment.garment_sewings[self.sewing_index]
            current_sewing.flip = not current_sewing.flip

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif (self.event.type == 'RET' or self.event.type == 'SPACE') and self.event.value == 'PRESS':
            return 'main'

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.cancel_restore_sewing()
            return 'main'


    # finish pin_from/pin_to and close
    @CookieCutter_FSM.FSM_State('pick_source_segment')
    def pick_source_segment(self):
        # finish sewing_from/sewing_to and close
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            # draw the points and pick
            if self.closest_segment_id >= 0:
                current_sewing = self.garment.garment_sewings[self.sewing_index]
                current_sewing.from_segment_idx = self.closest_segment_id
                current_sewing.from_spline_idx = self.closest_spline_id
                self.closest_segment_id = -1
                self.closest_spline_id = -1
                return 'pick_target_pattern'
            else:
                return #stay in the same state

        elif self.event.type == 'MOUSEMOVE':
            current_sewing = self.garment.garment_sewings[self.sewing_index]
            self.closest_spline_id, self.closest_segment_id = self.find_closest_segment(current_sewing.source_obj.name)
            # return {'PASS_THROUGH'}

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.cancel_restore_sewing()
            return 'main'


    # finish pin_from/pin_to and close
    @CookieCutter_FSM.FSM_State('pick_target_pattern')
    def pick_target_pattern(self):
        #finish picking curve pattern for sewing source/target
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            hit_c, loc_w, curve_obj = self.find_curve_under_cursor(self.event, curves_2d=True, curves_3d=True)
            if hit_c:
                curve_obj.select_set(True)
                current_sewing = self.garment.garment_sewings[self.sewing_index]
                current_sewing.target_obj = curve_obj
                self.assign_target_obj_to_garment(curve_obj)
                return 'pick_target_segment'
            else:
                self.report({'INFO'}, 'No close pattern detected. Click Again')
                return #stay in same state

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.cancel_restore_sewing()
            return 'main'


    @CookieCutter_FSM.FSM_State('pick_target_segment')
    def pick_target_segment(self):
        current_sewing = self.garment.garment_sewings[self.sewing_index]
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            # draw the points and pick
            if self.closest_segment_id >= 0:
                current_sewing.to_segment_idx = self.closest_segment_id
                current_sewing.to_spline_idx = self.closest_spline_id
                self.closest_segment_id = -1
                self.closest_spline_id = -1
                self.update_sewings_bvht()
                self.deselect_all()
                return 'main'
            else:
                self.report({'INFO'}, 'No close sewing. Click Again')
                return  # stay in the same state

        elif self.event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE', 'F'} and self.event.value == 'PRESS':
            current_sewing.flip = not current_sewing.flip

        elif self.event.type == 'MOUSEMOVE':
            self.closest_spline_id, self.closest_segment_id = self.find_closest_segment(current_sewing.target_obj.name)
            # return {'PASS_THROUGH'}

        elif self.event.type == 'MIDDLEMOUSE':
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.cancel_restore_sewing()
            return 'main'


    #cancall currently edited sewing
    def backup_current_sewing(self):
        current_sewing = self.garment.garment_sewings[self.sewing_index]
        self.backup_sewing = {}
        self.backup_sewing['target_obj'] = current_sewing.target_obj
        self.backup_sewing['source_obj'] = current_sewing.source_obj
        self.backup_sewing['from_segment_idx'] = current_sewing.from_segment_idx
        self.backup_sewing['from_spline_idx'] = current_sewing.from_spline_idx
        self.backup_sewing['to_segment_idx'] = current_sewing.to_segment_idx
        self.backup_sewing['to_spline_idx'] = current_sewing.to_spline_idx
        self.backup_sewing['flip'] = current_sewing.flip


    def cancel_restore_sewing(self):
        if self.is_new_sewing:
            self.garment.garment_sewings.remove(self.sewing_index)
            self.sewing_id_offset = 0
            self.sewing_index = -1
            # self.update_sewings_bvht() no need since when creating sewing, self.sewings_bvht was not updated
        else:
            current_sewing = self.garment.garment_sewings[self.sewing_index]
            current_sewing.target_obj = self.backup_sewing['target_obj']
            current_sewing.source_obj = self.backup_sewing['source_obj']
            current_sewing.from_segment_idx = self.backup_sewing['from_segment_idx']
            current_sewing.from_spline_idx = self.backup_sewing['from_spline_idx']
            current_sewing.to_segment_idx = self.backup_sewing['to_segment_idx']
            current_sewing.to_spline_idx = self.backup_sewing['to_spline_idx']
            current_sewing.flip = self.backup_sewing['flip']
            self.backup_sewing.clear()

        self.closest_segment_id = -1
        self.closest_spline_id = -1
        self.deselect_all()



    def before_start(self):
        sewing_count = len(self.garment.garment_sewings)
        visible_sewing_ids = self.get_visible_sewing()
        visible_sewing_count = len(visible_sewing_ids)
        #init sewing_index from visible_sewing, or pattern and visible_sewing
        self.sewing_id_offset = 0
        self.sorted_hover_sewings = []
        self.sewing_index = -1
        self.sewings_bvht = []

        if sewing_count == 0 or visible_sewing_count == 0:
            return

        self.update_sewings_bvht()
        if visible_sewing_count == 1:
            self.sewing_index = visible_sewing_ids[0]
            return
        self.sewing_index = visible_sewing_ids[0]

        self.back_draw_sewings = self.garment.sewings_overlay
        self.garment.sewings_overlay = False

    def modal(self, context, event):
        self.context = context
        self.event = event
        context.area.tag_redraw()

        if self.done:
            return {self.done}

        self.fsm_update()
        if self.allow_nav:
            self.allow_nav = False
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def update_sewings_bvht(self):
        # resampled_splines_segments = self.patter_segments_cache[obj.name]
        out_bvht = []
        for sewing in self.garment.garment_sewings:
            if sewing.source_obj and sewing.target_obj:
                if sewing.source_obj.name not in self.patter_segments_cache or sewing.target_obj.name not in self.patter_segments_cache:
                    out_bvht.append(None)
                    continue
                vert_stich_01_list = self.patter_segments_cache[sewing.source_obj.name]
                vert_stich_02_list = self.patter_segments_cache[sewing.target_obj.name]

                verts_from = vert_stich_01_list[sewing.from_spline_idx][sewing.from_segment_idx]
                verts_to = vert_stich_02_list[sewing.to_spline_idx][sewing.to_segment_idx]
                if sewing.flip:
                    verts_to = list(reversed(verts_to))
            else:
                out_bvht.append(None)
                continue

            bm = bmesh.new()   # create an empty BMesh
            prev_from = None
            prev_to = None

            for (point_from, point_to) in zip(verts_from, verts_to):
                new_from = bm.verts.new(point_from)
                new_to = bm.verts.new(point_to)
                if prev_from:
                   new_face = bm.faces.new([prev_from, prev_to, new_to, new_from])
                prev_from = new_from
                prev_to = new_to

            #* adding thickness to sewings bvh
            circumference = (verts_from[0]-verts_from[-1]).length + (verts_to[0]-verts_to[-1]).length + (verts_from[0]-verts_to[0]).length + (verts_from[-1]-verts_to[-1]).length
            thickness = circumference / 4 / 10
            bm.normal_update()
            for v in bm.verts:
                v.co = v.co + v.normal*thickness/2
            bmesh.ops.solidify(bm, geom=bm.faces, thickness=thickness)
            # bmesh.ops.triangle_fill(bm, use_beauty=True, use_dissolve=False, edges=bm.edges)
            out_bvht.append(BVHTree.FromBMesh(bm))
            bm.free()
        self.sewings_bvht = out_bvht


    def invoke(self, context, event):
        self.garment_index = context.scene.garment_index
        if context.space_data.type == 'VIEW_3D':
            if event.shift:
                garment = context.scene.cloth_garment_data[self.garment_index]
                garment.sewings_overlay = not garment.sewings_overlay
                on_of = 'Enabled' if garment.sewings_overlay else 'Disabled'
                self.report({'INFO'}, f'{on_of} sewing overlay')
                return {'FINISHED'}
            self.invoke_garment(context, event)
            self.fsm_init()
            self.before_start()
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}




