'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import bmesh
import numpy as np
from mathutils import Vector
from .utils.helper_functions import update_keyframe_prop
from .edit_cloth_preset import get_default_preset, calculate_cloth_preset_vg_weights


def three_d_bincount_add(out ,e_k, weights):
    out[:,0] += np.bincount(e_k, weights[:,0], out.shape[0])
    out[:,1] += np.bincount(e_k, weights[:,1], out.shape[0])
    out[:,2] += np.bincount(e_k, weights[:,2], out.shape[0])

def new_bincount(out, e_k, weights = None):
    out += np.bincount(e_k, weights, out.shape[0])


def close_bmesh(context,  bm, source):
    bm.normal_update()
    if context.object.mode == "EDIT":
        bmesh.update_edit_mesh(source, tessface=False, destructive=False)
    else:
        bm.to_mesh(source)
        bm.free()
        # source.update()

def get_bmesh(context, mesh):
    bm = bmesh.new()
    if context.active_object.mode == 'OBJECT':
        bm.from_mesh(mesh)
    elif context.active_object.mode == 'EDIT':
        bm =  bmesh.from_edit_mesh(mesh)
    return bm

class GTOOL_OT_GarmentSmooth(bpy.types.Operator):
    bl_idname = 'object.cloth_smooth'
    bl_label = 'Cloth Smooth'
    bl_description = 'Smooth'
    bl_options = {'REGISTER', 'UNDO'}

    iteration_amount: bpy.props.IntProperty(name="Iterations", description="Number of times to repeat the smoothing", default=3, min=1, max=10, step=1)

    def inflate_numpy_smooth(self,context):
        bm = get_bmesh(context, context.active_object.data)
        bm.verts.ensure_lookup_table()
        bm.edges.ensure_lookup_table()

        o_co =  np.array([vert.co for vert in bm.verts])
        vert_count =  len(bm.verts)

        edge_keys = np.array([ [edge.verts[0].index, edge.verts[1].index] for edge in  bm.edges])
        freeze_verts = np.array([vert.is_boundary or not vert.is_manifold for vert in bm.verts]) #manifold for pocket sewing edges

        p_co = np.copy(o_co) #current new vertex position

        #sum edges that have vert X  and either end of edge (Va, Vb)
        v_connections = np.zeros(vert_count, dtype=np.float32)
        new_bincount(v_connections, edge_keys[:,1] )
        new_bincount(v_connections, edge_keys[:,0])

        for _ in range(self.iteration_amount):
            q_co = np.copy(p_co) #previous iteration vert position
            p_co.fill(0)
            #calc average position weighted...
            three_d_bincount_add(p_co, edge_keys[:,1], q_co[edge_keys[:,0]])
            three_d_bincount_add(p_co, edge_keys[:,0], q_co[edge_keys[:,1]])

            p_co =p_co / v_connections[:,None]   #new aver posision

            p_co[freeze_verts] = q_co[freeze_verts] # not selected ver - reset to original pos o_co
        # source.vertices.foreach_set('co', p_co.ravel())
        for v in bm.verts:
            v.co = p_co[v.index]
        close_bmesh(context, bm, context.active_object.data)


    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'MESH')

    def execute(self, context):
        self.inflate_numpy_smooth(context)
        return {'FINISHED'}



def init_simulation(context, garment):
    if garment.target_cloth_object is None:
        return
    sim_time = garment.sim_time #use for gravity ? or as bake cache time?
    sewing_fac = garment.warmup_factor
    garment_obj = garment.target_cloth_object
    garment_obj.animation_data_clear() #maybe jutt delete cloth mod actions keyfremes

    cloth_mod = garment_obj.modifiers.get('Cloth')
    if not cloth_mod:
        cloth_mod = garment_obj.modifiers.new(name='Cloth', type='CLOTH')
        cloth_mod.settings.use_sewing_springs = True
        cloth_mod.collision_settings.distance_min = 0.005

    _, _, max_cloth_params = calculate_cloth_preset_vg_weights(garment)

    if cloth_mod.settings.vertex_group_shrink != '' and max_cloth_params:
        update_keyframe_prop(garment_obj, 'modifiers["Cloth"].settings.shrink_max', frame=int(sim_time*sewing_fac*1.6), max_val=max_cloth_params['shrink'])
        # NOTE: old way
        # garment_obj.keyframe_insert(data_path='modifiers["Cloth"].settings.shrink_max', frame=int(sim_time*sewing_fac*1.6)) # 40% delay
        # cloth_mod.settings.shrink_max = 0.0
        # garment_obj.keyframe_insert(data_path='modifiers["Cloth"].settings.shrink_max', frame=0)

    if cloth_mod.settings.quality <= 10:
        cloth_mod.settings.quality = 18
    # context.scene.animation_data_clear()


    # sewing_force
    default_preset = get_default_preset(garment)
    if default_preset and max_cloth_params:
        update_keyframe_prop(garment_obj, 'modifiers["Cloth"].settings.sewing_force_max',
                             frame=int(sim_time*sewing_fac),
                             max_val=default_preset.sewing_force, min_val=0.4*default_preset.sewing_force)
        # NOTE: old way...
        # cloth_mod.settings.sewing_force_max = 0.2
        # garment_obj.keyframe_insert(data_path='modifiers["Cloth"].settings.sewing_force_max', frame=0)
        # cloth_mod.settings.sewing_force_max = 2.0
        # garment_obj.keyframe_insert(data_path='modifiers["Cloth"].settings.sewing_force_max', frame=int(sim_time*sewing_fac))
    else: # no default preset - just animate sewing force to 2
        update_keyframe_prop(garment_obj, 'modifiers["Cloth"].settings.sewing_force_max',
                             frame=int(sim_time*sewing_fac),
                             max_val=2, min_val=0.4)

    # pressure_factor
    if cloth_mod.settings.use_pressure:
        cloth_mod.settings.pressure_factor = 1.0
        garment_obj.keyframe_insert(data_path='modifiers["Cloth"].settings.pressure_factor', frame=int(sim_time*sewing_fac*1.2))
        cloth_mod.settings.pressure_factor = 0
        garment_obj.keyframe_insert(data_path='modifiers["Cloth"].settings.pressure_factor', frame=0)

    #* Gravity
    if context.scene.animation_data:
        if context.scene.animation_data.action is None:
            context.scene.animation_data.action = bpy.data.actions.new('SceneAction')
        for fcurv in context.scene.animation_data.action.fcurves:
            if fcurv.data_path == 'gravity':
                context.scene.animation_data.action.fcurves.remove(fcurv)
    context.scene.gravity[2] = -0.8
    context.scene.keyframe_insert(data_path='gravity', frame=0)
    context.scene.gravity[2] = -9.8
    context.scene.keyframe_insert(data_path='gravity', frame=sim_time)

    # garment_obj.modifiers['Cloth'].point_cache.frame_end = int(sim_time*1.4)


class GTOOL_OT_SetupKeyframes(bpy.types.Operator):
    bl_idname = "garment.init_simulation"
    bl_label = "Initialize Simulation"
    bl_description = 'Initialize simulation variables (like gravity, sewing strength) by animating their slow increase over time\n\
This helps to avoid cloth falling too fast on floor'
    bl_options = {'REGISTER', 'UNDO'}


    def execute(self, context):
        garment_index = context.scene.garment_index
        garment = context.scene.cloth_garment_data[garment_index]
        init_simulation(context, garment)
        return {'FINISHED'}


class GTOOL_OT_BakeCloth(bpy.types.Operator):
    bl_idname = "garment.simulate_cloth"
    bl_label = "Simulate"
    bl_description = 'Bake simulation to cache\nCannot be paused or cancelled!'
    bl_options = {'REGISTER', 'UNDO'}


    def execute(self, context):
        garment_index = context.scene.garment_index
        garment = context.scene.cloth_garment_data[garment_index]
        if garment.target_cloth_object is None:
            self.report({'INFO'}, 'Cannot find Garment object: ' + garment.name)
            return {'CANCELLED'}
        garment_obj = garment.target_cloth_object
        for modifier in garment_obj.modifiers:
            if modifier.type == 'CLOTH':
                override = {'scene': context.scene, 'active_object': garment_obj, 'point_cache': modifier.point_cache}
                with bpy.context.temp_override(**override):
                    bpy.ops.ptcache.free_bake()
                    bpy.ops.ptcache.bake(bake=True)
                break
        return {'FINISHED'}


class GTOOL_OT_FinalizeClothMesh(bpy.types.Operator):
    bl_idname = "garment.finalize_cloth_mesh"
    bl_label = "Finalize cloth"
    bl_description = "Applies modifiers, collapse sewing edges, and generate new mesh with uv and vertex groups"
    bl_options = {"REGISTER","UNDO"}

    weld_sewings: bpy.props.BoolProperty(name='Weld Sewings', description='Weld adjacent sewing patterns borders together', default=True)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH'


    def execute(self, context):
        # garment_index = context.scene.garment_index
        # garment_obj = context.scene.cloth_garment_data[garment_index].target_cloth_object if garment_index >= 0 else context.active_object
        garment_obj = context.active_object

        output_obj = garment_obj.copy()
        mesh_copy = output_obj.data.copy()#uninstace
        output_obj.data = mesh_copy

        for col in garment_obj.users_collection:
            col.objects.link(output_obj)

        is_cloth_sim_on = any([mod.show_viewport for mod in garment_obj.modifiers if mod.type == 'CLOTH' and mod.show_viewport])

        constructive_modifiers = ['SUBSURF', 'BEVEL', 'MIRROR', 'MASK', 'EDGE_SPLIT', 'SOLIDIFY']
        disabled_mods = []
        for mod in output_obj.modifiers:  # disable constructive modifiers - cant write to shape if new geo is created
            if mod.type in constructive_modifiers and not mod.show_viewport:
                disabled_mods.append(mod.name)
                mod.show_viewport = False

        if is_cloth_sim_on:
            # we baked cloth from from sim - remove existing shape keys
            if output_obj.data.shape_keys:
                for shape_key in reversed(output_obj.data.shape_keys.key_blocks): # reversed - to remove custom shape, and reset to base - last
                    output_obj.shape_key_remove(shape_key)

            depsgraph = bpy.context.evaluated_depsgraph_get()
            cloth_obj_eval = garment_obj.evaluated_get(depsgraph)

            #* copy verts from eval_obj, to new_mesh
            for new_v, src_v in zip(mesh_copy.vertices, cloth_obj_eval.data.vertices):
                new_v.co = src_v.co
        else:
            # we baked cloth from from sim - remove existing shape keys
            if output_obj.data.shape_keys:
                for shape_key in output_obj.data.shape_keys.key_blocks: # forward dir - apply base - then build up shapes on top of base..
                    output_obj.shape_key_remove(shape_key)


        mesh_copy.update()
        output_obj.data = mesh_copy

        for mod in output_obj.modifiers:  # remove modifiers stored to mesh
            if mod.name not in disabled_mods: # not-disabled mod ==  remove deforming mods (cos they were applied)
                output_obj.modifiers.remove(mod)

        #* merge wire edges.
        bm = bmesh.new()   # create an empty BMesh
        bm.from_mesh(mesh_copy)   # fill it in from a Mesh

        # remove loose edges
        wire_edges = [e for e in bm.edges if e.is_wire]
        pure_wire_edges = [e for e in wire_edges if len(e.verts[0].link_faces) == 0 and len(e.verts[1].link_faces) == 0]
        bmesh.ops.delete(bm, geom=pure_wire_edges, context='EDGES')

        # remove pin edges - detect by e.v1 or e.v2 - are they linked to flat manifold patch
        wire_edges = [e for e in bm.edges if e.is_wire]
        pin_edges = [] # when two places in mesh are sewn
        for e in wire_edges:
            sub_edge_is_bound = False # does any v.e is boundary. If not - e must be pin edge
            for vert_e in e.verts[0].link_edges[:] + e.verts[1].link_edges[:]:
                if vert_e.is_boundary and vert_e != e:
                    sub_edge_is_bound = True
                    break
            if not sub_edge_is_bound:
                pin_edges.append(e)
        bmesh.ops.delete(bm, geom=pin_edges, context='EDGES')


        # loose_verts = [v for v in bm.verts if len(v.link_faces)==0]
        # bmesh.ops.delete(bm, geom=loose_verts, context='VERTS')
        # mark all boundary edges with sharp and crease
        if bpy.app.version < (4, 0, 0):
            crease_layer = bm.edges.layers.crease.verify()
        else:
            crease_layer = bm.edges.layers.float.get('crease_edge', None)
            # crease_edge_attr = obj.data.attributes.get('crease_vert')
            if not crease_layer:
                crease_layer = bm.edges.layers.float.new('crease_edge')
                # crease_edge_attr = obj.data.attributes.new('crease_vert', 'FLOAT','POINT')

        for e in bm.edges:
            if e.is_boundary:
                e.smooth = False
                e[crease_layer] = 1.0

        wire_edges = [e for e in bm.edges if e.is_wire]
        if self.weld_sewings:
            if wire_edges:
                bmesh.ops.collapse(bm, edges=wire_edges, uvs = True)
            # extra step - get non boundary but creased edges, and smooth them with adjacent edges
            edges_to_smooth = [e for e in bm.edges if e[crease_layer] > 0.9 and not e.is_boundary]
            # convert edges to verts, and smooth them
            verts_to_smooth = set()
            for e in edges_to_smooth:
                if not e.verts[0].is_boundary:
                    verts_to_smooth.add(e.verts[0])
                if not e.verts[1].is_boundary:
                    verts_to_smooth.add(e.verts[1])

            adj_verts = {}
            b = 0.4
            for _ in range(2):  # repeat 3 times
                for v in verts_to_smooth:
                    # adj_co = np.mean([e.other_vert(v).co for e in v.link_edges if e[crease_layer] > 0.9 and not e.is_boundary], axis=0)
                    adjv = adj_verts.get(v.index)
                    if not adjv:
                        adjv = [e.other_vert(v) for e in v.link_edges if e[crease_layer] > .9 and not e.is_boundary]
                        adj_verts[v.index] = adjv
                    adj_aver = sum((v.co for v in adjv), Vector())/len(adjv)
                    v.co = b*v.co + (1.0 - b)*adj_aver

        else:
            for e in wire_edges:
                center = (e.verts[0].co + e.verts[1].co)/2
                e.verts[0].co = center
                e.verts[1].co = center
            bmesh.ops.delete(bm, geom=wire_edges, context='EDGES')


        bm.to_mesh(mesh_copy)
        bm.free()  # free and prevent further access

        for mod_name in disabled_mods:  # enable back constructive modifiers
            output_obj.modifiers[mod_name].show_viewport = True

        for obj in context.selected_objects:
            obj.select_set(False)
        output_obj.select_set(True)
        output_obj.data.use_auto_smooth = True
        output_obj.data.auto_smooth_angle = 3.14159
        context.view_layer.objects.active = output_obj

        return {"FINISHED"}
