# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
# Pattern flattening based on https://gitlab.com/thomaskole/blender-seams-to-sewing-pattern/-/blob/master/op_seams_to_sewingpattern.py

import bpy
import bmesh
# from mathutils.bvhtree import BVHTree
from mathutils.kdtree import KDTree
from mathutils import Vector, Matrix
from mathutils.geometry import intersect_line_plane, intersect_point_line
from math import sqrt, radians





def euclidean_len(points):
    prev_co = points[0]
    total_len = 0
    for p in points:
        total_len += (p - prev_co).length
        prev_co = p
    return total_len

def point_line_dist(point1, point2, pointn):
    cut_pt, _ = intersect_point_line(pointn, point1, point2)
    return (cut_pt - pointn).length

def iterate(points, out_pt_ids, error):
    new = []
    for seg_idx in range(len(out_pt_ids) - 1):  # for each sub  segment in the simplified line
        furthest_pts_ids = 0
        max_dist = 0
        # for subsegment  inner pts
        for i, test_point in enumerate(points[out_pt_ids[seg_idx] + 1: out_pt_ids[seg_idx + 1]]):
            # point_line_dist(first_pt, last_pt, test_point)
            current_dist = point_line_dist(points[out_pt_ids[seg_idx]], points[out_pt_ids[seg_idx + 1]], test_point)
            if current_dist > max_dist:
                max_dist = current_dist
                if max_dist >= error:
                    furthest_pts_ids = i + 1 + out_pt_ids[seg_idx]
        if furthest_pts_ids: # add furthest pt, for each sub segment
            new.append(furthest_pts_ids)
    return False if new == [] else new

### Ramer-Douglas-Peucker algorithm ###
def simplify_RDP(points, error):
    ''' Return simplified verts ids '''
    out_pt_ids = [0, len(points) - 1]
    euclid_len = euclidean_len(points) # to  make simplify_RDP independent on spl length (normalize kind of)
    if euclid_len < 0.0001:
        return out_pt_ids

    err = error/100 * euclid_len
    iter_pts_ids = 1
    while iter_pts_ids is not False:
        iter_pts_ids = iterate(points, out_pt_ids, err)
        if iter_pts_ids:
            out_pt_ids += iter_pts_ids
            out_pt_ids.sort()
    return out_pt_ids


class GTOOL_OT_CURVE_OT_CSimplify(bpy.types.Operator):
    bl_idname = "object.garment_pattern_simplify"
    bl_label = "Simplify Garmetn Pattern"
    bl_description = "Simplify Pattern"
    bl_options = {'REGISTER', 'UNDO'}

    error: bpy.props.FloatProperty(name="Error", description="Maximum allowed distance error in Blender Units", min=0, max=10, default=0.0, precision=2)
    sharp_angle_threshold: bpy.props.FloatProperty(name="Sharp Angle Threshold", description="Threshold for sharp angles", min=0, max=180, default=30.0, precision=1)

    @staticmethod
    def simplify_curve(curve, error=0.5, angle_threshold=0.0):
        backup_mat_indices = [spline.material_index for spline in curve.data.splines]
        simplified_splines = []

        # TODO: for now it takes onliy curve with one spline. May need to target only specyfic spl_idx
        for spline in curve.data.splines:
            if len(spline.points) > 1 or len(spline.bezier_points) > 1:  # skip single points
                if spline.type == 'BEZIER':
                    points = [point.co.xyz for point in spline.bezier_points]
                else:
                    points = [point.co.xyz for point in spline.points]
                out_pt_ids = simplify_RDP(points, error)
                simplified_pts = [points[idx] for idx in out_pt_ids]
                simplified_splines.append(simplified_pts)

        curve.data.splines.clear()
        curve.data.resolution_u = 3
        # write new simplified splines points
        for simplified_spline in simplified_splines:
            spline = curve.data.splines.new('BEZIER')
            spline.use_cyclic_u = True
            spline.bezier_points.add(len(simplified_spline) - 1)
            spl_len = len(simplified_spline)
            prev_pt = simplified_spline[-1]

            angle_rad = radians(angle_threshold)
            for i, point in enumerate(simplified_spline):
                # calculate angle between prev and point,  and point and next point
                next_pt = simplified_spline[(i + 1) % spl_len]
                angle = (point - prev_pt).angle((next_pt - point))
                spline.bezier_points[i].co = point
                prev_pt = point
                if abs(angle) < angle_rad:
                    spline.bezier_points[i].handle_left_type = 'AUTO'
                    spline.bezier_points[i].handle_right_type = 'AUTO'
                else:
                    spline.bezier_points[i].handle_left_type = 'VECTOR'
                    spline.bezier_points[i].handle_right_type = 'VECTOR'


    def execute(self, context):
        self.simplify_curve(context.active_object, self.error, self.sharp_angle_threshold)
        return {'FINISHED'}



def get_verts_islands(bm):
    ''' get islands as vert lists - ignore faces linked by wire edges'''
    islands = []
    i = -1
    for v in bm.verts:
        v.tag = False

    for v in bm.verts:
        if not v.tag:
            island = [v]
            v.tag = True
            i += 1
            islands.append([])
            while len(island) > 0:
                v = island.pop()
                islands[i].append(v)
                for e in v.link_edges:
                    if len(e.link_faces)>0: # ignore wire
                        for v2 in e.verts:
                            if not v2.tag:
                                v2.tag = True
                                island.append(v2)
    return islands

# get_faces_islands
def get_faces_islands(bm):
    ''' get islands as face lists - ignore faces linked by wire edges'''
    islands = []
    i = -1
    for f in bm.faces:
        f.tag = False

    for f in bm.faces:
        if not f.tag:
            island = [f]
            f.tag = True
            i += 1
            islands.append([])
            while len(island) > 0:
                f = island.pop()
                islands[i].append(f)
                for e in f.edges:
                    if len(e.link_faces)>1: # at least 2 faces
                        for f2 in e.link_faces:
                            if not f2.tag:
                                f2.tag = True
                                island.append(f2)
    return islands


def get_boundary_edges(verts_in_island):
    '''returns boundary loop edges (sorted by topoogy)'''
    # find first boundary vert
    first_loop_vert = None
    for v in verts_in_island:
        for e in v.link_edges:
            e.tag = False
        if v.is_boundary:
            first_loop_vert = v
            break
    # follow linked edges loop

    # get first boundary edge
    first_loop_edge = None
    for e in first_loop_vert.link_edges:
        if e.is_boundary:
            first_loop_edge = e
            break

    # follow linked edges loop
    loop_edges = [first_loop_edge]
    current_edge = first_loop_edge
    current_edge.tag = True
    prev_vert = first_loop_vert
    while True:
        prev_edge = current_edge
        next_vert = current_edge.other_vert(prev_vert) # go from  (prev_vert) - e -> (next_vert)
        for e in next_vert.link_edges:
            if e.is_boundary and not e.tag:
                loop_edges.append(e)
                current_edge = e
                current_edge.tag = True
                prev_vert = next_vert
                break
        if current_edge == prev_edge: # no more edges
            break

    island_boundary_verts = []
    prev_edge = loop_edges[-1]
    for e in loop_edges:
        shared_vert = e.verts[0] if prev_edge.other_vert(e.verts[0]) else e.verts[1]
        island_boundary_verts.append(shared_vert)

    return loop_edges, island_boundary_verts


def get_seam_edges(bm):
    '''get edges that are seams but non border '''
    seam_edges = []
    # uv_lay = bm.loops.layers.uv['UB']
    uv_lay = bm.loops.layers.uv.active
    for e in bm.edges:
        #no need to split this since already border
        # if len(e.link_faces) == 1:
        #     e.select = True
        if len(e.link_faces) == 2:
            e_loop1 = e.link_loops[0]
            e_loop2 = e.link_loops[1]
            if e_loop1[uv_lay].uv != e_loop2.link_loop_next[uv_lay].uv or e_loop2[uv_lay].uv != e_loop1.link_loop_next[uv_lay].uv:
                seam_edges.append(e)
    return seam_edges

class GT_OT_UVShapekey(bpy.types.Operator):
    bl_label = "Patterns from Mesh UV"
    bl_idname = "cloth.uv_to_pattern"
    bl_description = "Use objects UV to create sewing Patterns"
    bl_options = {'REGISTER', 'UNDO'}

    error: bpy.props.FloatProperty(name="Decimation Amount", description="Simplify generated strands", min=0, max=10, default=0.5, precision=2)
    sharp_angle_threshold: bpy.props.FloatProperty(name="Sharp Angle Threshold", description="Threshold for sharp angles", min=0, max=180, default=30.0, precision=1)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH' and len(context.active_object.data.uv_layers) > 0

    @staticmethod
    def create_flat_shape(bm, island_verts):
        # shape_lay = bm.verts.layers.shape.get("uv_flat")
        # if not shape_lay:
        #     shape_lay = bm.verts.layers.shape.new("uv_flat")


        uv_layer = bm.loops.layers.uv.active
        bm.verts.ensure_lookup_table()
        loose_verts = [] # what for?
        for vert in island_verts:
            if vert.link_loops:
                co = vert.link_loops[0][uv_layer].uv
                # vert[shape_lay] = Vector((co[0], co[1], 0))
                vert.co = Vector((co[0], co[1], 0))
            else:
                loose_verts.append(vert)
        for v in reversed(loose_verts):
            bm.verts.remove(v)

    @staticmethod
    def calc_avg_co_norm_per_island(island_verts):
        # bm.verts.normal_update()
        island_pos = Vector((0, 0, 0))
        island_norm = Vector((0, 0, 0))
        for v in island_verts:
            island_pos += v.co
            island_norm += v.normal
        island_pos /= len(island_verts)
        island_norm.normalize()

        return island_pos, island_norm

    def bbox_calc(self, bm):
        # get min and max vertex position, for bounding box calculation
        min_co = Vector((1000000, 1000000, 1000000))
        max_co = Vector((-1000000, -1000000, -1000000))
        for v in bm.verts:
            min_co.x = min(min_co.x, v.co.x)
            min_co.y = min(min_co.y, v.co.y)
            min_co.z = min(min_co.z, v.co.z)
            max_co.x = max(max_co.x, v.co.x)
            max_co.y = max(max_co.y, v.co.y)
            max_co.z = max(max_co.z, v.co.z)
        return max_co - min_co


    def flatten_mesh(self, face_islands, uv_layer):
        '''Using uv layer'''
        islands_init_area = []
        islands_aver_norm = []
        for _,island_faces in enumerate(face_islands):
            area_before = 0
            # NOTE: calculate the area, average position
            average_position = Vector((0, 0, 0))
            f_cnt = 0
            for f in island_faces:
                f.select = True
                area_before += f.calc_area()
                average_position += f.calc_center_median()
                f_cnt += 1

            average_position /= f_cnt

            average_tangent = Vector((0, 0, 0))
            average_bitangent = Vector((0, 0, 0))

            # calculate a rough tangent and a bitangent
            average_uv_position = Vector((0, 0))
            uv_position_samples = 0

            for face in island_faces:
                for loop in face.loops:
                    uv = loop[uv_layer].uv
                    uv_position_samples += 1
                    average_uv_position += uv
                    delta = loop.vert.co - average_position
                    average_tangent += delta * (uv.x - 0.5)
                    average_bitangent += delta * (uv.y - 0.5)

            # reorient the tangent and bitangent

            average_uv_position /= uv_position_samples
            average_tangent = average_tangent.normalized()
            average_bitangent = average_bitangent.normalized()
            average_normal = average_tangent.cross(average_bitangent).normalized()

            halfvector = average_bitangent + average_tangent
            halfvector /= 2
            halfvector.normalize()
            # NOTE: make sure there is 90deg between tangent and bitangent (by +/- 45deg to halfvector)
            cw = Matrix.Rotation( radians(45.0+180), 4, average_normal)
            ccw = Matrix.Rotation( radians(-45.0+180), 4, average_normal)

            average_tangent = Vector(halfvector)
            average_tangent.rotate(ccw)  #  half - 45 deg

            average_bitangent = Vector(halfvector)
            average_bitangent.rotate(cw) #  half + 45 deg  (90 deg offset in total)

            # new_3d_face.co =  tan * uv.x + bitan * uv.y + average_position
            for face in island_faces:
                for loop in face.loops:
                    uv = loop[uv_layer].uv
                    vert = loop.vert
                    # pos = Vector((0, 0, 0))
                    pos = average_position.copy()
                    pos += average_tangent * -(uv.x - average_uv_position.x)
                    pos += average_bitangent * -(uv.y - average_uv_position.y)
                    # arbitrary - should probably depend on object scale?
                    pos += average_normal * 0.3
                    vert.co = pos
            islands_init_area.append(area_before)
            islands_aver_norm.append(average_normal)
        return islands_init_area, islands_aver_norm



    def execute(self, context):
        obj = context.active_object
        context.scene.tool_settings.particle_edit.use_emitter_deflect = False
        if not obj.data.uv_layers:  # create new one
            self.report({'ERROR'}, 'Object does not have UVs! Cancelling')
            return {"CANCELLED"}
        if not obj.data.shape_keys:
            obj.shape_key_add(name="Basis")
        bm = bmesh.new()
        bm.from_mesh(obj.data)


        if not bm.verts.layers.shape:
            bm.verts.layers.shape.new("Basis")
        # check if we are flattening the mesh to UV

        seam_edges = get_seam_edges(bm)
        if seam_edges:
            ret = bmesh.ops.split_edges(bm, edges=seam_edges)


        # NOTE: find doubles on split seam edges
        bm.verts.ensure_lookup_table()
        boundary_verts = [v for v in bm.verts if v.is_boundary]
        kd = KDTree(len(boundary_verts))
        for v in boundary_verts:
            kd.insert(v.co, v.index)
        kd.balance()

        for v in boundary_verts:
            for (co, index, dist) in kd.find_range(v.co, 0.00001):
                linked_vert = bm.verts[index]
                if linked_vert != v:
                    try:
                        # print(f"new edge from: v.{v.index}-v.{linked_vert.index}")
                        bm.edges.new([v, linked_vert])
                    except:
                        pass


        # NOTE: start handling all the islands
        uv_layer = bm.loops.layers.uv.active
        # all_island_verts = get_verts_islands(bm)
        all_face_islands = get_faces_islands(bm)

        # NOTE: bmesh will be flat now (still merged as one object though)
        islands_init_area, islands_aver_norm = self.flatten_mesh(all_face_islands, uv_layer)

        # NOTE: get islands outline verts, and normalize scale
        all_islands_boundary_verts = []
        all_islands_centers = []
        for i, (area_before, island_faces) in enumerate(zip(islands_init_area, all_face_islands)):
            for v in bm.verts: v.tag = False
            island_verts = []
            for f in island_faces:
                for v in f.verts:
                    if not v.tag:
                        v.tag = True
                        island_verts.append(v)

            # NOTE: calc current center
            center = Vector((0, 0, 0))
            for v in island_verts:
                center += v.co
            center /= len(island_verts)

            # NOTE: compensate scale shrinkage after flattening
            area_after = sum(f.calc_area() for f in island_faces)
            scale_factor = sqrt(area_before[i] / area_after)
            scale_mat = Matrix.Scale(scale_factor, 4)
            to_orign_mat = Matrix.Translation(center)
            bmesh.ops.transform(bm, matrix=to_orign_mat @ scale_mat @ to_orign_mat.inverted(), verts=island_verts)
            # for v in boundary_verts: v.tag = False
            # for e in bm.edges: e.tag = False

            island_boundary_loop, island_boundary_verts = get_boundary_edges(island_verts) # returns boundary loop edges (sorted by topology)

            center = Vector((0, 0, 0))
            for v in island_boundary_verts:
                center += v.co
            center /= len(island_verts)
            all_islands_centers.append(center)
            all_islands_boundary_verts.append(island_boundary_verts)


        # TODO: we should preprocess all boundary verts - find sharp, or linked with multiple vire edges - pairs... or triplets...  v1 - vx,  v1-ny
        # NOTE: go over boundary loops, and select verts with edges that create sharp angles
        # NOTE: get verts that have more than 1 wire edge connected -
        # - they are corners of the pattern and can be marked as sewings boundary, should not be removed
        for i, (average_normal, island_boundary_verts) in enumerate(zip(all_islands_boundary_verts)):

            # NOTE: build curve using verts that are sharp (connecting edges are not parallel)
            # boundary_verts = [v for v in bm.verts if v.is_boundary]
            island_boundary_verts_candidates = []
            junction_verts = {}
            for shared_vert in island_boundary_verts:
                wire_edges = [e for e in shared_vert.link_edges if e.is_wire]

                if len(wire_edges) > 1: # must be a uv corner vert
                    shared_vert.select = True
                    junction_verts[v.index] = True
                    for e in wire_edges:
                        other_vert = e.other_vert(shared_vert)
                        junction_verts[other_vert.index] = True

                    island_boundary_verts_candidates.append(shared_vert)
                else:
                    linked_edges = [e for e in shared_vert.link_edges if not e.is_wire and e.is_boundary] # get lined boundary verts to calc adjacent edges angle
                    p_edge_dir = shared_vert.co - linked_edges[0].other_vert(shared_vert).co
                    p_edge_dir.normalize()
                    edge_dir = linked_edges[1].other_vert(shared_vert).co - shared_vert.co
                    edge_dir.normalize()
                    angle = p_edge_dir.angle(edge_dir)
                    if angle > 0.1: # NOTE: this sucks, use decimate curve from Hair Tool
                        # print(f"sharp angle: {angle} at vert: {shared_vert.index}")
                        shared_vert.select = True
                        island_boundary_verts_candidates.append(shared_vert)
        print(f"junction verts: {junction_verts}")


        return {'FINISHED'}
        for i, (area_before, average_normal, island_boundary_verts) in enumerate(zip(islands_init_area, islands_aver_norm, all_islands_boundary_verts)):
            # print(f"sharp verts: {island_sharp_verts}")
            # NOTE: for each sharp ver loop, create a curve
            if not island_boundary_verts_candidates:
                continue

            # NOTE: create curve object and link it to scene collection
            curve_data = bpy.data.curves.new(name=obj.name+"_pattern"+str(i), type='CURVE')
            curve_data.dimensions = '3D'
            curve_data.resolution_u = 2
            curve_obj = bpy.data.objects.new(obj.name+"_curve_pattern_"+str(i), curve_data)
            context.collection.objects.link(curve_obj)

            curve_spline = curve_data.splines.new('POLY')
            curve_spline.use_cyclic_u = True
            curve_spline.points.add(len(island_boundary_verts_candidates) - 1)
            for i, v in enumerate(island_boundary_verts_candidates):
                curve_spline.points[i].co = (v.co[0], v.co[1], v.co[2], 1)

            # create rot 90 around x axis matrix
            up_vec = Vector((0, 0, 1))
            rot_diff = up_vec.rotation_difference(average_normal)
            rot_mat = Matrix.Identity(3)
            # rot_mat.col[0].xyz = average_normal.xyz
            # rot_mat.col[1].xyz = average_tangent.xyz
            # rot_mat.col[2].xyz = average_bitangent.xyz

            #DONE: project pattern center to XY plane
            print(f"{center=}")
            center_projected = intersect_line_plane(Vector((0,0,0)), Vector((0,0,0)) + average_normal, center, average_normal) # project cartesian center to pattern plane, along normal
            print(f"{center_projected=}")
            to_orign_projected_mat = Matrix.Translation(center_projected)

            rot_mat.rotate(rot_diff)
            mat = rot_mat.to_4x4().inverted() @ to_orign_projected_mat.inverted()
            mat_inv = mat.inverted()
            curve_obj.matrix_world = mat_inv
            # apply mat_inv to all curve points
            for p in curve_spline.points:
                p.co = mat @ p.co

            # TODO: set sharp angle, from before simplify... from higher res mesh/curve..
            # TODO: generate sewings from wire edges..
            curve_obj.data.splines.update()
            GTOOL_OT_CURVE_OT_CSimplify.simplify_curve(curve_obj, self.error, self.sharp_angle_threshold)    # WARNING: simplify maybe in mesh state, to not discard corner verts pairs...
            curve_obj.data.dimensions = '2D'
            curve_obj.data.fill_mode = 'BOTH'


        bm.to_mesh(obj.data)
        # copy flat shape to temp bm
        # shape_lay = bm.verts.layers.shape.get("uv_flat")
        # for vert in bm.verts:
        #     vert.co = vert[shape_lay]

        bm.free()
        # shapeKey = obj.data.shape_keys.key_blocks["uv_flat"]
        # shapeKey.value = make_flat
        return {"FINISHED"}
