# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

from mathutils import Vector
from bpy_extras import view3d_utils
from bpy.app.handlers import persistent
from gpu_extras.batch import batch_for_shader
import gpu
import blf
import bpy
from ..utils.curve_helpers import get_segments_len
from .spline_base import BezierSplines

shader = gpu.shader.from_builtin('UNIFORM_COLOR')
DEPSGRAPH = None
PATTER_SEGMENTS_CACHE = {}

def draw_len_txt(self, context):
    region = context.region
    rv3d = context.region_data

    if bpy.app.version < (4, 0, 0):
        blf.size(0, 16, 60)
    else: # for blender 4
        blf.size(0, 16)

    blf.color(0, 0.8, 0.8, 0.7, 0.8)
    units_scale = context.scene.unit_settings.scale_length
    gpu.state.depth_test_set("LESS")
    for g_idx, garment in enumerate(bpy.context.scene.cloth_garment_data):
        if garment.sewings_overlay:
            for pattern in garment.sewing_patterns:
                pattern_obj = pattern.pattern_obj
                if not pattern.is_enabled or pattern_obj is None or pattern_obj.name not in PATTER_SEGMENTS_CACHE or pattern_obj not in bpy.context.selected_objects:
                    continue
                for spl_id, spl in enumerate(pattern_obj.data.splines):
                    for seg_id in range(len(spl.bezier_points)):
                        position1 = view3d_utils.location_3d_to_region_2d(region, rv3d, PATTER_SEGMENTS_CACHE[pattern_obj.name][spl_id][seg_id][3])
                        seg_len1 = PATTER_SEGMENTS_CACHE[pattern_obj.name+'_len'][spl_id][seg_id]

                        len1 = seg_len1*100
                        blf.position(0, position1[0], position1[1], 0)
                        blf.draw(0, str(round(len1*units_scale, 2)))  # + 'cm'

    gpu.state.depth_test_set("NONE")


def draw_px(self, context):
    # draw_len_txt(self, context)
    # bgl.glDepthFunc(bgl.GL_LEQUAL)  # fix 2.91 second layer dept bug
    gpu.state.blend_set("ALPHA")
    gpu.state.point_size_set(8)
    gpu.state.line_width_set(2.0)
    gpu.state.depth_test_set("LESS")
    # drawing  sewing
    for g_idx, garment in enumerate(bpy.context.scene.cloth_garment_data):
        if garment.sewings_overlay:
            points = []
            if garment.enable_pins:
                [points.extend([pin.source_obj.matrix_world @ Vector(pin.source_co), pin.target_obj.matrix_world @ Vector(pin.target_co)]) for pin in garment.pins if pin.source_obj and pin.target_obj]
                shader.bind()
                batch = batch_for_shader(shader, 'LINES', {"pos": points})
                shader.uniform_float("color", (0.95, 0.95, 0.95, 0.6))
                batch.draw(shader)
                batch = batch_for_shader(shader, 'POINTS', {"pos": points})
                batch.draw(shader)

            for sewing in garment.garment_sewings:
                if sewing.source_obj is None or sewing.source_obj not in context.visible_objects or sewing.source_obj.name not in PATTER_SEGMENTS_CACHE:
                    continue
                if sewing.target_obj is None or sewing.target_obj not in context.visible_objects or sewing.target_obj.name not in PATTER_SEGMENTS_CACHE:
                    continue
                vert_stich_01_list = PATTER_SEGMENTS_CACHE[sewing.source_obj.name]
                vert_stich_02_list = PATTER_SEGMENTS_CACHE[sewing.target_obj.name]

                seg_len1 = PATTER_SEGMENTS_CACHE[sewing.source_obj.name+'_len'][sewing.from_spline_idx][sewing.from_segment_idx]
                seg_len2 = PATTER_SEGMENTS_CACHE[sewing.target_obj.name+'_len'][sewing.to_spline_idx][sewing.to_segment_idx]
                diff_ratio = 2*abs((seg_len1-seg_len2)/(seg_len1+seg_len2)) # to abs value?

                verts_form = vert_stich_01_list[sewing.from_spline_idx][sewing.from_segment_idx]
                verts_to = vert_stich_02_list[sewing.to_spline_idx][sewing.to_segment_idx]
                if sewing.flip:
                    verts_to = list(reversed(verts_to))

                points = []
                for (point_from, point_to) in zip(verts_form, verts_to):
                    points.extend([(point_from.x, point_from.y, point_from.z), (point_to.x, point_to.y, point_to.z)])
                batch = batch_for_shader(shader, 'LINES', {"pos": points})
                shader.bind()
                col = (0.85, max(0, 0.85-2*diff_ratio), max(0, 0.85-2*diff_ratio), 0.55) if diff_ratio > 0.002 else (0.95, 0.95, 0.95, 0.65)
                shader.uniform_float("color", col)
                batch.draw(shader)

    gpu.state.depth_test_set("NONE")
    gpu.state.line_width_set(1)
    gpu.state.point_size_set(1)
    gpu.state.blend_set("NONE")


def any_draw_sew_on():
    return any([garment.sewings_overlay for garment in bpy.context.scene.cloth_garment_data])

def disable_all_draw_sewings():
    for garment in bpy.context.scene.cloth_garment_data:
        garment.sewings_overlay = False

def get_sewing_patterns_flat():
    return [pattern.pattern_obj for garment in bpy.context.scene.cloth_garment_data if garment.sewings_overlay for pattern in garment.sewing_patterns if pattern.pattern_obj and pattern.is_enabled]


def get_sewing_patterns():
    ''' return list of [(garment_idx, [patterns]), ...] '''
    return [(g_idx, [pattern.pattern_obj for pattern in garment.sewing_patterns if pattern.pattern_obj and pattern.is_enabled]) for g_idx, garment in enumerate(bpy.context.scene.cloth_garment_data) if garment.sewings_overlay]

@persistent
def draw_sew_scn_update(scene):
    check_we_need_update()

@persistent
def depsgraph_update(scene):
    global DEPSGRAPH, PATTER_SEGMENTS_CACHE
    DEPSGRAPH = bpy.context.evaluated_depsgraph_get()
    PATTER_SEGMENTS_CACHE.clear()
    for pattern_obj in get_sewing_patterns_flat():
        update_pattern_segs(pattern_obj)

handle_SpaceView3D = None
handle_SpaceView2D = None

def handle_handlers_draw_sewings():
    global handle_SpaceView3D, handle_SpaceView2D
    if any_draw_sew_on():
        if handle_SpaceView3D is None:
            args = (None, bpy.context)  # u can pass arbitrary class as first param  Instead of (self, context)
            handle_SpaceView3D = bpy.types.SpaceView3D.draw_handler_add(draw_px, args, 'WINDOW', 'POST_VIEW')

        if handle_SpaceView2D is None:
            args = (None, bpy.context)  # u can pass arbitrary class as first param  Instead of (self, context)
            handle_SpaceView2D = bpy.types.SpaceView3D.draw_handler_add(draw_len_txt, args, 'WINDOW', 'POST_PIXEL')

        if draw_sew_scn_update not in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.append(draw_sew_scn_update)

        if depsgraph_update not in bpy.app.handlers.undo_post:
            bpy.app.handlers.undo_post.append(depsgraph_update)

        if depsgraph_update not in bpy.app.handlers.redo_post:
            bpy.app.handlers.redo_post.append(depsgraph_update)
    else:
        PATTER_SEGMENTS_CACHE.clear()

        if handle_SpaceView3D is not None:
            bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView3D, 'WINDOW')
            handle_SpaceView3D = None

        if handle_SpaceView2D is not None:
            bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView2D, 'WINDOW')
            handle_SpaceView2D = None

        if draw_sew_scn_update in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.remove(draw_sew_scn_update)

        if depsgraph_update in bpy.app.handlers.undo_post:
            bpy.app.handlers.undo_post.remove(depsgraph_update)

        if depsgraph_update in bpy.app.handlers.redo_post:
            bpy.app.handlers.redo_post.remove(depsgraph_update)


def update_pattern_segs(pattern_obj):
    global PATTER_SEGMENTS_CACHE
    for obj in get_sewing_patterns_flat():
        #* update all instances, cos we using world space pts cache (so 2 instances with different mat_world == 2 caches)
        if obj.data == pattern_obj.data:
            # print(f'Updating {pattern_obj.name} cache')
            b_spl = BezierSplines(obj)
            resampled_splines_segments = b_spl.generate_splines_pts_cloud(8, world_space=True)
            segment_len = [get_segments_len(resampled_segments) for resampled_segments in resampled_splines_segments]
            PATTER_SEGMENTS_CACHE[obj.name] = resampled_splines_segments  # pattern_segments_cache[obj.name][spline_idx][segment_idx]
            PATTER_SEGMENTS_CACHE[obj.name+'_len'] = segment_len   # patter_segments_cache[obj.name][spline_idx][segment_idx]

def check_we_need_update():
    active_obj = bpy.context.active_object
    any_draw_sew_on = any([garment.sewings_overlay for garment in bpy.context.scene.cloth_garment_data])
    if any_draw_sew_on:
        if hasattr(DEPSGRAPH, 'updates'):
            sew_patterns = get_sewing_patterns_flat()
            for update in DEPSGRAPH.updates:
                if (update.id.original in sew_patterns and (update.is_updated_geometry or update.is_updated_transform)):
                    # print((f'ID {update.id}, geom {update.is_updated_geometry}, xform {update.is_updated_transform}'))
                    update_pattern_segs(update.id.original)
                    # return

def check_enable_draw_sewings(self, context):
    ''' called only once form garment.draw_overlay update '''
    global DEPSGRAPH, PATTER_SEGMENTS_CACHE
    DEPSGRAPH = context.evaluated_depsgraph_get()
    handle_handlers_draw_sewings()
    if self.sewings_overlay: # we enabeld it in garment props
        g_idx = int(self.path_from_id().split('.')[-1][-2])
        bpy.ops.garment.cleanup(garment_index=g_idx) #mostly for cleaning up broken pins- if user removed empty
        for pattern in self.sewing_patterns:
            if pattern.is_enabled and pattern.pattern_obj:
                    update_pattern_segs(pattern.pattern_obj)
    else:
        PATTER_SEGMENTS_CACHE.clear()

@persistent
def LoadPostDisableCurvDebug(scn):
    # handle_handlers_draw_sewings()
    #since it dosent work anyway on scene load. Plus prevets Draw Xray from running in background
    disable_all_draw_sewings()

def register():
    bpy.app.handlers.load_post.append(LoadPostDisableCurvDebug)

def unreg():
    global handle_SpaceView3D, handle_SpaceView2D
    if handle_SpaceView3D is not None:
        bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView3D, 'WINDOW')
        handle_SpaceView3D = None

    if handle_SpaceView2D is not None:
        bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView2D, 'WINDOW')
        handle_SpaceView2D = None

    if draw_sew_scn_update in bpy.app.handlers.depsgraph_update_post:
        bpy.app.handlers.depsgraph_update_post.remove(draw_sew_scn_update)

    bpy.app.handlers.load_post.remove(LoadPostDisableCurvDebug)
    for garment in bpy.context.scene.cloth_garment_data:
        garment.sewings_overlay = False

