# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import numpy as np
import mathutils
from mathutils import Vector, Matrix
from collections import OrderedDict
import itertools
from copy import copy, deepcopy
from bpy_extras import view3d_utils
from ..utils.helper_functions import *
from ..utils.curve_helpers import *
from math import sqrt, cos, pi, acos, degrees

class CutPointInfo:
    ''' placement in ['INSIDE','BORDER' ] of sewing pattern '''
    def __init__(self, spl_id, p_id, click_pos, placement='BORDER'):
        self.spl_id = spl_id  # spline id where cut is to be performed
        self.p_id = p_id  # point id where cut is to be performed
        self.click_pos = click_pos # where cut was placed
        self.placement = placement # Inside, or on border of pattern?

class SewingPtr():
    def __init__(self, sewing=None):
        self.spl_id, self.seg_id = (None, None)
        self.change_dir = False # will we flip garment.sewing[x].flip - sewing direction. => garment.sewing[x].flip != self.flip
        if sewing:
            if len(sewing) == 3:
                self.spl_id, self.seg_id, self.change_dir = sewing[0], sewing[1], sewing[2]
            else:
                self.spl_id, self.seg_id  = sewing[0], sewing[1]
                self.change_dir = False

    def reset(self):
        self.spl_id = -1
        self.seg_id = -1
        self.change_dir = False


class BezierPoint(object):  # do we even need it...
    def __init__(self, co, handle_left_type='AUTO', handle_left=Vector((0, 0, 0)), handle_right_type='AUTO', handle_right=Vector((0, 0, 0)), select = False, lh_select = False, rh_select = False, sewing=None):
        ''' ‘FREE’, ‘VECTOR’, ‘ALIGNED’, ‘AUTO’ '''
        self.co = deepcopy(co)
        self.handle_left = deepcopy(handle_left)
        self.handle_left_type = handle_left_type
        self.select_left_handle = lh_select
        self.handle_right = deepcopy(handle_right)
        self.handle_right_type = handle_right_type
        self.select_right_handle = rh_select

        self.sewing = SewingPtr(sewing) if sewing else SewingPtr((-1, -1, False))
        self.select_control_point = select
        self.tag = False #helper

    @classmethod
    def from_bezier_point(cls, bezier_point, sewing=None):
        if not sewing and isinstance(bezier_point, BezierPoint):
            sewing = (bezier_point.sewing.spl_id, bezier_point.sewing.seg_id, bezier_point.sewing.change_dir) if bezier_point.sewing else None
        return cls(bezier_point.co, bezier_point.handle_left_type, bezier_point.handle_left, bezier_point.handle_right_type, bezier_point.handle_right,
             select=bezier_point.select_control_point, lh_select=bezier_point.select_left_handle, rh_select=bezier_point.select_right_handle, sewing=sewing)

    def reverse_handles(self):
        back_left_type = self.handle_left_type
        back_left_pos = self.handle_left
        #! maybe first set type like in copy_to_target ???
        self.handle_left_type = self.handle_right_type
        self.handle_left = self.handle_right

        self.handle_right_type = back_left_type
        self.handle_right = back_left_pos

    def mirror_handles(self, use_right=True):
        ''' use_right - > copy right handle to left handle '''
        if use_right:
            if self.handle_right_type in {'ALIGNED', 'AUTO'}:
                self.handle_right.y = self.co.y  # make it horizontal
            self.handle_left_type = self.handle_right_type
            self.handle_left = copy(self.handle_right)
            self.handle_left.x *= -1
        else:
            # self.handle_right_type = self.handle_left_type
            # if self.handle_left_type in {'ALIGNED', 'FREE'}: # we can write co. onlyt to 'FREE' and 'ALIGNED' handles..
            #     self.handle_left.y = self.co.y  # make it horizontal ?
            if self.handle_left_type in {'ALIGNED', 'AUTO'}:
                self.handle_left.y = self.co.y  # make it horizontal
            self.handle_right_type = self.handle_left_type
            self.handle_right = copy(self.handle_left)
            self.handle_right.x *= -1

    def mirror_point(self):
        # if self.handle_left_type == 'FREE':
        self.handle_left.x = -1 * self.handle_left.x
        # if self.handle_right_type == 'FREE':
        self.handle_right.x = -1 * self.handle_right.x
        self.co.x = -1 * self.co.x

    def copy_to_target(self, bezier_point):
        bezier_point.co = self.co

        bezier_point.handle_left_type = self.handle_left_type
        bezier_point.handle_right_type = self.handle_right_type

        bezier_point.handle_left = self.handle_left
        bezier_point.handle_right = self.handle_right
        bezier_point.select_left_handle = self.select_left_handle
        bezier_point.select_right_handle = self.select_right_handle
        bezier_point.select_control_point = self.select_control_point

exp3_1 = np.array([3, 2, 1, 0], 'f')
exp4 = np.array([[4], [3], [2], [1], [0]], 'f')
boundaries = np.array([0, 1], 'f')
cubic_coeffs = np.array([[-1, 3, -3, 1], [3, -6, 3, 0], [-3, 3, 0, 0], [1, 0, 0, 0]], 'f')

class BezierSpline(object):
    def __init__(self, b_points=None, sewings=[None], cyclic=True, space='2D'):
        ''' takes data.spline '''
        self.points = []
        if b_points:
            self.points.extend([BezierPoint.from_bezier_point(p, sewing) for p, sewing in itertools.zip_longest(b_points, sewings[:len(b_points)])])  # ?zip to p_points len
        self.cyclic = cyclic
        self.space = space

    @classmethod
    def from_obj_spline(cls, obj, spline_id):
        spl = obj.data.splines[spline_id]
        def_sewings = [(spline_id, i, False) for i in range(len(spl.bezier_points))]
        return cls(spl.bezier_points, def_sewings, space=obj.data.dimensions, cyclic=spl.use_cyclic_u)

    @property
    def length(self):
        return len(self.points)

    def euclidean_length(self, res=9, seg_id=None, matrix=None):
        seg_ids = [seg_id] if seg_id else list(range(self.seg_count))
        if matrix:
            mat = matrix if type(matrix) == np.ndarray else np.array(matrix, 'f')
            out_dim = 3

        total_len = 0
        for seg_id in seg_ids:
            coeffs = self.__create_coefficients(seg_id)
            tn = np.linspace(0, 1, num=res, endpoint=True)
            np_seg_pts = self.evaluate(coeffs, tn)  # shape(res, nD)  3D or 2D

            # * multiply by mat (use einsum)
            if matrix:
                mat_dim_diff = mat.shape[0] - np_seg_pts.shape[1]  # for matching mat_dimension to pts dim
                if mat_dim_diff:  # add col to np_seg_pts (dim_diff cols)
                    np_seg_pts = self.__expand_cols_np(np_seg_pts, mat_dim_diff)
                    if mat.shape[0] == 4:  # for point (x,y,z, 1)
                        np_seg_pts[:,-1] = 1
                np_seg_pts = np.einsum('ij,aj->ai', mat, np_seg_pts)  # to world space

            seg_lens = np.apply_along_axis(np.linalg.norm, 1, np_seg_pts[:-1, :]-np_seg_pts[1:, :]) # (..,1) 4th axis does not affect length it seems
            total_len += np.sum(seg_lens)
        return total_len

    @property
    def seg_count(self):
        return self.length if self.cyclic else self.length -1

    def __delitem__(self, idx):
        del self.points[idx]

    def __getitem__(self, idx):  # overload [] get
        return self.points[idx]

    def __setitem__(self, idx, point, sewing=None):
        ''' point can be blender bezier, or BezierPoint '''
        self.points[idx] = BezierPoint.from_bezier_point(point, sewing)

    def get_prev_point(self, idx):
        return self.points[idx-1] if idx > 0 else self.points[-1] if self.cyclic else None

    def get_next_point(self, idx):
        return self.points[idx+1] if idx < self.length else self.points[0] if self.cyclic else None

    def rotate(self, n):
        if n > 0:
            for _ in range(n):
                # tmp = self.points
                self.points.insert(0, self.points.pop())
        elif n < 0:
            for _ in range(-1*n):
                self.points.append(self.points.pop(0))

    @staticmethod
    def solve_cubic_roots(coeffs):
        #https://www.particleincell.com/2013/cubic-line-intersection/
        ''' replaced with numpy roots '''
        def filter_ts(li):
            return [t for t in li if 0 <= t and t <= 1.0]

        a = coeffs[0]
        b = coeffs[1]
        c = coeffs[2]
        d = coeffs[3]

        eps = 0.0001
        if abs(a) < eps:
            if abs(b) < eps:
                if abs(c) < eps:  # no solutions
                    return []

                # Linear formula
                return filter_ts([-1 * (d / c)])

            #Quadratic formula detected
            q = pow(c, 2) + 4*b*d  # quadratic discriminant
            if q >= 0:
                q = sqrt(q)
                return filter_ts([(- q - c)/(2 * b), (q - c)/(2 * b)])
            else:
                return []

        A = b/a
        B = c/a
        C = d/a

        # print coeffs
        from math import sqrt, acos, cos, pi

        Q = (3*B - A*A)/9
        R = (9*A*B - 27*C - 2*A*A*A)/54
        D = Q*Q*Q + R*R  # polynomial discriminant

        ts = []

        # complex or duplicate roots
        if D >= 0:
            S = sgn(R + sqrt(D))*pow(abs(R + sqrt(D)), (1./3.))
            T = sgn(R - sqrt(D))*pow(abs(R - sqrt(D)), (1./3.))

            ts.append(-A/3 + (S + T))      # real root
            ts.append(-A/3 - (S + T)/2)    # real part of complex root
            ts.append(-A/3 - (S + T)/2)    # real part of complex root
            Im = abs(sqrt(3)*(S - T)/2)    # complex part of root pair

            #discard complex roots
            if Im != 0:
                del(ts[1])
                del(ts[2])
        else:
            try:
                th = acos(R/sqrt(-Q*Q*Q))
            except ZeroDivisionError:
                th = acos(sgn(R) * 1e20)

            ts.append(2*sqrt(-Q)*cos(th/3) - A/3)
            ts.append(2*sqrt(-Q)*cos((th + 2*pi)/3) - A/3)
            ts.append(2*sqrt(-Q)*cos((th + 4*pi)/3) - A/3)

        return filter_ts(ts)

    def get_uniform_tns(self, seg_points, res):  # return array <0,1> - where 1 is for longest strand
        # uniform_spacing: #adjust uniform spaced t (non xyz uniform), by segments legnths - makes points evennly distributed in Euclidean space
        points_count = seg_points.shape[0]  # 2x knots + handles
        domain = np.linspace(0, 1, points_count, endpoint=True) #defines ouput pt count
        seg_lens = np.apply_along_axis(np.linalg.norm, 1, seg_points[:-1, :]-seg_points[1:, :])  # do an - an-1 and get length
        t_spat = np.insert(seg_lens, 0, 0).cumsum()  # add zero and sum
        t_spat = t_spat/t_spat[-1]  #it will be now proportional to segments distances - bigger point spacing - more time.
        t_in = np.linspace(0, 1, res)
        corr_t_ins = np.interp(t_in,  t_spat, domain)   # convert time t_in, using spatial spacing t_spat, to output domain
        return corr_t_ins  # remap from <0,1> to <0, point_count-1>

    def __create_coefficients(self, seg_id):
        '''f(t) = pts[0] * t^3 + pts[1] * t^2 + pts[2] * t^1 + pts[3]
        seg_points - bezier p[seg_id], handle_r, handle_l, p[seg_id+1]
        cubic_coeffs (4,4) @ pts (4,2) -> [4,2] (2 for 2d )
        # return bezier coefficeints
        coeff[:,0] - x
        coeff[:,1] - y
        coef[0,:] - t^3
        coef[1,:] - t^2
        coef[2,:] - t
        coef[3,:] - 1
        '''
        pts = self.get_segment_points(seg_id)
        seg_points = np.array([p.xy for p in pts], 'f') if self.space == '2D' else np.array(pts, 'f')
        return np.einsum('ij,jk->ik', cubic_coeffs, seg_points)

    def evaluate_at(self, seg_id, t=None):
        if t == None: #treat seg_id as t_n (float)
            seg, t = divmod(seg_id, 1)
            seg_id = int(seg)
        t = np.array([t])
        coeffs = self.__create_coefficients(seg_id)
        tn = np.power(t[:, None], exp3_1)  # eg tn is eg (8, 4) , coeffs are (4, 2)
        return Vector(np.einsum('ij,jk->ik', tn, coeffs)[0].tolist()).to_3d()

    def evaluate(self, coeffs, t):
        ''' return len(t)  2d points '''
        if type(t) != np.ndarray:
            t = np.array(t, 'f')
        tn = np.power(t[:,None], exp3_1) # eg tn is eg (8, 4) , coeffs are (4, 2)
        return np.einsum('ij,jk->ik', tn, coeffs) #mul j's -> output (8,2)

    @staticmethod
    def __expand_cols_np(in_array, shape_diff):
        # https://stackoverflow.com/a/8505658/13588868
        rows_cnt = in_array.shape[0]
        return in_array[:, :shape_diff] if shape_diff < 0 else np.c_[in_array, np.zeros((rows_cnt, shape_diff))]

    @staticmethod
    def __np_mat_multiply(pts, mat):
        mat_dim_diff = mat.shape[0] - pts.shape[1]  # for matching mat_dimension to pts dim
        if mat_dim_diff:  # add col to np_seg_pts (dim_diff cols)
            pts = BezierSpline.__expand_cols_np(pts, mat_dim_diff)
            if mat.shape[0] == 4:  # for point (x,y,z, 1)
                pts[:, -1] = 1
        return np.einsum('ij,aj->ai', mat, pts)

    def eval_resampled(self, resolution, matrix=None, uniform=True, out_dim=2):
        ''' generate n pts from eval spline'''
        seg_points = []
        if matrix:
            mat = np.array(matrix, 'f')
            out_dim = 3
        seg_cnt = self.seg_count
        for seg_id in range(seg_cnt):
            coeffs = self.__create_coefficients(seg_id)
            res = resolution[seg_id] if type(resolution) is list else resolution  # res may be list for each seg_id
            tn = np.linspace(0, 1, num=res, endpoint=True)
            np_seg_pts = self.evaluate(coeffs, tn) # shape(res, 2)  2 - xy
            if uniform:
                tn_adj = self.get_uniform_tns(np_seg_pts, res)  # adjust tn to get even spacing
                np_seg_pts = self.evaluate(coeffs, tn_adj)  # shape(res, 2)  2 - xy

            # * multiply by mat (use einsum)
            if matrix:
                np_seg_pts = self.__np_mat_multiply(np_seg_pts, mat)

            #* matc=h vector out_dim size
            target_dim_diff = out_dim - np_seg_pts.shape[1]
            if target_dim_diff:
                np_seg_pts = self.__expand_cols_np(np_seg_pts, target_dim_diff)
            seg_points.append(np_seg_pts)
        return [[Vector(p) for p in pts] for pts in seg_points]

    def __spl_pt_distance(self, coeffs,  pt):
        # https://stackoverflow.com/questions/2742610/closest-point-on-a-cubic-bezier-curve
        # These are the coefficients of the derivatives d/dx and d/(d/dx).
        da = 6*np.sum(coeffs[0]*coeffs[0])
        db = 10*np.sum(coeffs[0]*coeffs[1])
        dc = 4*(np.sum(coeffs[1]*coeffs[1]) + 2*np.sum(coeffs[0]*coeffs[2]))
        dd = 6*(np.sum(coeffs[0]*(coeffs[3]-pt)) + np.sum(coeffs[1]*coeffs[2]))
        de = 2*(np.sum(coeffs[2]*coeffs[2])) + 4*np.sum(coeffs[1]*(coeffs[3]-pt))
        df = 2*np.sum(coeffs[2]*(coeffs[3]-pt))

        dda = 5*da
        ddb = 4*db
        ddc = 3*dc
        ddd = 2*dd
        dde = de
        dcoeffs = np.stack([da, db, dc, dd, de, df])
        ddcoeffs = np.stack([dda, ddb, ddc, ddd, dde]).reshape(-1, 1)

        # Calculate the real extremes, by getting the roots of the first
        # derivativ of the distance function.
        extrema = self.__np_real_roots(dcoeffs)
        # Remove the roots which are out of bounds of the clipped range [0, 1].
        # [future reference] https://stackoverflow.com/questions/47100903/deleting-every-3rd-element-of-a-tensor-in-tensorflow
        dd_clip = (np.sum(ddcoeffs * np.power(extrema, exp4)) >= 0) & (extrema > 0) & (extrema < 1)
        minima = extrema[dd_clip]

        # Add the start and end position as possible positions.
        potentials = np.concatenate((minima, boundaries))

        # Calculate the points at the possible parameters t and
        # get the index of the closest
        points = self.evaluate(coeffs, potentials)
        distances = np.sum(np.square(points - pt), axis=1)
        index = np.argmin(distances)

        return points, distances, potentials, index


    def get_segment_points(self, seg=0):
        count = len(self.points)
        p0 = self.points[seg].co
        p1 = self.points[seg].handle_right
        p2 = self.points[(seg+1) % count].handle_left
        p3 = self.points[(seg+1) % count].co
        return p0, p1, p2, p3

    def closest_point_view_2d(self, context, mouse_2d, matrix=None):
        ''' Return the closest POINT_3d, t; to the point mouse_2d (view space 2d)
        and the curve.'''
        best_dist = 100 # px max
        best_point = None
        best_t = None
        if self.space == '2D':  # cos location_3d_to_region_2d won work on 2d curves..., also not required
            print('closest_pt_view() not for 2d curves!')
            return None,None
        region = context.region
        rv3d = context.region_data

        # view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, mouse_2d)
        # ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, mouse_2d)
        def create_view_coefficients(self, seg_id): #* get spline in to flat 2d screen space
            pts = self.get_segment_points(seg_id)
            if matrix is not None: #usually to world space
                pts = self.__np_mat_multiply(np.array(pts,'f'), matrix)  # gives 4d
                pts = [Vector(p) for p in self.__expand_cols_np(pts, -1)]  # shrink back to 3d
            np_seg_pts = np.array([view3d_utils.location_3d_to_region_2d(region, rv3d, p) for p in pts], 'f')
            if np.isnan(np.sum(np_seg_pts)): #fastest way to check nan
                return None
            return np.einsum('ij,jk->ik', cubic_coeffs, np_seg_pts)

        for seg_id in range(self.seg_count):
            coeffs_2d_view = create_view_coefficients(self, seg_id)
            if coeffs_2d_view is None:
                return None, None
            points, distances, ts, index = self.__spl_pt_distance(coeffs_2d_view, mouse_2d)
            if distances[index] < best_dist:
                best_dist = distances[index]
                best_point = Vector(points[index])
                best_t = ts[index]+seg_id
        return best_point, best_t


    def closest(self, pt):
        ''' Return the closest POINT, T, between the point pt (local space 2d or 3d)
        and the curve.'''
        best_dist = 9999
        best_point = None
        best_t = None
        pt = np.array(pt.xy, 'f') if self.space == '2D' else np.array(pt.xyz, 'f')
        for seg_id in range(self.seg_count):
            coeffs = self.__create_coefficients(seg_id)
            points, distances, ts, index = self.__spl_pt_distance(coeffs, pt)
            if distances[index] < best_dist:
                best_dist = distances[index]
                best_point = Vector(points[index])
                best_t = ts[index]+seg_id
        return best_point, best_t

    @staticmethod
    def __np_real_roots(coefficients, EPSILON=1e-6):
        r = np.roots(coefficients)
        return r.real[abs(r.imag) < EPSILON]

    # #computes intersection between a cubic spline and a line segment*/
    def __computeIntersections(self, seg_id, l_a, l_b):
        # https://stackoverflow.com/a/57315396
        A = l_b.y-l_a.y  # A=y2-y1
        B = l_a.x-l_b.x  # B=x1-x2
        C = l_a.x*(l_a.y-l_b.y) + l_a.y*(l_b.x-l_a.x)  # C=x1*(y1-y2)+y1*(x2-x1)
        abc_line = np.array([A, B, C], 'f')  # line eq
        coeffs = self.__create_coefficients(seg_id)

        # have the equation Ax+By+C = 0.
        # We can however rewrite both x and y as functions of t.
        # This will then give you A*x(t) + B*y(t) + C = 0, which is now an equation in only a single unknown(t) that we can find root(s) for.
        # A Bezier curve can be converted to form x = bx[0]*t ^ 3 + bx[1]*t ^ 2 + bx[2]*t + bx[3](ignore the coefficient ordering, in hindsight they should be numbered in reverse).
        # Similar equation exists for y.
        # You can plug these into the A*x(t) + B*y(t) + C = 0 equation. You will then obtain the following eq
        # (A*bx[0]+B*by[0])*t ^ 3 + (A*bx[1]+B*by[1])*t ^ 2 + (A*bx[2]+B*by[2])*t + (A*bx[3]+B*by[3]+C) = 0, or P0*t ^ 3 + P1*t ^ 2 + P2*t + P3 = 0.

        ceff2 = np.append(coeffs, np.array([[0], [0], [0], [1]], 'f'), axis=1)  # bezier(x) + bezier(y) + bezier_C
        # A * coef2[:,0] + B * coef2[:,1] + C * coef2[:,2]
        P = np.sum(abc_line * ceff2, axis=1)  # line ABC eq with bezier line coeefs
        # verify the roots are in bounds of the linear segment
        extrema_t = self.__np_real_roots(P)
        delta = 0.005
        extrema_t = extrema_t[(extrema_t > 0+delta) & (extrema_t < 1-delta)]  # remove out of segment extremas
        if len(extrema_t) == 0:
            return []
        spl_eval_pts = self.evaluate(coeffs, extrema_t) #2d pts list
        intersections = []
        for t, pt in zip(extrema_t.tolist(), spl_eval_pts.tolist()):
            # above is intersection point assuming infinitely long line segment,
            #  make sure we are also in bounds of the line*/
            if (l_b.x-l_a.x) != 0:  # if not vertical line*/
                s = (pt[0]-l_a.x)/(l_b.x-l_a.x)
            else:
                s = (pt[1]-l_a.y)/(l_b.y-l_a.y)

            if 0 < s and s < 1.0:# in bounds?
                intersections.append((t,Vector(pt)))
        return intersections


    def spline_line_intersection(self, line_a, line_b):
        ''' return spline line intersection points '''
        all_intersections = []
        for seg_id in range(self.seg_count):
            inters = self.__computeIntersections(seg_id, line_a, line_b)
            all_intersections.extend([(seg_id+t,pt) for t,pt in inters])
        return all_intersections

    def get_euclid_t(self, seg_id, t_s):  # return array <0,1> - where 1 is for longest strand
        ''' convert single t - in time domain, to euclidean distance t_spatial '''
        coeffs = self.__create_coefficients(seg_id)
        tn = np.linspace(0, 1, num=12, endpoint=True)
        seg_points = self.evaluate(coeffs, tn)  # helper pts to calc euclid dist.
        # uniform_spacing: #adjust uniform spaced t (non xyz uniform), by segments legnths - makes points evennly distributed in Euclidean space
        points_count = seg_points.shape[0]  # 2x knots + handles
        domain = np.linspace(0, 1, points_count, endpoint=True)  # defines ouput pt count
        seg_lens = np.apply_along_axis(np.linalg.norm, 1, seg_points[:-1, :]-seg_points[1:, :])  # do an - an-1 and get length
        t_spatial = np.insert(seg_lens, 0, 0).cumsum()  # add zero and sum
        t_spatial = t_spatial/t_spatial[-1]  # it will be now proportional to segments distances - bigger point spacing - more time.
        corr_t_spat = np.interp(t_s,  t_spatial, domain)  # convert t_s, using spatial spacing t_spatial, to output domain
        return corr_t_spat

    def split_seg_t(self, t_n, spatial_space = False):
        seg_count = self.seg_count
        seg_id, t = divmod(t_n, 1)
        seg_id = int(seg_id)
        if spatial_space:
            t = self.get_euclid_t(seg_id, t)  # adjust time  t to euclid t_spatial
            # np_seg_pts = self.evaluate(coeffs, tn_adj)  # shape(res, 2)  2 - xy

        p0, p1, p2, p3 = self.get_segment_points(seg_id)
        p01 = p0.lerp(p1, t)
        p12 = p1.lerp(p2, t)
        p23 = p2.lerp(p3, t)
        p01_12 = p01.lerp(p12, t)
        p12_23 = p12.lerp(p23, t)
        p0112_1223 = p01_12.lerp(p12_23, t)

        firsthalf = [p0, p01, p01_12, p0112_1223]
        secondhalf = [p0112_1223, p12_23, p23, p3]

        self.points[seg_id].handle_left_type = 'FREE'
        self.points[seg_id].handle_right_type = 'FREE'
        self.points[seg_id].handle_right = p01
        self.points[(seg_id+1) % seg_count].handle_right_type = 'FREE'
        self.points[(seg_id+1) % seg_count].handle_left_type = 'FREE'
        self.points[(seg_id+1) % seg_count].handle_left = p23
        new_pt = BezierPoint(p0112_1223, handle_left=p01_12, handle_right=p12_23, handle_left_type='FREE', handle_right_type='FREE', select=True, lh_select=True, rh_select=True)
        new_pt_id = seg_id+1 # no mod- just insert at last pos [len(arra)]
        self.points.insert(new_pt_id, new_pt)
        return new_pt_id

    def get_center(self):
        '''aver spl center pos'''
        result = Vector((0,0,0))
        for pt in self.points:
            result += pt.co
        return result/self.length

    def get_signed_area(self):
        ''' + clockwise, - ccw'''
        # sum += x1*y2 - x2*y1 + x2*y3 - x3*y2 + ...
        count = self.seg_count #? maybe should be always len(self.pts)? cyclic only support..
        result = 0
        for i in range(count):
            result += self.points[i].co.x * self.points[(i+1) % count].co.y - self.points[(i+1) % count].co.x * self.points[i].co.y
        return -result

    def append(self, point, sewing=None):
        self.points.append(BezierPoint.from_bezier_point(point, sewing))

    def insert(self, index, points, sewings=[None]):
        temp_points = [BezierPoint.from_bezier_point(p, sewing) for p, sewing in itertools.zip_longest(points, sewings)]
        self.points[index:index] = temp_points

    def reset_sewings(self):
        for p in self.points:
            p.sewing.spl_id = -1

    def mirror(self):
        [p.mirror_point() for p in self.points]

    def mirrored(self):
        new_b_spl = BezierSpline()
        new_b_spl.points = deepcopy(self.points)
        new_b_spl.mirror()
        new_b_spl.cyclic = self.cyclic
        return new_b_spl

    def split_by_two_ids(self, splice_points):
        ''' return two BezierSpline(s)  created from spliting  blender_bezier_points '''
        if len(splice_points) != 2:
            return
        sp = sorted(splice_points)  # if we have split by points (3,6), we have to do points[3:6+1] - to get [3,4,5,6]
        spl_a_b = BezierSpline(deepcopy(self.points[sp[0]:sp[1]+1]))
        spl_b_a = BezierSpline(deepcopy(self.points[sp[1]:] + self.points[:sp[0]+1]))
        return (spl_a_b, spl_b_a)

    def resample_pts(self, res, mat=None):
        ''' generate resampled spl pts - cheaper/simpler ver of self.eval_resampled()
        !Does not include segments endpoints!!
        '''
        seg_points = []
        for seg_id in range(self.seg_count):
            coeffs = self.__create_coefficients(seg_id)
            seg_points.append(self.evaluate(coeffs, np.linspace(0, 1, num=res, endpoint=False)))  # !endpoint=False
        if mat:
            return [[mat @ Vector(p).to_3d() for p in pts] for pts in seg_points]
        else:
            return [[Vector(p) for p in pts] for pts in seg_points]


    def is_point_inside_winding(self, pt):
        pt = pt.to_2d()
        flat_pts = [p for seg_pts in self.resample_pts(8) for p in seg_pts]
        rad = 0
        prev_V = flat_pts[-1]
        for i in range(len(flat_pts)):
            next_V = flat_pts[i]
            rad += (prev_V-pt).angle_signed((next_V-pt))
            prev_V = flat_pts[i]
        if abs(abs(rad) - 6.283) < 0.1:  # is it close to 2pi= 6.283 ?
            return True
        return False

    def pointIsInPoly_raycast(self, p):
        isInside = False
        pts = [p for seg_pts in self.resample_pts(8) for p in seg_pts]
        minX = pts[0].x
        maxX = pts[0].x
        minY = pts[0].y
        maxY = pts[0].y
        for point in pts:
            minX = min(point.x, minX)
            maxX = max(point.x, maxX)
            minY = min(point.y, minY)
            maxY = max(point.y, maxY)

        if (p.x < minX or p.x > maxX or p.y < minY or p.y > maxY):
            return False

        i = 0
        j = len(self.points) - 1
        for i in range(len(pts)):
            if ((pts[i].y > p.y) != (pts[j].y > p.y) and p.x < (pts[j].x - pts[i].x) * (p.y - pts[i].y) / (pts[j].y - pts[i].y) + pts[i].x):
                isInside = not isInside
            j = i
        return isInside

    def __rev_sewings(self, pts):
        ''' Required flipping sewing when spline direction is reversed'''
        #this offsets +1 sewing, since  we have:
        #A - 0 - B - 1 - C - 2 -(A)  but after reverse
        #C - 2 - B - 1 - A - 0
        first_sewing_back = copy(pts[0].sewing)
        list_len = len(pts)
        for i, p in enumerate(pts):
            p.sewing.spl_id, p.sewing.seg_id = pts[(i+1) % list_len].sewing.spl_id, pts[(i+1) % list_len].sewing.seg_id
            p.sewing.change_dir = not p.sewing.change_dir #we need to flip sewing direction too
        pts[-1].sewing.spl_id, pts[-1].sewing.seg_id = first_sewing_back.spl_id, first_sewing_back.seg_id

    def reverse(self):
        ''' Reverse point order '''
        self.points.reverse()
        [p.reverse_handles() for p in self.points]  # ? we are not using deep copy, so this affects original splie too...
        self.__rev_sewings(self.points)

    def reversed(self):
        ''' Reverse point order '''
        # rev = list(self.points.__reversed__()) #do not change the original points

        rev = self.points[::-1]  # do not change the original points
        [p_rev.reverse_handles() for p_rev in rev]  # ? we are not using deep copy, so this affects original splie too...
        self.__rev_sewings(rev)
        return rev

    def __set_blender_spl_p_count(self, obj,  spl_id):
        blender_spline = obj.data.splines[spl_id]
        target_point_count = self.length
        new_points = target_point_count - len(blender_spline.bezier_points)
        if new_points > 0:  # orig spline has not enough
            blender_spline.bezier_points.add(count=new_points)
        elif new_points < 0:  # orig spline have too many points
            spline_remove_point(obj, blender_spline, [i for i in range(abs(new_points))])

    def write_to_blender_spl(self, obj, spl_id, old_pattern=False):
        if spl_id > len(obj.data.splines)-1:
            sp = obj.data.splines.new("BEZIER")
            sp.use_cyclic_u = self.cyclic
            spl_id = len(obj.data.splines) - 1
        self.__set_blender_spl_p_count(obj, spl_id)
        blender_spline = obj.data.splines[spl_id]
        for i, p in enumerate(self.points):
            p.copy_to_target(blender_spline.bezier_points[i])

        old_sewings = [(p.sewing.spl_id, p.sewing.seg_id) for p in self.points]
        change_dirs = [p.sewing.change_dir for p in self.points]  # do we flip sewing.flip ?
        new_sewings = [(spl_id, i) for i in range(self.length)]
        old_new_li = [(old, new + (flip,)) for old, new, flip in zip(old_sewings, new_sewings, change_dirs)if (old != new or old_pattern)]
        # old_new_map = dict(zip(old_sewings, new_sewings))
        old_new_map = dict(old_new_li)
        if old_pattern:  # we created new object obj.  So remap sewings from old_pattern to new_obj
            remap_sewing_to_new_obj(bpy.context, old_pattern, old_new_map, new_obj=obj)
        else:
            remap_sewing(bpy.context, obj, old_new_map)

         #* remap sewings in current spl pts
        for i,p in enumerate(self.points): #basically init with spl_id, seg_id = [0,1,2...]
            p.sewing.spl_id = spl_id
            p.sewing.seg_id = i



class BezierSplines(object):
    def __init__(self, curve_obj):
        self.splines = [BezierSpline.from_obj_spline(curve_obj, spl_id) for spl_id in range(len(curve_obj.data.splines))]
        self.curve_obj = curve_obj
        self.dimensions = curve_obj.data.dimensions

    @property
    def length(self):
        return len(self.splines)

    def __delitem__(self, idx):
        del self.splines[idx]

    def __getitem__(self, idx):  # overload [] get
        return self.splines[idx]

    def __setitem__(self, idx, value):
        ''' point can be blender bezier, or BezierPoint '''
        self.splines[idx] = value

    def append(self, spl):
        self.splines.append(spl)

    def write_to_obj(self, obj = None):
        write_obj = obj if obj else self.curve_obj
        for spl_id, spl in enumerate(self.splines):
            spl.write_to_blender_spl(write_obj, spl_id)

    def get_closest_ctrl_point(self, search_pos):
        closets_dist = 999
        best_pt = None
        best_co = None
        for spl_id, spl in enumerate(self.splines):
            for p_id, b_point in enumerate(spl.points):
                dist = (search_pos-b_point.co).length_squared
                if dist < closets_dist:
                    best_pt = (spl_id, p_id)
                    closets_dist = dist
                    best_co = b_point.co
        return best_pt, sqrt(closets_dist), best_co

    def closest_spl_point(self, search_pos):
        closets_dist = 999
        best = (None, None, None,None)
        for spl_id, spl in enumerate(self.splines):
            best_point, best_t = spl.closest(search_pos) #t is accumulated
            dist = (search_pos.xy - best_point).length
            if dist < closets_dist:
                best = (spl_id, best_point.to_3d(), best_t, dist)
                closets_dist = dist
        return best

    def closest_spl_point_view_space(self, context, event, max_dist = 20, world_space = False):
        ''' mostly for picking 3d curves under cursor... based on distance to spline outline'''
        best = (None,None,None,None)
        mouse_2d = event.mouse_region_x, event.mouse_region_y
        mat = np.array(self.curve_obj.matrix_world, 'f') if world_space else None
        for spl_id, spl in enumerate(self.splines):
            best_point_2d, best_ts = spl.closest_point_view_2d(context, mouse_2d, matrix=mat)  # t is accumulated
            if best_point_2d:
                dist_2d = (Vector(mouse_2d) - best_point_2d).length
                if dist_2d < max_dist:
                    best_point_3d_l = spl.evaluate_at(best_ts)  #? t for flattened curve should be same for original 3d curve?
                    best = (spl_id, best_point_3d_l, best_ts, dist_2d)
                    max_dist = dist_2d
        return best


    def sort_pts_along_line(self, pts, l_a, l_b):
        ''' sort pts in direction of line AB. Return [spl_id,t] list '''
        pts_len = [((pt-l_a).length_squared, spl_id, t) for t, pt, spl_id in pts]  # bigger the value the more forwared the finger is
        sorted_pts = [[spl_id, t] for pt, spl_id, t in sorted(pts_len, key=lambda ele: ele[0], reverse=False)]
        return sorted_pts

    def is_point_inside_pattern(self, pt):
        is_inside_cnt = 0
        for spl in self.splines:
            is_inside_cnt += spl.is_point_inside_winding(pt)
        return is_inside_cnt % 2 != 0

    def splines_line_intersect(self, line_a, line_b):
        ''' return newly created intersection points '''
        pts_with_ids = []
        sorted_intersections = []
        for spl_id, spl in enumerate(self.splines):
            inters = spl.spline_line_intersection(line_a, line_b) # return (t, pt) list
            pts_with_ids.extend([(t, pt, spl_id) for t,pt in inters]) #intersection pts with spl_id

        sorted_cut_pts = self.sort_pts_along_line(pts_with_ids, line_a.to_2d(), line_b.to_2d())
        new_pts_ids = []
        #* perform actuall cut using split_seg_t (adds pt at time t)
        for i,(spl_id, t_n) in enumerate(sorted_cut_pts):
            new_pt_id = self.splines[spl_id].split_seg_t(t_n)

            #* offest existing output points to right, if new cut was before them
            for k,added_pt in enumerate(new_pts_ids):
                if added_pt.spl_id == spl_id and added_pt.p_id >= new_pt_id:
                    new_pts_ids[k].p_id += 1
            # print(f'Slice intersect pt: {spl_id}:{new_pt_id}')
            new_pts_ids.append(CutPointInfo(spl_id=spl_id, p_id=new_pt_id, placement='BORDER', click_pos=self.splines[spl_id].points[new_pt_id].co))

            #* Offset any remaining cuts on current spli_id +=1
            cut_seg_id, t_cut = divmod(t_n, 1)
            cut_seg_id = int(cut_seg_id)
            for j,(remaining_spl_id, remaining_t_n) in enumerate(sorted_cut_pts[i+1:]):
                j += i+1
                if remaining_spl_id == spl_id:
                    remain_seg_id, remain_t = divmod(remaining_t_n, 1)
                    remain_seg_id = int(remain_seg_id)
                    if remain_seg_id > cut_seg_id:  # offset cuts after current segment +=1
                        sorted_cut_pts[j][1] += 1 # eval_time += 1 == jump next seg
                    if remain_seg_id == cut_seg_id:  # if cuttings same segment
                        if remain_t > t_cut: # after cut_t
                            sorted_cut_pts[j][1] = (remain_t - t_cut)/(1 - t_cut) + cut_seg_id + 1  # jump +1 seg, and scale remain_t
                        else: #next cut before t_cut
                            sorted_cut_pts[j][1] = remain_t/t_cut + cut_seg_id  # jump +1 seg, and scale remain_t


        spl_ids = {spl_id for (spl_id, t_total) in sorted_cut_pts}
        for spl_id in spl_ids:
            self.splines[spl_id].write_to_blender_spl(self.curve_obj, spl_id)

        # intersections.append({'spl_id':spl_id, 'ts':ts})
        return new_pts_ids

    def generate_splines_pts_cloud(self, res, world_space = False):
        ''' generate HQ resampled points for each spline.
        Used for: kd find closest segment, bgl drawing, ... caching low res spl pts. Uses endpoints!
        '''
        if world_space:
            return [spl.eval_resampled(res, matrix=self.curve_obj.matrix_world, uniform=True) for spl in self.splines]  # 3d
        else:
            return [spl.eval_resampled(res, uniform=True) for spl in self.splines]  # 3d or 2d, force out_dim?

        # return [spl.resample_pts(res, mat=...) for spl in self.splines]


