# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import mathutils
from mathutils import Vector, Matrix
from copy import deepcopy
from ..utils.helper_functions import *
from ..utils.curve_helpers import *
from .spline_base import *

class GTOOL_OT_SeparatePatternBySelection(bpy.types.Operator):
    bl_idname = "curve.separate_pattern_sel"
    bl_label = "Separate Pattern"
    bl_description = "Separate selected pattern points, into new object (preserves sewings)"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    @staticmethod
    def separate_selection(source_obj):
        ''' separate selected b_poiints int new obj'''
        new_obj_splines = [] #splines for new obj
        remaining_splines = [] #remaining spl
        b_spl = BezierSplines(source_obj)
        for spl_id, spl in enumerate(b_spl.splines): # we use BezierSplines() cls
            separate_pts = []
            remaining_pts = []
            p_count = len(spl.points)
            prev_point = spl.get_prev_point(0) # [-1] if cyclic or None
            #deepcopy to unlink sewings.
            for p_id, point in enumerate(spl.points):
                if point.select_control_point:
                    separate_pts.append(deepcopy(point))
                    if prev_point and not prev_point.select_control_point: #fix sliced corner verts handlers
                        remaining_pts.append(deepcopy(point))  # add first selected point, to remaining_pts pts.
                        remaining_pts[-1].handle_right_type = 'VECTOR'  #adj to cut edge
                        remaining_pts[-1].handle_left_type = 'FREE'  # fixes eg 'AUTOMATIC' aligning 45deg to corner
                        remaining_pts[-1].sewing.reset() #clear slice edge sewings

                        separate_pts[-1].handle_left_type = 'VECTOR' #adj to cut edge
                        separate_pts[-1].handle_right_type = 'FREE' #fixes eg 'AUTOMATIC' aligning 45deg to corner
                else:
                    if prev_point and prev_point.select_control_point:  # fix sliced corner verts handlers
                        remaining_pts.append(deepcopy(prev_point))  # add last selected point (i-1), to remaining_pts pts.
                        remaining_pts[-1].handle_left_type = 'VECTOR'  #adj to cut edge
                        remaining_pts[-1].handle_right_type = 'FREE' #fixes eg 'AUTOMATIC' aligning 45deg to corner

                        separate_pts[-1].handle_right_type = 'VECTOR'  # adj to cut edge
                        separate_pts[-1].handle_left_type = 'FREE'  # fixes eg 'AUTOMATIC' aligning 45deg to corner
                        separate_pts[-1].sewing.reset()  # clear slice edge sewings

                    remaining_pts.append(deepcopy(point))
                prev_point = point
            if separate_pts:
                new_obj_splines.append(BezierSpline(separate_pts, cyclic=spl.cyclic, space=spl.space))
            if remaining_pts:
                remaining_splines.append(BezierSpline(remaining_pts, cyclic=spl.cyclic, space=spl.space))

        curveData = bpy.data.curves.new(source_obj.data.name+'split', 'CURVE') #? will auto adds .001?
        curveData.dimensions = source_obj.data.dimensions
        curveData.fill_mode = source_obj.data.fill_mode

        split_obj = bpy.data.objects.new(source_obj.name + 'split', curveData)
        link_to_same_collection(source_obj, split_obj)
        split_obj.select_set(True)
        for i, b_spl in enumerate(new_obj_splines):
            b_spl.write_to_blender_spl(split_obj, i, source_obj)
        split_obj.matrix_world = source_obj.matrix_world

        #write not selected points back to active bezier obj
        for i, b_spl in enumerate(remaining_splines):
            b_spl.write_to_blender_spl(source_obj, i)
        while len(source_obj.data.splines) > len(remaining_splines):  # remove splines data that may have been left after split.
            source_obj.data.splines.remove(source_obj.data.splines[-1])

        return split_obj


    def execute(self, context):
        self.separate_selection(context.active_object)
        return {"FINISHED"}

class GTOOL_OT_JoinSewingPatterns(bpy.types.Operator):
    bl_idname = "curve.join_patterns"
    bl_label = "Join patterns"
    bl_description = "Join patterns (preserves sewings)"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def invoke(self, context, event):
        for g_id in range(len(context.scene.cloth_garment_data)):
            bpy.ops.garment.cleanup(garment_index=g_id)
        return self.execute(context)

    def execute(self, context):
        active_obj = context.active_object
        if active_obj.type != 'CURVE':
            self.report({'WARNING'}, 'Select at least two sewing patterns (curves). Cancelling!')
            return {'CANCELLED'}
        sel_objs = [obj for obj in context.selected_objects if obj.type == 'CURVE' and obj != active_obj]
        if not sel_objs:
            self.report({'WARNING'}, 'Select at least two sewing patterns (curves). Cancelling!')
            return {'CANCELLED'}

        active_mw_inv = active_obj.matrix_world.inverted()
        for obj_for_merging in sel_objs:
            for spl_id, spl in enumerate(obj_for_merging.data.splines):
                new_spline = BezierSpline()
                for i,p in enumerate(spl.bezier_points):
                    new_point = BezierPoint.from_bezier_point(p, (spl_id, i, False))
                    new_point.co = active_mw_inv @ obj_for_merging.matrix_world @ p.co
                    new_point.co.z = 0
                    new_point.handle_left = active_mw_inv @ obj_for_merging.matrix_world @ p.handle_left
                    new_point.handle_left.z = 0
                    new_point.handle_right = active_mw_inv @ obj_for_merging.matrix_world @ p.handle_right
                    new_point.handle_right.z = 0
                    new_spline.append(new_point)
                new_spline.write_to_blender_spl(active_obj, len(active_obj.data.splines), obj_for_merging)

        for obj_for_merging in reversed(sel_objs):
            bpy.data.objects.remove(obj_for_merging)
        return {"FINISHED"}


class GTOOL_OT_CurveCutHalf(bpy.types.Operator):
    bl_label = "Cut pattern in half"
    bl_idname = "curve.cut_half"
    bl_description = "Split sewing patter in two. Split if performed by center X axis.\
        \nSpline has to have two points on center of X axis, to be cut correctly."
    bl_options = {"REGISTER", "UNDO"}

    use_positive_x: bpy.props.BoolProperty(name='Use Positive X', default=True)
    merge_distance: bpy.props.FloatProperty(name='Merge distance', description="Snap points to OX axis, if they are within distance", default=0.001, min=0, max=0.1, subtype='DISTANCE')
    instance: bpy.props.BoolProperty(name='Instance', default=False)
    cache_sewings = []
    first_run = True

    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    def execute(self, context):
        bpy.ops.curve.slice_on_axis(merge_distance=self.merge_distance)
        if self.first_run:
            self.first_run = False
            self.cache_sewings = backup_sewings(context)  # cos on redo operator dosent restore custom props
        else:
            load_backup_sewings(context, self.cache_sewings)
        sel_obj = context.active_object
        max_y_cut_threshold = 999999.0

        def get_on_axis_pts():
            on_axis_points_ids = []
            for spl_id, current_spl in enumerate(sel_obj.data.splines):
                on_axis_points_ids.extend([(spl_id, p_i) for p_i, p in enumerate(current_spl.bezier_points) if abs(p.co.x) <= 0.001 and p.co.y < max_y_cut_threshold])
            on_axis_sorted_pts = sorted(on_axis_points_ids, key=lambda pt: sel_obj.data.splines[pt[0]].bezier_points[pt[1]].co.y)  # sort by Y.
            return on_axis_sorted_pts

        on_axis_sorted_pts = get_on_axis_pts()
        if len(on_axis_sorted_pts) % 2 != 0:
            return {'CANCELLED'}

        while_loops = 0
        def deselect_all():
            for spl in sel_obj.data.splines:
                for p in spl.bezier_points:
                    p.select_control_point = False

        while(on_axis_sorted_pts):
            deselect_all()  # select two on axis points
            pt1 = on_axis_sorted_pts.pop()
            pt2 = on_axis_sorted_pts.pop()
            sel_obj.data.splines[pt1[0]].bezier_points[pt1[1]].select_control_point = True
            sel_obj.data.splines[pt2[0]].bezier_points[pt2[1]].select_control_point = True
            max_y_cut_threshold = sel_obj.data.splines[pt2[0]].bezier_points[pt2[1]].co.y - 0.00001  # cut off points higher that new threshold

            bpy.ops.curve.connect_points()
            on_axis_sorted_pts = get_on_axis_pts()  # after cut, we will have new splines. Redo whole thing again

            while_loops += 1
            if while_loops > 10:  # backup, so we wont stuck in loop
                print('Maximum loop count exceeded.')
                break

        right_pts = []
        flip = 1 if self.use_positive_x else -1
        deselect_all()
        for spl_id, current_spl in enumerate(sel_obj.data.splines):
            (avg_co, std_dev) = get_avg_pos_deviation(current_spl.bezier_points)  # get center of patters / hole
            if flip * avg_co.x > 0:
                for p in current_spl.bezier_points:
                    p.select_control_point = True
        # bpy.ops.curve.separate_pattern_sel()
        new_obj = GTOOL_OT_SeparatePatternBySelection.separate_selection(sel_obj)
        if self.instance:
            for spl_id, sel_obj_spline in enumerate(sel_obj.data.splines):
                if spl_id < len(new_obj.data.splines):  # ? maybe instenad of matching by spl_id, use matching by spline get_avg_pos_deviation data? Except 0
                    clone_sp = BezierSpline.from_obj_spline(new_obj, spl_id)  # spline from clone
                    oring_sp = BezierSpline.from_obj_spline(sel_obj, spl_id)  # spline from original obj (instanced obj)
                    new_obj_spl = new_obj.data.splines[spl_id]
                    sel_obj_spl = sel_obj.data.splines[spl_id]
                    clone_pts_count = len(new_obj_spl.bezier_points)  # ! make sure spl_id < len (splines)
                    i = 0
                    while (i < clone_pts_count):  # rotate copy obj, spline, till first points are close
                        if (clone_sp.points[0].co.reflect(Vector((1, 0, 0))) - sel_obj_spline.bezier_points[0].co).length < 0.001:  # -1 cos copy is mirrored for now
                            if clone_sp.get_signed_area() * oring_sp.get_signed_area() > 0:  # make sure they both have same direction
                                clone_sp.reverse()
                            old_sewings = [(p.sewing.spl_id, p.sewing.seg_id) for p in clone_sp.points]  # use rotated new_obj points sewings
                            new_sewings = [(spl_id, p_i+1) for p_i in range(len(sel_obj_spl.bezier_points))]  # map those old copy sewings, to instance sewings
                            old_new_li = [(old, new) for old, new in zip(old_sewings, new_sewings)if (old != new)]  # cos flipping dir changes order??
                            old_new_map = dict(old_new_li)
                            remap_sewing(bpy.context, new_obj, old_new_map)
                            break
                        else:
                            clone_sp.rotate(-1)
                            i += 1

            new_obj.data = sel_obj.data
            matForFlip = Matrix.Identity(4)
            matForFlip[0][0] = -matForFlip[0][0]  # flix scale in x -1 to get mirror (local space)
            new_obj.matrix_world = matForFlip @ new_obj.matrix_world
        new_obj.name = sel_obj.name + '.r'
        sel_obj.name = sel_obj.name + '.l'
        return {'FINISHED'}


class GTOOL_OT_CurveSymmetrize(bpy.types.Operator):
    bl_label = "Symmetrize"
    bl_idname = "curve.curve_symmetrize"
    bl_description = "shift + click = use destructive algorithm (break sewings, but works better on complex geometry)."
    bl_options = {"REGISTER", "UNDO"}

    use_positive_x: bpy.props.BoolProperty(name='Use Positive X', default=True)
    merge_distance: bpy.props.FloatProperty(name='Merge distance', description="Snap points to OX axis, if they are within distance", default=0.001, min=0, max=0.1, subtype='DISTANCE')
    cache_sewings = []
    first_run = True

    def find_two_on_axis_pair_ids(self, active_spline):
        kd = mathutils.kdtree.KDTree(len(active_spline.bezier_points))  # for marging adding
        for i, point in enumerate(active_spline.bezier_points):
            kd.insert(Vector((point.co.x, 0, 0)), i)
        kd.balance()
        # return points close to center.
        return [index for (co, index, dist) in kd.find_n(Vector((0, 0, 0)), 2) if dist < 0.001]  # ! make distance dependent on garment size maybe?

    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    shift_clicked = False
    def invoke(self, context, event):
        if event.shift:
            self.shift_clicked = True
        return self.execute(context)

    def mirror_spline(self, context, sel_obj):
        flip = 1 if self.use_positive_x else -1
        mirrored_splines_ids = []
        spl_mirror_pair = find_curve_pairs(sel_obj.data.splines)
        for spl_id, current_spl in enumerate(sel_obj.data.splines):
            current_spl = sel_obj.data.splines[spl_id] #fix bug it seems....
            if spl_id in mirrored_splines_ids:
                continue  # spline was already handled, go to next

            # get two selected points (point_id, point)
            on_axis_pair_ids = self.find_two_on_axis_pair_ids(current_spl)
            b_spl = BezierSpline.from_obj_spline(sel_obj, spl_id)
            if len(on_axis_pair_ids) == 2:  # symmetrize if spl has two points on center
                spl_ab, spl_ba = b_spl.split_by_two_ids(on_axis_pair_ids)
                #which spline after cut is the right (left) one?
                (avg_co_ab, _) = get_avg_pos_deviation(spl_ab.points)
                (avg_co_ba, _) = get_avg_pos_deviation(spl_ba.points)
                if (avg_co_ab.x > avg_co_ba.x and self.use_positive_x) or (avg_co_ab.x < avg_co_ba.x and not self.use_positive_x):
                    pass
                else:
                    back = spl_ba
                    spl_ba = spl_ab
                    spl_ab = back
                mirrored = spl_ab.mirrored()
                mirrored.reverse()
                for mir, removed_half in zip(mirrored, spl_ba):  # copy sewings from removed part, to symmetrized points later
                    mir.sewing = removed_half.sewing
                del mirrored[0], mirrored[-1]  # ! maybe move up?

                # handle center points
                spl_ab[0].mirror_handles(use_right=True)  # really? - guess due to split_by_two_ids
                spl_ab[-1].mirror_handles(use_right=False)  # really?
                # is_cw = spl_ab.get_signed_area() > 0
                # if (flip > 0): # damm, this is ok, if spline is clockwise... (XXX: may not work on CCW splines? )
                #     spl_ab[0].mirror_handles(use_right=True)
                #     spl_ab[-1].mirror_handles(use_right=False)
                # else:
                #     spl_ab[0].mirror_handles(use_right=not is_cw)
                #     spl_ab[-1].mirror_handles(use_right= is_cw)

                if mirrored.length:
                    spl_ab.insert(spl_ab.length, mirrored)
                spl_ab.write_to_blender_spl(sel_obj, spl_id=spl_id)
                mirrored_splines_ids.append(spl_id)

            elif all([flip*p.co.x > 0 for p in current_spl.bezier_points]):  # mirror if all spl point are on one side
                mirrored = b_spl.mirrored()
                if spl_mirror_pair[spl_id]:  # if we get pair on other side, write to it
                    mirrored_spl_id = spl_mirror_pair[spl_id]
                    mirrored.reset_sewings()  # we do not want mirror sewing
                    target_spline = BezierSpline.from_obj_spline(sel_obj, mirrored_spl_id)
                    for p, target_p in zip(mirrored.points, target_spline.points):  # copy  sewings from left spline
                        target_p.sewing = p.sewing
                    mirrored.write_to_blender_spl(sel_obj, spl_id=mirrored_spl_id)
                    mirrored_splines_ids.append(mirrored_spl_id)
                else:  # else wirte to new spline
                    new_spl_id = len(sel_obj.data.splines)
                    mirrored.write_to_blender_spl(sel_obj, spl_id=new_spl_id)
                    mirrored_splines_ids.append(new_spl_id)

            else:
                #? should I mabye remove this spline???
                print(f'Spline nr {spl_id} is not on center (does not have 2 points on X axis) and not on side. Cant mirror. Skipping')
        for spl_id, current_spl in enumerate(sel_obj.data.splines):  # remove splines that were on bad side
            if all([flip*p.co.x < 0 for p in current_spl.bezier_points]) and spl_id not in mirrored_splines_ids:
                sel_obj.data.splines.remove(sel_obj.data.splines[spl_id])
                sewing_rem_spline_id(context, sel_obj, spl_id)

    def execute(self, context):
        if self.shift_clicked: #use different algorightm
            bpy.ops.curve.curve_symmetrize_adv(use_positive_x = self.use_positive_x)
            return {'FINISHED'}

        selected_objs = [o for o in context.selected_objects if o.type == 'CURVE']
        bpy.ops.curve.slice_on_axis(merge_distance=self.merge_distance)
        if self.first_run:
            self.first_run = False
            self.cache_sewings = backup_sewings(context)  # cos on redo operator dosent restore custom props
        else:
            load_backup_sewings(context, self.cache_sewings)
        for sel_obj in selected_objs:
            self.mirror_spline(context, sel_obj)

        return {'FINISHED'}

class GTOOL_OT_NewSymm(bpy.types.Operator):
    '''use destructive algorithm (break sewings, but works better on complex geometry)'''
    bl_label = "Destructive symmetrize"
    bl_idname = "curve.curve_symmetrize_adv"
    bl_description = "Split sewing patter in two. Split if performed by center X axis.\
        \nSimilar to normal symmetrize but does not preserves sewings."
    bl_options = {"REGISTER", "UNDO"}

    merge_distance: bpy.props.FloatProperty(name='Merge distance', description="Snap points to OX axis, if they are within distance", default=0.001, min=0, max=0.1, subtype='DISTANCE')
    use_positive_x: bpy.props.BoolProperty(name='Use Positive X', default=True)
    cache_sewings = []
    first_run = True

    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    @staticmethod
    def get_on_axis_pts(sp, first_run = True):
        ''' return top most and bottom points that are on x Axis '''
        on_axis_points_ids = [p_id for p_id, p in enumerate(sp.points) if abs(p.co.x) <= 0.001]
        if on_axis_points_ids and len(on_axis_points_ids)%2 == 0:
            sorted_pts = sorted(on_axis_points_ids, key=lambda p_id: sp.points[p_id].co.y)  # sort by Y.
            if first_run: #get most extreme points
                return (sorted_pts[0], sorted_pts[-1])
            else: # ignore extremes. Take second best
                if len(sorted_pts) >= 4:
                    return (sorted_pts[1], sorted_pts[-2])
                else:
                    return None
        return None

    @staticmethod
    def find_last_two_on_axis(sp):
        on_axis_points_ids = [p_id for p_id, p in enumerate(sp.points) if abs(p.co.x) <= 0.001]
        if on_axis_points_ids and len(on_axis_points_ids)%2 == 0:
            return (on_axis_points_ids[1], on_axis_points_ids[2])

    def is_proper_side(self, current_spl):
        is_proper_side = [p.co.x > -0.001 for p in current_spl.points] if self.use_positive_x else [p.co.x < 0.001 for p in current_spl.points]
        return all(is_proper_side)

    def execute(self, context):
        bpy.ops.curve.slice_on_axis(merge_distance=self.merge_distance)
        if self.first_run:
            self.first_run = False
            self.cache_sewings = backup_sewings(context) #cos on redo operator dosent restore custom props
        else:
            load_backup_sewings(context, self.cache_sewings)
        sel_obj = context.active_object
        flip = 1 if self.use_positive_x else -1

        for_symmetrization_splines = [] #splines that should be symmetrized
        for_mirroring_splines = [] #splines that should be symmetrized
        mirrored_splines_ids = []
        spl_pairs = find_curve_pairs(sel_obj.data.splines)
        for spl_id, spl in enumerate(sel_obj.data.splines):
            def slice_repeat(current_spl, depth = 0):
                if depth > 10: #backup, so we wont stuck in loop
                    print('max recursion exceeded')
                    return

                first = True if depth == 0 else False
                center_pts = self.get_on_axis_pts(current_spl, first)
                if center_pts is None: #cant slice current spline?
                    if self.is_proper_side(current_spl):
                        for_symmetrization_splines.append(current_spl)
                    return

                #? will this work always???? Seems so
                if set([center_pts[0],center_pts[1]]) == set((0, current_spl.length - 1)):
                    r1,r2 = self.find_last_two_on_axis(current_spl)
                    a = BezierSpline(current_spl.points[:r1+1])
                    if self.is_proper_side(a):
                        for_symmetrization_splines.append(a)
                    b = BezierSpline(current_spl.points[r2:])
                    if self.is_proper_side(b):
                        for_symmetrization_splines.append(b)
                    return

                a, b = current_spl.split_by_two_ids((center_pts[0], center_pts[1]))
                #? check if current_spl has at least 1 pt?
                if a.length > 2:
                    slice_repeat(a, depth+1)
                if b.length > 2:
                    slice_repeat(b, depth+1)

            spline = BezierSpline.from_obj_spline(sel_obj, spl_id)
            if all([flip*p.co.x > 0.0 for p in spline.points]):  # curve is on right, no symmetrize, just mirror...
                # if spl_pairs[spl_id]:  # if we get pair on other side, write to it
                    # mirrored_splines_ids.append(spl_pairs[spl_id])
                for_mirroring_splines.append(spline)

            else:
                slice_repeat(spline)
        #symmetrization of splines
        i = 0
        for i, b_spl in enumerate(for_symmetrization_splines):
            mirrored = b_spl.mirrored()
            mirrored.reverse()
            del mirrored[0], mirrored[-1]  # ! maybe move up?
            mirrored.reset_sewings()
            if flip > 0:
                b_spl[0].mirror_handles(use_right=True)
                b_spl[-1].mirror_handles(use_right=False)
            else:
                b_spl[0].mirror_handles(use_right=True)
                b_spl[-1].mirror_handles(use_right=False)
            b_spl.insert(b_spl.length, mirrored)
            b_spl.write_to_blender_spl(sel_obj, spl_id=i)

        i += 1
        #mirroring of splines
        for j, b_spl in enumerate(for_mirroring_splines):
            b_spl.write_to_blender_spl(sel_obj, spl_id=i+2*j)
            mirrored = b_spl.mirrored()
            mirrored.reset_sewings()
            mirrored.write_to_blender_spl(sel_obj, spl_id=i+2*j+1)

        return {'FINISHED'}


class GTOOL_OT_PatternFlip(bpy.types.Operator):
    bl_idname = "curve.pattern_flip"
    bl_label = "Flip Pattern"
    bl_description = "Duplicate and mirrors selected curve pattern.\
        \nIf used second time, previously mirroed copy will be updated.\
        \nUse F6 to access additional options"
    bl_options = {"REGISTER", 'UNDO'}

    use_3dCursor: bpy.props.BoolProperty(name="Use 3d Cursor", default=False)
    clone: bpy.props.BoolProperty(name="Create duplicate", default=True)
    flip_axis: bpy.props.EnumProperty(name="Flip Axis", description="", default="X",
                                       items=[("X", "X", ""),
                                              ("Y", "Y", ""),
                                              ("Z", "Z", "")])

    @classmethod
    def poll(cls, context):
        return context.mode == 'OBJECT'


    def execute(self, context):
        for obj in context.selected_objects:
            clone = None
            if self.clone:
                if is_name_r_l(obj.name):
                    mirrored_name = get_mirrored_name(obj.name)
                    if mirrored_name in context.scene.objects.keys():
                        clone = bpy.data.objects[mirrored_name]
                    else:
                        clone = obj.copy()
                        link_to_same_collection(obj, clone)
                        clone.name = mirrored_name
                else:
                    obj.name = obj.name + '.r'
                    clone = obj.copy()
                    link_to_same_collection(obj, clone)
                    clone.name = obj.name[:-2] + '.l'

            matForFlip = Matrix.Identity(4)
            objMat = deepcopy(obj.matrix_world)
            i = 0
            if self.flip_axis == 'X':
                i = 0
            if self.flip_axis == 'Y':
                i = 1
            if self.flip_axis == 'Z':
                i = 2
            matForFlip[i][i] = -matForFlip[i][i]  # flix scale in x -1 to get mirror (local space)

            if self.use_3dCursor:
                cursorLoc = context.scene.cursor.location
                matForFlip[i][3] = cursorLoc[i]  # flip loc
                objMat[i][3] = objMat[i][3] - cursorLoc[i]  # if 3d cursor flip
            if clone is not None:
                clone.matrix_world = matForFlip@objMat  # mirror in local space @ to obj space
                clone.select_set(False)
            else:
                obj.matrix_world = matForFlip@objMat  # mirror in local space @ to obj space
        return {"FINISHED"}


class GTOOL_OT_WeldVerts(bpy.types.Operator):
    bl_label = "Stitch Points"
    bl_idname = "curve.weld_points"
    bl_description = "Stitch two curve points"
    bl_options = {"REGISTER", "UNDO"}

    weld_position: bpy.props.EnumProperty(name='Weld position', description='Weld Position',
        items=[
            ('FIRST', 'First', ''),
            ('SECONOD', 'Second', ''),
            ('CENTER', 'Center', '')
        ], default='FIRST')

    def self_find_points_pairs(splines, selected_bpoints):
        p_count = len(selected_bpoints)
        kd = mathutils.kdtree.KDTree(p_count)
        for i, bpoint in enumerate(selected_bpoints):
            kd.insert(bpoint.co, i)
        kd.balance()
        (avg_co, std_dev) = get_avg_pos_deviation(selected_bpoints)
        # Find the closest point to the center
        found_doubles = []
        for i,p in enumerate(selected_bpoints):
            co, index, dist = kd.find(p)
            if index and dist < std_dev/20: # use 1/20 of std_dev to find doubles
                pair = set(i, index)
                if pair not in found_doubles:
                    found_doubles.append(pair)
        return found_doubles

    def execute(self, context):
        selObj = context.active_object
        weld_pts = []
        selected_points_per_spl = {}
        total_sel_count = 0
        for spl_id, spl in enumerate(selObj.data.splines):
            weld_pts.extend([BezierPoint.from_bezier_point(p, (spl_id, p_id, False)) for (p_id, p) in enumerate(spl.bezier_points) if p.select_control_point])
            selected_points_per_spl[spl_id] =  [p_id for (p_id, p) in enumerate(spl.bezier_points) if p.select_control_point]
        if len(weld_pts) != 2:
            self.report({'WARNING'}, 'Select two points for welding! Cancelling')
            return {'CANCELLED'}
        #?: add welding for n sequential points on single spline.... No need, we can just manualy delete points, almost the same
        if weld_pts[0].sewing.spl_id == weld_pts[1].sewing.spl_id:  # seems we selected 2 points on same spline
            spl_id = weld_pts[0].sewing.spl_id
            spline_point_count = len(selObj.data.splines[spl_id].bezier_points)
            spline = BezierSpline.from_obj_spline(selObj, spl_id)
            a, b = spline.split_by_two_ids((weld_pts[0].sewing.seg_id, weld_pts[1].sewing.seg_id))
            center = (a[-1].co + a[0].co)/2

            def write_spl(x, write_id):
                x[0].co = center + 0.001*x[0].co
                #? x[0].handle_right_type = 'FREE' #! may be broken on aligned handles..
                x[0].handle_left = x[-1].handle_left
                x[0].handle_left_type = x[-1].handle_left_type
                del x[-1]
                x.write_to_blender_spl(selObj, spl_id=write_id)

            #! what if we rewrite old spl ids, then second empty_mw may override new sewing (treating first new override sewings as old sew)
            #Sollution make sure to write to last spline first. Then nothing will be overriden
            if a.length > 3 and b.length > 3: #pick order - first write to new last[-1] spline
                write_a_id = spl_id if a.length > b.length else len(selObj.data.splines)
                write_b_id = spl_id if a.length <= b.length else len(selObj.data.splines)
                if write_a_id > write_b_id: #first write to last (new) spline
                    write_spl(a, write_a_id)
                    write_spl(b, write_b_id)
                else:
                    write_spl(b, write_b_id)
                    write_spl(a, write_a_id)
            else: #write just one spl
                if a.length > 3:
                    write_spl(a, spl_id)
                if b.length > 3:
                    write_spl(b, spl_id)

        #weld two holes..
        elif weld_pts[0].sewing.spl_id != weld_pts[1].sewing.spl_id:  # seems we selected 2 points on 2 different splines
            spl_id1 = weld_pts[0].sewing.spl_id
            spl_id2 = weld_pts[1].sewing.spl_id

            spline1 = BezierSpline.from_obj_spline(selObj, spl_id1)

            center = (weld_pts[0].co + weld_pts[1].co)/2
            spline1[weld_pts[0].sewing.seg_id].co = center
            spline1[weld_pts[0].sewing.seg_id].handle_left_type = 'VECTOR'
            spline1[weld_pts[0].sewing.seg_id].handle_right_type = 'VECTOR'

            spline2 = BezierSpline.from_obj_spline(selObj, spl_id2)
            spline2[weld_pts[1].sewing.seg_id].co = center
            spline2[weld_pts[1].sewing.seg_id].handle_left_type = 'VECTOR'
            spline2[weld_pts[1].sewing.seg_id].handle_right_type = 'VECTOR'

            #for outline with hole #!old ? remove...?
            flip = 1 if weld_pts[0].sewing.spl_id == 0 or weld_pts[1].sewing.spl_id == 0 else -1

            #deal with - outline + hole, vs (hole - hole or patch-patch)
            # if weld_pts[0].sewing.spl_id == 0 or  weld_pts[1].sewing.spl_id == 0:
            (avg_co_1, dev1) = get_avg_pos_deviation(spline1.points)
            (avg_co_2, dev2) = get_avg_pos_deviation(spline2.points)
            if (avg_co_2-avg_co_1).length < dev1: #they are inside each other so outline-hole case
                flip = -1 #flip if two
                # print('Outline hole case. Flip')
            else:
                flip = 1
                # print('hole-hole or patch-patch case. NO Flip')

            if flip*spline1.get_signed_area() * spline2.get_signed_area() > 0:
                spline2.rotate(-1*weld_pts[1].sewing.seg_id - 1)
                spline1.insert(weld_pts[0].sewing.seg_id+1, spline2)
            else:
                spline2.rotate(-1*weld_pts[1].sewing.seg_id)
                spline1.insert(weld_pts[0].sewing.seg_id+1, spline2.reversed())

            spline1.write_to_blender_spl(selObj, spl_id1)
            selObj.data.splines.remove(selObj.data.splines[spl_id2])
            sewing_rem_spline_id(context, selObj, spl_id2, remap_removed_spline=False)
        return {'FINISHED'}


class GTOOL_OT_ConnectVertspair(bpy.types.Operator):
    bl_label = "Connect points"
    bl_idname = "curve.connect_points"
    bl_description = "Connect selected points"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        selObj = context.active_object
        spl_slice = []
        for spl_id, spl in enumerate(selObj.data.splines):
            spl_slice.extend([{'spl_id': spl_id, 'p_id': p_id} for p_id, p in enumerate(spl.bezier_points) if p.select_control_point])
        if len(spl_slice)!=2:
            self.report({'WARNING'}, 'Select only two points! Cancelling')
            return {'CANCELLED'}

        # 'SPLIT_SLICE' - when we cut outline in half
        if spl_slice[0]['spl_id'] == spl_slice[1]['spl_id'] :  # we split outline in 2 parts it seems !what about holes ?
            spl_id = spl_slice[0]['spl_id']
            outline_cut_points = (spl_slice[0]['p_id'], spl_slice[1]['p_id'] )
            outline_spl = BezierSpline.from_obj_spline(selObj, spl_id)
            spl_ab, spl_ba = outline_spl.split_by_two_ids(outline_cut_points)

            spl_ab.points[-1].sewing.spl_id = -1
            spl_ba.points[-1].sewing.spl_id = -1

            spl_ba[0].handle_left_type = 'VECTOR'
            spl_ba[0].handle_right_type = 'FREE'
            spl_ba[-1].handle_right_type = 'VECTOR'
            spl_ba[-1].handle_left_type = 'FREE'

            spl_ab[0].handle_left_type = 'VECTOR'
            spl_ab[0].handle_right_type = 'FREE'
            spl_ab[-1].handle_right_type = 'VECTOR'
            spl_ab[-1].handle_left_type = 'FREE'

            #! first write to new spline (-1), so that second write_to_blender_spl wont override first wtbs sewings
            spl_ba.write_to_blender_spl(selObj, spl_id=len(selObj.data.splines)) #will crate new one, cos id = len
            spl_ab.write_to_blender_spl(selObj, spl_id=spl_id)

        # slice_type = 'CONNECT_SLICE' - connect 2 holes
        elif spl_slice[0]['spl_id'] != spl_slice[1]['spl_id']: #we connect 2 holes in 1 it seems
            hole1_spl_idx, hole1_cut_idx = (spl_slice[0]['spl_id'], spl_slice[0]['p_id'])
            hole2_spl_idx, hole2_cut_idx = (spl_slice[1]['spl_id'], spl_slice[1]['p_id'])

            hole1_spl = BezierSpline.from_obj_spline(selObj, hole1_spl_idx)
            #* add pt to hole 1
            if len(hole1_spl.points) > 1:
                hole1_spl.insert(hole1_cut_idx, [hole1_spl[hole1_cut_idx]]) # we split cut point on hole_bezier spline on 2
                hole1_spl[hole1_cut_idx].sewing.spl_id = -1  # unlink one of split points #? maybe bridge this...
                hole1_spl[hole1_cut_idx].handle_left_type = 'FREE'
                hole1_spl[hole1_cut_idx].handle_right_type = 'VECTOR'
                hole1_spl[hole1_cut_idx+1].handle_right_type = 'FREE'
                hole1_spl[hole1_cut_idx+1].handle_left_type = 'VECTOR'
            else:
                hole1_spl[hole1_cut_idx].handle_left_type = 'VECTOR'
                hole1_spl[hole1_cut_idx].handle_right_type = 'VECTOR'

            #* add pt to hole 2
            hole2_spl = BezierSpline.from_obj_spline(selObj, hole2_spl_idx)
            if len(hole2_spl.points) > 1:
                hole2_spl.insert(hole2_cut_idx, [hole2_spl[hole2_cut_idx]])  # we split cut point on hole_bezier spline on 2
                hole2_spl.rotate(-1*hole2_cut_idx-1) #move cut to begining of slice, minus one for vert clone

                hole2_spl[0].handle_right_type = 'FREE'
                hole2_spl[0].handle_left_type = 'VECTOR'
                hole2_spl[-1].handle_left_type = 'FREE'
                hole2_spl[-1].handle_right_type = 'VECTOR'
            else:
                hole2_spl[0].handle_right_type = 'VECTOR'
                hole2_spl[0].handle_left_type = 'VECTOR'

            #* merge separate splines.
            flip = -1 if hole1_spl.is_point_inside_winding(hole2_spl.get_center()) or hole2_spl.is_point_inside_winding(hole1_spl.get_center()) else 1
            if flip*hole1_spl.get_signed_area() * hole2_spl.get_signed_area() > 0:  # for cutting hole in outline, both splines have to have opposite direction (sig_area)
                hole2_spl[0].sewing.spl_id = -1  # unlink one of split points
                hole1_spl.insert(hole1_cut_idx+1, hole2_spl)  # in outline split place we put the hole spline, +1 to offset vert clone
            else:
                hole2_spl[-1].sewing.spl_id = -1  # unlink one of split points
                hole1_spl.insert(hole1_cut_idx+1, hole2_spl.reversed())  # in outline split place we put the hole spline

            hole1_spl.write_to_blender_spl(selObj, spl_id=min(hole1_spl_idx, hole2_spl_idx))
            selObj.data.splines.remove(selObj.data.splines[max(hole1_spl_idx, hole2_spl_idx)])
            sewing_rem_spline_id(context, selObj, max(hole1_spl_idx, hole2_spl_idx), remap_removed_spline=False)

        return {'FINISHED'}

class GTOOL_OT_DeletePoint(bpy.types.Operator):
    bl_label = "Delete Vert GT"
    bl_idname = "curve.curve_del_vert_gt"
    bl_description = "Delete Vert from GT - prevents moving sewings after removing curve points"
    bl_options = {"REGISTER", "UNDO"}
    ''' delete sewing left to selected points, substract segment_idx for each removed point for sewing right to it '''

    #Running this from F3 always works cos init properly reads the garment sewing
    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    def execute(self, context):
        selObj = context.active_object
        splines_for_removal = []
        for spl_id, spl in enumerate(selObj.data.splines):
            selected_points = [p for p in spl.bezier_points if p.select_control_point]
            if len(selected_points) == len(spl.bezier_points):
                splines_for_removal.append(spl_id)
                continue
        for spl_rem_id in reversed(splines_for_removal): #removing all points, so removing current spline
            selObj.data.splines.remove(selObj.data.splines[spl_rem_id])
            sewing_rem_spline_id(context, selObj, spl_rem_id)

        for spl_id, spl in enumerate(selObj.data.splines):
            removed_seg_ids = [(spl_id, i) for (i, p) in enumerate(spl.bezier_points) if p.select_control_point]
            if removed_seg_ids:
                remove_sewings_by_sewid(context, selObj, removed_seg_ids)
            not_selected_pts = BezierSpline() #points that were not removed
            not_selected_pts.points = [BezierPoint.from_bezier_point(p, (spl_id, i, False)) for (i, p) in enumerate(spl.bezier_points) if not p.select_control_point]
            if not_selected_pts.length != len(spl.bezier_points):
                not_selected_pts.write_to_blender_spl(selObj, spl_id=spl_id)

        return {'FINISHED'}


class GTOOL_OT_SubdivideSegment(bpy.types.Operator):
    bl_label = "Subdivide Segments GT"
    bl_idname = "curve.curve_subdivide_gt"
    bl_description = "Subdivide segments while preserving garment sewing from moving"
    bl_options = {"REGISTER", "UNDO"}

    number_cuts : bpy.props.IntProperty(name='Number of cuts', default=1, min=1, soft_max=6)
    factor: bpy.props.FloatProperty(name='Factor', description='', default= 0.5, min=0, max=1, subtype='FACTOR')

    cache_sewings = []
    first_run = True
    #Running this from F3 always works cos init properly reads the garment sewing
    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'number_cuts')
        if self.number_cuts == 1:
            layout.prop(self, 'factor')

    def execute(self, context):
        if self.first_run:
            self.first_run = False
            self.cache_sewings = backup_sewings(context)
        else:
            load_backup_sewings(context, self.cache_sewings)
        if self.factor==0 or self.factor == 1 and self.number_cuts == 1:
            return {'FINISHED'}
        selObj = context.active_object
        b_spl = BezierSplines(selObj)
        for spl_id, spl in enumerate(selObj.data.splines):
            s_len = len(spl.bezier_points)
            prev_point = spl.bezier_points[-1] if spl.use_cyclic_u else None
            seg_to_subdivie = []
            for p_id, point in enumerate(spl.bezier_points):
                if point.select_control_point is True and prev_point and prev_point.select_control_point:
                    seg = p_id-1 if p_id > 0 else s_len-1
                    seg_to_subdivie.append(seg)
                prev_point = point

            seg_offset = 0
            for seg_subd in seg_to_subdivie:
                t_seg = seg_subd + seg_offset  # -1 cos we subdivide prev_point / current_pt
                tn_split = [self.factor] if self.number_cuts < 2 else [(self.number_cuts-n) + 1/(n+1)
                                                                                for n in range(self.number_cuts, 0, -1)]
                for t in tn_split:
                    b_spl[spl_id].split_seg_t(t_seg + t, spatial_space=True)  # uniform in euclid space
                seg_offset += self.number_cuts
            if seg_to_subdivie:
                b_spl[spl_id].write_to_blender_spl(selObj, spl_id)
        return {'FINISHED'}



class GTOOL_OT_SplitByLine(bpy.types.Operator):
    bl_label = "Curve line cut GT"
    bl_idname = "curve.slice_on_axis"
    bl_description = "Cut 2d curve, along OX plane, by adding new points on curve center line"
    bl_options = {"REGISTER", "UNDO"}

    cache_sewings = []
    first_run = True
    merge_distance: bpy.props.FloatProperty(name='Merge distance', description="Snap points to OX axis, if they are within distance", default=0.001, min=0, max=0.1, subtype='DISTANCE')

    #Running this from F3 always works cos init properly reads the garment sewing
    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    @staticmethod
    def slice_curve_plane(spline, plane_co, plane_no, margin):
        intersection_pts = {}
        splines_segs_resampled = spline.eval_resampled(20, uniform=True, out_dim=3)  # need local space
        # margin = 0.004  # this is as factor... - use meters instead - (dist to ox axis)
        for seg_id, seg_resampled in enumerate(splines_segs_resampled):
            sub_segment_len = len(seg_resampled)
            for sub_id in range(sub_segment_len-1):  #  segments.cnt to pts.cnt-1
                sub_point = seg_resampled[sub_id]
                next_sub_point = seg_resampled[sub_id+1]
                #intersection projected tuple (on first line, on second line)
                intersections_line_plane = mathutils.geometry.intersect_line_plane(sub_point, next_sub_point, plane_co, plane_no)  # this is for segments :::
                touching_left = 1 if sub_id == 0 else 0  # does sub_segment touch spline point on left
                touching_right = 1 if sub_id + 1 == (sub_segment_len-1) else 0  # does sub_segment touch spline point on right
                if intersections_line_plane:
                    inter_pt, inter_fac = mathutils.geometry.intersect_point_line(intersections_line_plane, sub_point, next_sub_point) # as factor from first pt
                    # inter_fac within (0+d, 1-d) range ?
                    # convert from factor to absolute distance using - plane normal as x axis,
                    # if (0+margin*touching_left) < inter_fac <= (1-margin*touching_right):  # factor - if on line then it is (0, 1)
                    # hit_dist_x_axis <= margin):  # factor - if on line then it is (0, 1)
                    if (0 < inter_fac <= 1):   # factor  is (0, 1) - segment is intersecting
                        pt_hit_dist_x_axis = (sub_point - plane_co).project(plane_no).length
                        next_pt_hit_dist_x_axis = (next_sub_point - plane_co).project(plane_no).length
                        if touching_left and pt_hit_dist_x_axis < margin:  # only snap  seg_id for first sub segment pts
                            print(f'snapping first vert id: {seg_id} to axis')
                            spline.points[seg_id].co = inter_pt
                        elif touching_right and next_pt_hit_dist_x_axis < margin:  # only snap  seg_id for first sub segment pts
                            print(f'snapping second vert id: {(seg_id+1) % spline.length} to axis')
                            spline.points[(seg_id+1) % len(splines_segs_resampled)].co = inter_pt
                        else: # not touching axis, but intersecting
                            if seg_id not in intersection_pts.keys():
                                intersection_pts[seg_id] = []
                            intersection_pts[seg_id].append(intersections_line_plane)
                    # else:  # check if we can snap point to cut line...???
                    #     if touching_left and -1*margin < hit_dist_x_axis < margin:  # only snap  seg_id for first sub segment pts
                    #         print(f'snapping first vert id: {seg_id} to axis')
                    #         spline.points[seg_id].co = inter_pt
                    #     if touching_right and 1 - margin < hit_dist_x_axis < 1 + margin:  # only snap  seg_id for last sub segment pts
                    #         print(f'snapping second vert id: {(seg_id+1) % spline.length} to axis')
                    #         spline.points[(seg_id+1) % len(splines_segs_resampled)].co = inter_pt
        return intersection_pts


    def execute(self, context):
        if self.first_run: # XXX: does it do anything? - does not seems to keep sewings if slice pts are added
            self.first_run = False
            self.cache_sewings = backup_sewings(context)
        else:
            load_backup_sewings(context, self.cache_sewings) # how did it work...

        sel_obj = context.active_object
        for spl_id, active_spline in enumerate(sel_obj.data.splines):
            b_spl = BezierSpline.from_obj_spline(sel_obj, spl_id)
            # min_x, max_x, min_y, max_y = spline_bbox(active_spline)
            # line = (Vector((0, max_y + 0.01, 0)), Vector((0, min_y - 0.01 , 0)))
            intersection_pts = self.slice_curve_plane(b_spl, Vector((0,0,0)), Vector((1,0,0)), self.merge_distance)
            offset = 0 #each new cut point will be put further into already extended line
            for seg_id  in intersection_pts.keys():
                for point_cut in intersection_pts[seg_id]:
                    ins_p = BezierPoint(point_cut)
                    b_spl.insert(seg_id+offset+1, [ins_p])
                    offset += 1
            b_spl.write_to_blender_spl(sel_obj, spl_id)
        return {'FINISHED'}



class GTOOL_OT_DuplicateSewed(bpy.types.Operator):
    bl_label = "Duplicate Sewed"
    bl_idname = "curve.duplicate_sewed"
    bl_description = "Duplicate pattern and sew it to original pattern (or copy sewings from original object)"
    bl_options = {"REGISTER", "UNDO"}

    instance: bpy.props.BoolProperty(name='Instance', description = 'Generate pattern instance vs copy', default=True)
    sewing_source: bpy.props.EnumProperty(name='Sewing source', description='Sewing source',
        items=[
            ('SEWINGS', 'Copy garment sewings', 'Copy garment sewings from source pattern to clone'),
            ('OBJECT', 'Stitch to source obj', 'Stitch clone to source obj')
        ], default='OBJECT')

    @classmethod
    def poll(cls, context):
        return (context.active_object and context.active_object.type == 'CURVE')

    def execute(self, context):
        selObj = context.active_object
        selObj.select_set(False)

        sel_obj_instance = selObj.copy()
        if not self.instance:
            sel_obj_instance.data = selObj.data.copy()
        link_to_same_collection(selObj, sel_obj_instance)
        sel_obj_instance.select_set(True)
        sel_obj_instance.matrix_world = sel_obj_instance.matrix_world
        if self.sewing_source == 'OBJECT':
            for spl_idx, active_spline in enumerate(selObj.data.splines):
                spline_len = len(active_spline.bezier_points)
                for garment in context.scene.cloth_garment_data:
                    existing_patterns = [g_mesh.pattern_obj.name for g_mesh in garment.sewing_patterns if g_mesh.pattern_obj]
                    if selObj.name in existing_patterns:  # find garments  that are using active_obj
                        for i in range(spline_len):
                            garment.garment_sewings.add()
                            garment.garment_sewings[-1].source_obj = selObj
                            garment.garment_sewings[-1].target_obj = sel_obj_instance
                            garment.garment_sewings[-1].from_spline_idx = spl_idx
                            garment.garment_sewings[-1].from_segment_idx = i
                            garment.garment_sewings[-1].to_spline_idx = spl_idx
                            garment.garment_sewings[-1].to_segment_idx = i

                    garment.sewing_patterns.add()
                    garment.sewing_patterns[-1].pattern_obj = sel_obj_instance
        else:
            for garment in context.scene.cloth_garment_data:
                existing_patterns = [p.pattern_obj for p in garment.sewing_patterns if p.pattern_obj]
                old_sewings = []
                if selObj in existing_patterns:  # find garments  that are using active_obj
                    for idx, sewing in enumerate(garment.garment_sewings):
                        if sewing.source_obj == selObj or sewing.target_obj == selObj:
                            old_sewings.append(sewing)
                garment.sewing_patterns.add()
                garment.sewing_patterns[-1].pattern_obj = sel_obj_instance

                for old_sew in old_sewings: #old sewing where source or target is in sew
                    new_sew = garment.garment_sewings.add()
                    new_sew.source_obj = old_sew.source_obj
                    new_sew.target_obj = old_sew.target_obj
                    new_sew.flip = old_sew.flip
                    new_sew.from_segment_idx = old_sew.from_segment_idx
                    new_sew.from_spline_idx = old_sew.from_spline_idx
                    new_sew.to_segment_idx = old_sew.to_segment_idx
                    new_sew.to_spline_idx = old_sew.to_spline_idx
                    if old_sew.source_obj == selObj:
                        new_sew.source_obj = sel_obj_instance
                    if old_sew.target_obj == selObj:
                        new_sew.target_obj = sel_obj_instance


        return {'FINISHED'}

