'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from math import ceil, floor, radians
from statistics import median
from .utils.helper_functions import find_create_collection, frange, get_scale_mat, import_node_group

import bmesh
import bpy
import mathutils
import numpy as np
import re
from mathutils import Matrix, Vector
from mathutils.bvhtree import BVHTree
from .spline_tools.spline_base import *
from .utils.curve_helpers import get_predicted_vert_count
from .utils.helper_functions import set_weights_uniform, set_vert_vec_attrib_uniform, set_vert_float_attrib_uniform
from .edit_cloth_preset import assign_cloth_preset_weights_to_all_patterns
from .mesh_cloth_functions import init_simulation
from os import path

bend_deform_node_group = 'BendDeform_GN'
bend_mod_name = 'Bend Deform GT'
BEND_DEFORM_GT_FILE = path.dirname(__file__)+"/bend_deform_gt.blend"

def get_node_group(node_name):
    node_group = bpy.data.node_groups.get(node_name)
    if not node_group:
        import_node_group(node_name, lib_file=BEND_DEFORM_GT_FILE)    # XXX: does it actually works (node.name vs node.node_tree.name)
        node_group = bpy.data.node_groups.get(node_name)
        node_group.use_fake_user = True
    return node_group


def kd_intersect(kd_search, verts_pair, cross_edges, search_range):
    close_inner_bound_edges = kd_search.find_range((verts_pair[0].co+verts_pair[1].co)/2, search_range)
    for (e_center, e_id, dist) in close_inner_bound_edges:
        if check_line_intersects_outline(verts_pair, [cross_edges[e_id]]):
            return True
    return False


def check_line_intersects_outline(verts_pair, cross_edges):
    for brd_ed_idx, cross_edge in enumerate(cross_edges):
        if verts_pair[0] in cross_edge.verts[:] or verts_pair[1] in cross_edge.verts[:]:  # ignore if line ends intersect
            continue
        if mathutils.geometry.intersect_line_line_2d(verts_pair[0].co, verts_pair[1].co, cross_edge.verts[0].co, cross_edge.verts[1].co):  # return vector or None
            return True
    return False


def do_tris_from_pointcloud(spl_outline_points, point_cloud, point_spacing, output_topology, silhouette_bvh, is_cyclic_spline):
    # XXX: point_cloud = point_cloud  + pocket_verts_co  # ver with pockets in past?
    bm = bmesh.new()
    border_verts = []
    border_edges = []  # for excluding edges that cross with outline in check_line_intersects_outline
    outline_edge_loops = []

    vert_boundary_id = bm.verts.layers.int.new('is_boundary_v')  # ?
    vert_boundary_id = bm.verts.layers.int['is_boundary_v']

    # * create verts for outline points, and connect them with edges
    vert_id = 0
    for outline_points, is_cyclic in zip(spl_outline_points, is_cyclic_spline):
        outline_edge_loops.append([])
        prev_vert = None
        for i, point_outline in enumerate(outline_points):
            new_vert = bm.verts.new(point_outline)
            new_vert.index = vert_id
            new_vert[vert_boundary_id] = 1
            if i == 0:
                first_vert = new_vert
            border_verts.append(new_vert)

            if prev_vert:
                new_border_edge = bm.edges.new([prev_vert, new_vert])
                border_edges.append(new_border_edge)  # border vert points N is member of border_edges [N-1] and [N]
                outline_edge_loops[-1].append(new_border_edge)
            prev_vert = new_vert
            vert_id += 1
        bm.verts.ensure_lookup_table()
        if is_cyclic:
            last_border_edge = bm.edges.new([first_vert, new_vert])
            border_edges.append(last_border_edge)
            outline_edge_loops[-1].append(last_border_edge)
        prev_vert = None

    # * creave point_cloud verts
    point_cloud_count = len(point_cloud)
    if point_cloud_count > 0:
        kd_point_cloud_verts = mathutils.kdtree.KDTree(point_cloud_count)  # for marging adding
        point_cloud_first_vert = vert_id
        p_cloud_verts = []
        for point in point_cloud:
            new_vert = bm.verts.new(point)
            new_vert.index = vert_id
            p_cloud_verts.append(new_vert)
            kd_point_cloud_verts.insert(point, vert_id)
            vert_id += 1
        kd_point_cloud_verts.balance()

        bm.verts.ensure_lookup_table()

        kd_border_edges = mathutils.kdtree.KDTree(len(border_edges))  # for marging adding
        for e_id, edge in enumerate(border_edges):
            kd_border_edges.insert((edge.verts[0].co + edge.verts[1].co)/2, e_id)
        kd_border_edges.balance()

        # * conntect point_cloud verts in range 1.2*point_spacing
        inner_edges = []
        inner_edge_keys = set()
        cloud_range = 1.2 * point_spacing
        for p_vert in p_cloud_verts:
            for (co, index, dist) in kd_point_cloud_verts.find_range(p_vert.co, cloud_range):
                if (index, p_vert.index) not in inner_edge_keys and dist > 0.0001:
                    # but check if new edge wont cross boundary verts...
                    if not kd_intersect(kd_border_edges, [p_vert, bm.verts[index]], border_edges, search_range=point_spacing):
                        inner_edge_keys.add((p_vert.index, index))

        for (v1, v2) in inner_edge_keys:
            inner_edges.append(bm.edges.new([bm.verts[v1], bm.verts[v2]]))


        if output_topology == 'TRIS': #connect protuding square verts
            inner_boudnary_verts = [v for v in p_cloud_verts if len(v.link_edges) != 6]  # not the outer boundary, but pts cloud boundary
            inner_boudnary_edges = [e for e in inner_edges if len(e.verts[0].link_edges) != 6 and len(e.verts[1].link_edges) != 6]  # not the outer boundary, but pts cloud boundary
        else:
            inner_boudnary_verts = [v for v in p_cloud_verts if len(v.link_edges) != 4]  # not the outer boundary, but pts cloud boundary
            inner_boudnary_edges = [e for e in inner_edges if len(e.verts[0].link_edges) != 4 and len(e.verts[1].link_edges) != 4]  # not the outer boundary, but pts cloud boundary

        # * kd for inner_boudnary_verts
        kd_inner_boundary_verts = mathutils.kdtree.KDTree(len(inner_boudnary_verts))  # for marging adding
        for vert in inner_boudnary_verts:
            kd_inner_boundary_verts.insert(vert.co, vert.index)
        kd_inner_boundary_verts.balance()

        if output_topology == 'QUAD': #connect protuding corners of quad point cloud
            tris_edges_keys = set() # create tris by merging corners
            corner_inner_boudnary_verts = [v for v in inner_boudnary_verts if len(v.link_edges) == 2]  # not the outer boundary, but pts cloud boundary
            max_dist = point_spacing * 1.42 #basically search across diagonal
            for corner_vert in corner_inner_boudnary_verts:
                v1_link_verts = [e.other_vert(corner_vert) for e in corner_vert.link_edges]
                for (co, index, dist) in kd_inner_boundary_verts.find_range(corner_vert.co, max_dist):  # we are mostly intereseted in diagonal verts . Square diagonal = 1.41 * edge_len
                    v2 = bm.verts[index]
                    if (index, corner_vert.index) not in tris_edges_keys and dist > point_spacing * 1.40:
                        v2_link_verts = [e.other_vert(v2) for e in v2.link_edges]
                        if len((set(v1_link_verts).intersection(v2_link_verts))) < 2: #? not remember why
                            if not kd_intersect(kd_border_edges, [corner_vert, bm.verts[index]], border_edges, search_range=point_spacing):
                                tris_edges_keys.add((corner_vert.index, index))

            for (v1, v2) in tris_edges_keys:
                new_triangle_edge = bm.edges.new([bm.verts[v1], bm.verts[v2]])
                inner_boudnary_edges.append(new_triangle_edge) #add it to quad inner_boudnary_edges edges list

        # * kd for inner_boudnary_edges
        kd_inner_boundary_edges = mathutils.kdtree.KDTree(len(inner_boudnary_edges))  # for marging adding
        for e_id, edge in enumerate(inner_boudnary_edges):
            kd_inner_boundary_edges.insert((edge.verts[0].co + edge.verts[1].co)/2, e_id)
        kd_inner_boundary_edges.balance()

        # ? Bridge pointcloud with border verts start
        # * bridge boundary verts with inner border verts
        bridge_edges = []
        edges_search_rad = 1.6 * point_spacing #if output_topology == 'TRIS' else 1.2 * point_spacing # quad diagonal is 1.41
        bridge_verts_search_rad = 1.6 * point_spacing if output_topology == 'TRIS' else 1.5 * point_spacing
        for brd_idx, border_vert in enumerate(border_verts):
            for (co, index, dist) in kd_inner_boundary_verts.find_range(border_vert.co, bridge_verts_search_rad):  # Previouly only added one, but now we do intersect_previous_border_edges check
                # *check outer border edges intersection
                if kd_intersect(kd_border_edges, [border_vert, bm.verts[index]], border_edges, search_range=edges_search_rad):
                    continue

                # ? change to kd search? with kd_inserts?
                intersect_previous_bridge_edges = check_line_intersects_outline([border_vert, bm.verts[index]], bridge_edges)
                if intersect_previous_bridge_edges:
                    continue

                # *check inner border edges intersection
                if kd_intersect(kd_inner_boundary_edges, [border_vert, bm.verts[index]], inner_boudnary_edges, search_range=edges_search_rad):
                    continue

                new_bride_edge = bm.edges.new([border_vert, bm.verts[index]])
                bridge_edges.append(new_bride_edge)
                # break

        # * finally fill edges with faces
        hole_faces = []
        for outline_n, is_cyclic in zip(outline_edge_loops, is_cyclic_spline): #prevents corrupted fill of mesh - fill holes only \ pts_cloud
            if is_cyclic:
                ret = bmesh.ops.contextual_create(bm, geom=outline_n)
                if ret['faces']:
                    hole_faces.extend(ret['faces'])
        bmesh.ops.contextual_create(bm, geom=bm.edges) #fill remaining geo (pts clould + bridged outline)
        for f in hole_faces:
            bm.faces.remove(f)

        if output_topology == 'TRIS':
            bmesh.ops.join_triangles(bm, faces=bm.faces, cmp_seam=False, cmp_sharp=False, cmp_uvs=False, cmp_vcols=False, cmp_materials=False, angle_face_threshold=3.1, angle_shape_threshold=3.1)
        bmesh.ops.triangulate(bm, faces=bm.faces, quad_method='BEAUTY', ngon_method='BEAUTY')
        if output_topology == 'QUAD': #somehow gives nice results after tirangulate then -> join_tris
            bmesh.ops.join_triangles(bm, faces=bm.faces, cmp_seam=False, cmp_sharp=False, cmp_uvs=False, cmp_vcols=False, cmp_materials=False, angle_face_threshold=3.1, angle_shape_threshold=3.1)

    else:
        bmesh.ops.contextual_create(bm, geom=bm.edges)
        bmesh.ops.triangulate(bm, faces=bm.faces, quad_method='BEAUTY', ngon_method='BEAUTY')  # here becaouse above line breaks triangulation
        if output_topology == 'QUAD':
            bmesh.ops.join_triangles(bm, faces=bm.faces, cmp_seam=False, cmp_sharp=False, cmp_uvs=False, cmp_vcols=False, cmp_materials=False, angle_face_threshold=3.1, angle_shape_threshold=3.1)

    mesh = bpy.data.meshes.new(name="Cloth")
    bm.to_mesh(mesh)
    bm.free()
    del bm
    return mesh


def poke_new_verts(target_mesh, pocket_verts_co_projected, seg_len):
    ''' Add extra verts from projected pocket outline or Pins, to sewing pattern '''
    if len(pocket_verts_co_projected) == 0:
        return
    bm = bmesh.new()   # create an empty BMesh
    bm.from_mesh(target_mesh)   # fill it in from a Mesh
    pattern_BVHTree = BVHTree.FromBMesh(bm, epsilon=0.0)

    # Modify the BMesh, can do anything here...
    bm.faces.ensure_lookup_table()
    faces_to_split = {}  # faces to perform poke on
    hit_faces = []
    pocket_hit_verts = []  # verts from pocket_ outline that hit target mesh

    # get faces that are intersecting with pocket outline verts
    for i, pocket_vert_co in enumerate(pocket_verts_co_projected):
        location, normal, index, dist = pattern_BVHTree.ray_cast(Vector((pocket_vert_co.x, pocket_vert_co.y, 0.01)), Vector((0, 0, -200)), 201) # cast ray from Vec1 in dir Vec_dir
        if index is not None:
            if index in faces_to_split.keys():
                faces_to_split[index].append(pocket_vert_co)
            else:
                faces_to_split[index] = [pocket_vert_co]
                hit_faces.append(bm.faces[index])

    bm.faces.ensure_lookup_table()
    new_verts = []
    splited_faces = []
    for face, (face_id, face_new_verts_co) in zip(hit_faces, faces_to_split.items()):
        ver_1 = face.verts[0]
        ver_2 = face.verts[1]
        vec_1_2 = (ver_2.co - ver_1.co)  # no need to normalize
        if len(face_new_verts_co) > 1:
            face_new_verts_co.sort(key=lambda x: vec_1_2.dot(x))  # sort by dot product len
        # split face from its ver[0] to vert2 - by adding loop throught coos
        (generated_face, generated_loop) = bmesh.utils.face_split(face, vert_a=ver_1, vert_b=ver_2, coords=face_new_verts_co,  use_exist=True)
        splited_faces.append(generated_face)
        new_verts.extend([v for v in face.verts if v.index == -1])  # old face will be connected now to new verts (with idx==-1)
        # new_faces.append(generated_face_loop[0])

    bmesh.ops.triangulate(bm, faces=splited_faces, quad_method='BEAUTY', ngon_method='BEAUTY')
    # bmesh.ops.beautify_fill(bm, faces=bm.faces, edges=bm.edges, method='AREA')

    # WARNING! THis may remove doubles on close boudnary verts (eg. on narrow vedges)
    # remove doubles execept new pocket verts
    boundary_verts = [vert for vert in bm.verts if vert.is_boundary]
    found_doubles = bmesh.ops.find_doubles(bm, verts=bm.verts, keep_verts=new_verts+boundary_verts, dist=seg_len/3)
    bmesh.ops.weld_verts(bm, targetmap=found_doubles['targetmap'])

    bm.to_mesh(target_mesh)
    bm.free()


def get_lattice_transformation(align_objs):
    ''' return (center, sizex, sizey, sizez) '''
    patter_count = len(align_objs)
    if patter_count == 1:  # get local space cos we will align lattice to it anyway
        points = [point.co for spl in align_objs[0].data.splines for point in spl.bezier_points]
    else:
        points = []
        for curve in align_objs:
            points.extend([curve.matrix_world@point.co for spl in curve.data.splines for point in spl.bezier_points])

    np_points = np.array(points)
    max_x = np.max(np_points[:, 0])
    max_y = np.max(np_points[:, 1])
    max_z = np.max(np_points[:, 2])
    min_x = np.min(np_points[:, 0])
    min_y = np.min(np_points[:, 1])
    min_z = np.min(np_points[:, 2])

    # center = Vector((max_x+min_x, max_y+min_y, max_z+min_z)) * 0.5
    vec_scale = Vector((max(max_x-min_x, 0.1), max(max_y-min_y, 0.1), max(max_z-min_z, 0.1))) * 0.5

    mat_loc = Matrix.Translation(((max_x+min_x)/2, (max_y+min_y)/2, (max_z+min_z)/2))
    mat_sca = get_scale_mat(vec_scale)
    # mat_rot = align_objs[0].matrix_world.to_quaternion().to_matrix().to_4x4()

    if patter_count == 1:
        out_mat = align_objs[0].matrix_world  @ mat_loc  @ mat_sca
    else:
        out_mat = mat_loc @ mat_sca

    return out_mat


def get_set_lattice(context, garment, name, align_objs):
    lattice_name = 'Lattice_' + name
    if lattice_name in bpy.data.objects.keys():
        lattice_ob = bpy.data.objects[lattice_name]
    else:
        lattice = bpy.data.lattices.new(lattice_name)
        lattice.points_u = 3
        lattice.points_v = 3
        lattice.points_w = 3
        lattice_ob = bpy.data.objects.new(lattice_name, lattice)

    output_collection = bpy.data.collections[garment.name]
    if lattice_ob.name not in output_collection.objects.keys():
        output_collection.objects.link(lattice_ob)

    # lattice_ob.rotation_euler = align_obj.rotation_euler
    new_mat_transform = get_lattice_transformation(align_objs)
    lattice_ob.matrix_world = new_mat_transform
    return lattice_ob


class GTOOL_OT_Triangulate(bpy.types.Operator):
    bl_idname = 'garment.garment_to_mesh'
    bl_label = 'Generate Mesh'
    bl_description = 'Generate triangulated (or quad) mesh\n(if warning icon is visible, generated polycount will be high, so it may freeze Blender!)'
    bl_options = {'REGISTER'}


    def cache_cloth_vg(self, target_cloth_obj):
        self.backup_cloth_vg = {}
        if target_cloth_obj:
            for mod in target_cloth_obj.modifiers:
                if mod.type == 'CLOTH':
                    self.backup_cloth_vg['vertex_group_mass'] = mod.settings.vertex_group_mass
                    # self.backup_cloth_vg['vertex_group_structural_stiffness'] = mod.settings.vertex_group_structural_stiffness
                    # self.backup_cloth_vg['vertex_group_shear_stiffness'] = mod.settings.vertex_group_shear_stiffness
                    # self.backup_cloth_vg['vertex_group_bending'] = mod.settings.vertex_group_bending
                    # self.backup_cloth_vg['vertex_group_shrink'] = mod.settings.vertex_group_shrink
                    # self.backup_cloth_vg['vertex_group_pressure'] = mod.settings.vertex_group_pressure
                    return

    def restore_cloth_vg(self, target_cloth_obj):
        if len(self.backup_cloth_vg) == 0:
            return
        for mod in target_cloth_obj.modifiers:
            if mod.type == 'CLOTH':
                if self.backup_cloth_vg['vertex_group_mass'] in target_cloth_obj.vertex_groups.keys():
                    mod.settings.vertex_group_mass = self.backup_cloth_vg['vertex_group_mass']
                else:  # there was no any vg in cloth.vertex_group_mass
                    if 'PIN' in target_cloth_obj.vertex_groups.keys(): #try setting to pin
                        mod.settings.vertex_group_mass = 'PIN'
                    else:
                        mod.settings.vertex_group_mass = ''

                # stiff_vg = self.backup_cloth_vg.get('vertex_group_structural_stiffness')
                # if stiff_vg in target_cloth_obj.vertex_groups.keys():
                #     mod.settings.vertex_group_structural_stiffness = stiff_vg
                # else:
                #     mod.settings.vertex_group_structural_stiffness = ''
                #
                # shear_stiffness = self.backup_cloth_vg.get('vertex_group_shear_stiffness')
                # if shear_stiffness in target_cloth_obj.vertex_groups.keys():
                #     mod.settings.vertex_group_shear_stiffness = shear_stiffness
                # else:
                #     mod.settings.vertex_group_shear_stiffness = ''
                #
                # vertex_group_bending = self.backup_cloth_vg.get('vertex_group_bending')
                # if vertex_group_bending in target_cloth_obj.vertex_groups.keys():
                #     mod.settings.vertex_group_bending = vertex_group_bending
                # else:
                #     mod.settings.vertex_group_bending = ''
                #
                # vertex_group_shrink = self.backup_cloth_vg.get('vertex_group_shrink')
                # if vertex_group_shrink in target_cloth_obj.vertex_groups.keys():
                #     mod.settings.vertex_group_shrink = vertex_group_shrink
                # else:
                #     mod.settings.vertex_group_shrink = ''
                #
                # vertex_group_pressure = self.backup_cloth_vg.get('vertex_group_pressure')
                # if vertex_group_pressure in target_cloth_obj.vertex_groups.keys():
                #     mod.settings.vertex_group_pressure = vertex_group_pressure
                # else:
                #     mod.settings.vertex_group_pressure = ''

                return


    @staticmethod
    def get_segments_spacing(context, points):
        '''Return normalized points spacing len '''
        segments_points_spacing = []
        segment_len = 0
        previous_point = points[0]
        for point in points:
            point_diff_vec = Vector(point) - Vector(previous_point)
            segment_len += point_diff_vec.length
            previous_point = point
        segment_len *= context.scene.unit_settings.scale_length
        segments_points_spacing.append(segment_len/len(points))
        median_spacing = median(segments_points_spacing)  # for point spacing  use the line with longest spacing legnth
        return median_spacing

    @staticmethod
    def get_segment_proportions(context, splinesSegments, resolution):
        '''Return normalized list of segment lengths * res '''
        length_of_segments = []
        segments_points_spacing = []
        for segment in splinesSegments:  # for islands
            segment_len = 0
            previous_point = segment[0]
            for point in segment:
                point_diff_vec = Vector(point) - Vector(previous_point)
                segment_len += point_diff_vec.length
                previous_point = point
            segment_len *= context.scene.unit_settings.scale_length
            length_of_segments.append(segment_len)
            segments_points_spacing.append(segment_len/len(segment))
        max_segment_len = max(length_of_segments)
        # median_spacing = median(segments_points_spacing)  # for point spacing  use the line with longest spacing legnth
        # normalize_maxSeg_len = 1/median_spacing  # 1meter/ max_diff_len = x.
        # N = S / 1meter  * Res   # s - len of segment
        return [max(ceil(resolution * seg_len), 2) for seg_len in length_of_segments]  # /1meter  - eg 12points/meter

    def calc_normalized_seg_res(self, context, curve_obj, resolution):
        '''Return list of segments lengths normalized  '''
        splines_seg_props = []
        for spline in curve_obj.data.splines:
            uniform_segments_points = []
            points_count = len(spline.bezier_points)
            mat_scale = get_scale_mat(curve_obj.matrix_world.to_scale())
            for i in range(points_count):
                knot1 = mat_scale@spline.bezier_points[i].co
                handle1 = mat_scale@spline.bezier_points[i].handle_right
                knot2 = mat_scale@spline.bezier_points[(i+1) % points_count].co
                handle2 = mat_scale@spline.bezier_points[(i+1) % points_count].handle_left
                uniform_segments_points.append(mathutils.geometry.interpolate_bezier(knot1, handle1, handle2, knot2, resolution))
            splines_seg_props.append(self.get_segment_proportions(context, uniform_segments_points, resolution))
        return splines_seg_props

    def curve_resample(self, curve_obj):
        '''Return list of points for each 2D curve segment. Assume curve obj, has segment count dict assigned '''
        mat_scale = get_scale_mat(curve_obj.matrix_world.to_scale())
        splines_normalized_points = []
        for spl_id, spline in enumerate(self.bsplines[curve_obj.name]):
            cached_seg_res = curve_obj['normalized_seg_res'][spl_id]  # this data is generated after sewing normalization step
            spl_segments_resampled = spline.eval_resampled(list(cached_seg_res), matrix=mat_scale, uniform= True, out_dim=3) #3d
            splines_normalized_points.append(spl_segments_resampled) # in wordd space...
        return splines_normalized_points

    def process_sewing_resolution(self, context, garment):
        ''' store map k: garment name V: list or segments lengths
        calculate even for disabled sewing patterns, for simplicity :/ '''
        orig_pattern_seg_res = {pattern.pattern_obj.name: self.calc_normalized_seg_res(context, pattern.pattern_obj,  garment.cloth_res) for pattern in garment.sewing_patterns}

        garment_sewings = context.scene.cloth_garment_data[self.garment_index].garment_sewings
        calculated_sewings = [] # for main while loop - mark finished sewings
        sewing_count = len(garment_sewings)
        sewing_to_check = [i for i in range(sewing_count)]

        def get_connected_sewings(sew_idx, search_dir):
            ''' get linnked sewings connected to sew_idx, in direction defined by search_dir
                Return list of (child_sewing_id, next_direction_to_serach - 'target' or 'source') list '''
            current_sitch = garment_sewings[sew_idx]

            #*  we will search for sewings, connected to target_pattern[target_spl_idx][target_seg_idx]
            target_pattern = current_sitch.source_obj.name if search_dir == 'source' else current_sitch.target_obj.name
            target_spl_idx = current_sitch.from_spline_idx if search_dir == 'source' else current_sitch.to_spline_idx
            target_seg_idx = current_sitch.from_segment_idx if search_dir == 'source' else current_sitch.to_segment_idx

            #* searc for target_pattern
            child_sewings = []
            for sewing_id in sewing_to_check:  # search sewing that are matching (source/target.patter, to_s_id/from_s_id) of current parent sewing
                if sewing_id != sew_idx:
                    if garment_sewings[sewing_id].target_obj.name == target_pattern and garment_sewings[sewing_id].to_spline_idx == target_spl_idx and garment_sewings[sewing_id].to_segment_idx == target_seg_idx:
                        # add sewing matching, and direction that we should follow next when searching for next children
                        child_sewings.append((sewing_id, 'source'))
                    if garment_sewings[sewing_id].source_obj.name == target_pattern and garment_sewings[sewing_id].from_spline_idx == target_spl_idx and garment_sewings[sewing_id].from_segment_idx == target_seg_idx:
                        # add sewing matching, and direction that we should follow next when searching for next children
                        child_sewings.append((sewing_id, 'target'))
            #they removed from sewing_to_check, later in sewing_get_linked_res
            return child_sewings

        def sewing_get_linked_res(sew_idx, search_dir):
            ''' Return sum of segments, and amount of connected sewing to searched sewing'''
            ''' search_dir - in which direction we go from current sewing - coming from  ('source' or 'target' ) '''

            #* get current sewing target segment res
            sewing = garment_sewings[sew_idx]
            sewing_pattern = sewing.source_obj.name if search_dir == 'source' else sewing.target_obj.name
            spl_id = sewing.from_spline_idx if search_dir == 'source' else sewing.to_spline_idx
            seg_id = sewing.from_segment_idx if search_dir == 'source' else sewing.to_segment_idx
            current_sewing_seg_res = orig_pattern_seg_res[sewing_pattern][spl_id][seg_id]

            #* mark current sew_id as traversed
            if sew_idx in calculated_sewings: #allow only for master root sewing (cos we have go twice - left and right for it)
                return (0, 0) #already checked
            calculated_sewings.append(sew_idx)
            sewing_to_check.remove(sew_idx)

            child_sewings = get_connected_sewings(sew_idx, search_dir=search_dir)
            sum_children_sewing_cnt = 0 # child res
            children_cumsum_res = 0   # child reses
            for (child_sewing_id, direction_to_search) in child_sewings:
                (child_cumsum_res, sum_child_sewing_cnt) = sewing_get_linked_res(child_sewing_id, search_dir=direction_to_search)
                children_cumsum_res += child_cumsum_res
                sum_children_sewing_cnt += sum_child_sewing_cnt
            return (children_cumsum_res + current_sewing_seg_res, sum_children_sewing_cnt + 1)  # + self

        # calculate connected sttiches point counts (res)
        sewing_to_res = {}  # maps  sewing_id: avg_res
        i = 0
        while len(sewing_to_check) > 0:
            sewing = sewing_to_check[0]
            (right_chain_cumsum_res, right_chain_cnt) = sewing_get_linked_res(sewing, search_dir='source')  # get sewing in direction of sewing source
            calculated_sewings.remove(sewing) #mark as not processed again for second pass
            sewing_to_check.append(sewing)
            (left_chain_cumsum_res, left_chain_cnt) = sewing_get_linked_res(sewing, search_dir='target')  # get sewing in direction of sewing target
            aver_res = int(round((right_chain_cumsum_res + left_chain_cumsum_res)/(right_chain_cnt + left_chain_cnt)))
            for calculated_sewings in calculated_sewings:
                sewing_to_res[calculated_sewings] = aver_res
            calculated_sewings = [] #we can, since in one pass, we touch all linked sewings.
            i += 1
            if i > sewing_count: #sanity check
                break

        #first pockets. Then in can be override by finall for loop
        if garment.enable_pockets:
            for pocket in garment.pockets:  # calc pattern without normalization (no sewing )
                pocket.pattern_obj['normalized_seg_res'] = self.calc_normalized_seg_res(context, pocket.pattern_obj, garment.cloth_res)

        # then transfer segments  point count from sewing to pattern segmets counts
        for garment_mesh in garment.sewing_patterns:  # first use default/ original values  from curve_uniform_resample()
            garment_mesh.pattern_obj['normalized_seg_res'] = orig_pattern_seg_res[garment_mesh.pattern_obj.name]
        for sewing_id, avg_res in sewing_to_res.items():  # then assing res from computed sewing_to_res
            sewing = garment_sewings[sewing_id]
            sewing.source_obj['normalized_seg_res'][sewing.from_spline_idx][sewing.from_segment_idx] = avg_res
            sewing.target_obj['normalized_seg_res'][sewing.to_spline_idx][sewing.to_segment_idx] = avg_res

        # assing default values for pocket segments

    def draw(self, context):
        layout = self.layout
        row = layout.row(align=True)
        col = row.column(align=True)
        col.label(text='Generated vertex count is high', icon = 'ERROR')
        col.label(text='      It may take long time process')
        col.label(text='      Continue?')




    @staticmethod
    def set_pattern_attribs(mesh_pattern_obj, pattern_curve_obj):
        pattern_up_vec = pattern_curve_obj.matrix_world.col[2]
        facing_center_vec = pattern_curve_obj.matrix_world.translation.xy.to_3d()  # vector to center (0- transl.x, 0-trans.y, 0 - 0)
        scale_mat = get_scale_mat(pattern_curve_obj.matrix_world.to_scale())  # set size to (1,1,1)
        is_facing_center = facing_center_vec.dot(scale_mat@pattern_up_vec)  # if negative we should flip the Y vector direction (rotated 180 deg X)
        origin = pattern_curve_obj.matrix_world.col[3]
        y_dir = pattern_curve_obj.matrix_world.col[1]
        if is_facing_center > 0:
            x_dir = pattern_curve_obj.matrix_world.col[0]
        else:
            # y_dir = -1*pattern_curve_obj.matrix_world.col[1]
            x_dir = -1*pattern_curve_obj.matrix_world.col[0]

        set_vert_vec_attrib_uniform(mesh_pattern_obj, 'GT_Origin', origin)
        set_vert_vec_attrib_uniform(mesh_pattern_obj, 'GT_Y_Dir', y_dir)
        set_vert_vec_attrib_uniform(mesh_pattern_obj, 'GT_X_Dir', x_dir)
        # set_vert_float_attrib_uniform(gen_pattern_obj, 'GT_Bend_Factor', 1) # XXX: what about pockets?


    def invoke(self, context, event):
        garment_index = context.scene.garment_index
        garment = context.scene.cloth_garment_data[garment_index]
        self.garment = garment
        self.blend_units_rescale = context.scene.unit_settings.scale_length
        if get_predicted_vert_count(context, garment) > 15000:
            return context.window_manager.invoke_props_dialog(self)
        else:
            return self.execute(context)

    @staticmethod
    def assign_uvs(context, merged_obj):
        #? wont work if obj on hidden layer no matter what...
        if len(merged_obj.data.polygons) and merged_obj in context.visible_objects:
            override = {'active_object': merged_obj, 'selected_objects': [merged_obj], 'selected_editable_objects': [merged_obj], 'visible_objects': [merged_obj], 'editable_objects': [merged_obj]}
            with bpy.context.temp_override(**override):
                bpy.ops.object.editmode_toggle()
                bpy.ops.mesh.select_all()
                # bpy.ops.uv.unwrap(override, method='CONFORMAL', fill_holes=False, margin=0.1)
                bpy.ops.uv.unwrap(method='ANGLE_BASED', fill_holes=True, correct_aspect=True, margin=0.1)
                bpy.ops.mesh.select_all(action='DESELECT')
                bpy.ops.object.editmode_toggle()


    def execute(self, context):
        self.garment_index = context.scene.garment_index
        garment = self.garment
        bpy.context.scene.frame_current = 0
        self.generated_objs_names = {}  # k: curve pattern name, value: generated mesh name
        bpy.ops.garment.cleanup(garment_index = self.garment_index)

        self.enabled_patterns = [g_sewp.pattern_obj.name for g_sewp in garment.sewing_patterns if g_sewp.pattern_obj and g_sewp.is_enabled]

        # NOTE: helper bsplines  {obj_name: BezierSpline}
        self.bsplines = {g_sewp.pattern_obj.name:BezierSplines(g_sewp.pattern_obj) for g_sewp in garment.sewing_patterns if g_sewp.pattern_obj and g_sewp.is_enabled}
        if garment.enable_pockets:
            self.bsplines.update({pocket.pattern_obj.name: BezierSplines(pocket.pattern_obj) for pocket in garment.pockets if pocket.pattern_obj})
        self.process_sewing_resolution(context, garment)  #* stored in pattern_obj['normalized_seg_res']  - for pattern and pockets
        # convert the curve pattern to mesh pattern

        output_collection = find_create_collection(context, context.scene.collection, garment.name)
        sewing_patterns = [p.pattern_obj for p in garment.sewing_patterns if p.is_enabled]
        if garment.enable_pockets:
            for pocket in garment.pockets:
                if pocket.pattern_obj not in sewing_patterns:
                    gen_pattern_obj = self.create_mesh_cloth(context, pocket.pattern_obj)  # * generates pattern_obj['segments_verts']
                    self.generated_objs_names[pocket.pattern_obj.name] = gen_pattern_obj.name
                # sewing_objs_names.extend(self.create_pocket_sewing(context, pocket))
        for sewing_pattern in garment.sewing_patterns:
            if sewing_pattern.is_enabled:
                gen_pattern_obj = self.create_mesh_cloth(context, sewing_pattern.pattern_obj) # * generates pattern_obj['segments_verts']
                self.generated_objs_names[sewing_pattern.pattern_obj.name] = gen_pattern_obj.name
                set_vert_float_attrib_uniform(gen_pattern_obj, 'GT_Bend_Factor', sewing_pattern.bend_factor) # restore bend factor

        merged_objs = list(self.generated_objs_names.values())  # + sewing_objs_names+
        if not merged_objs:
            print('There are no meshes generated. Maybe wrong patterns are selected. Cancelling')
            return {'CANCELLED'}

        for obj_name in merged_objs:
            if obj_name not in output_collection.objects.keys():
                output_collection.objects.link(bpy.data.objects[obj_name])

        self.cache_cloth_vg(garment.target_cloth_object)
        if garment.merge_patterns and garment.name != '':
            #* create sewins bmedges, join, write data to garment.target_mesh_obj
            merged_obj = self.merge_objs(context, merged_objs, garment.smooth_strength, garment)
            merged_obj.scale = (1, 1, 1)
            self.setup_modifiers(context, merged_obj, garment) #bend, lattice
            if len(output_collection.objects) == 0:
                bpy.data.collections.remove(output_collection)
            for modifier in merged_obj.modifiers:
                if modifier.type == 'CLOTH':
                    override = {'scene': context.scene, 'active_object': merged_obj, 'point_cache': modifier.point_cache}
                    with bpy.context.temp_override(**override):
                        bpy.ops.ptcache.free_bake()
                    break
            self.restore_cloth_vg(garment.target_cloth_object)
            self.assign_uvs(context, merged_obj) #* finally do uv

            # write  cloth presets into patterns
            # precalculate all data required for assign_cloth_preset_to_all_patterns()
            assign_cloth_preset_weights_to_all_patterns(garment)



        self.report({'INFO'}, 'Garment triangulation finished')
        return {'FINISHED'}

    def setup_modifiers(self, context, merged_obj, garment):
        ''' add bend, lattice modifiers. Also make sure bend is before lattice, which is before cloth'''
        CLOTH_NOT_EXIST = 55
        LATTICE_NOT_EXIST = 54
        grouped_modifiers_stack = {}  # k -> V :  mod.type -> [(idx, mod.name),...]
        modifiers_names = []
        mod_len = len(merged_obj.modifiers)
        for i, mod in enumerate(merged_obj.modifiers):  # CLOTH, SIMPLE_DEFORM, LATTICE
            modifiers_names.append(mod.name)
            if mod.type not in grouped_modifiers_stack.keys():
                grouped_modifiers_stack[mod.type] = []
            grouped_modifiers_stack[mod.type].append((i, mod.name))

        first_cloth_idx = grouped_modifiers_stack['CLOTH'][0][0] if 'CLOTH' in grouped_modifiers_stack.keys() else CLOTH_NOT_EXIST
        first_lattice_idx = grouped_modifiers_stack['LATTICE'][0][0] if 'LATTICE' in grouped_modifiers_stack.keys() else LATTICE_NOT_EXIST

        # remove existing mods if hey are referencing groups, objects that no longer exits
        for mod_name in modifiers_names:
            mod = merged_obj.modifiers[mod_name]
            if hasattr(mod, 'vertex_group'):
                if mod.vertex_group and mod.vertex_group not in merged_obj.vertex_groups.keys():
                    merged_obj.modifiers.remove(mod)
                    mod = None
            if mod and hasattr(mod, 'object'):
                if mod.object and mod.object not in context.scene.objects.keys():
                    merged_obj.modifiers.remove(mod)
                    mod = None

        def check_remove_mod(name):
            '''Remove modifier if it was disabled by user'''
            mod = merged_obj.modifiers.get(name)
            if mod:
                merged_obj.modifiers.remove(mod)
            helper_obj = bpy.data.objects.get(name)
            if helper_obj and helper_obj.type in {'LATTICE', 'EMPTY'}:  # remove also empty or lattice helper
                bpy.data.objects.remove(helper_obj)


        def update_mod_position(modifier):
            nonlocal mod_len, first_cloth_idx, first_lattice_idx  # gives ability to update nonlacol var
            modifier.show_expanded = False
            new_mod_id = mod_len
            # move it befor cloth mods
            while (new_mod_id > first_cloth_idx and new_mod_id >= 1):
                bpy.ops.object.modifier_move_up(modifier=modifier.name)
                new_mod_id -= 1
            if modifier.type == 'NODES':  # move GN_Bend before lattice mods
                while (new_mod_id >= first_lattice_idx and new_mod_id >= 1):  # move it befor lattice and cloth mods
                    bpy.ops.object.modifier_move_up(modifier=modifier.name)
                    new_mod_id -= 1

            mod_len += 1
            # move up lattice and cloth mods if they exist
            first_cloth_idx = first_cloth_idx + 1 if first_cloth_idx != CLOTH_NOT_EXIST else first_cloth_idx
            first_lattice_idx = first_lattice_idx + 1 if first_lattice_idx != LATTICE_NOT_EXIST else first_lattice_idx


        def set_mod(name, align_objs):
            nonlocal mod_len, first_cloth_idx, first_lattice_idx  # gives ability to update nonlacol var
            mod_name = 'Lattice_' + name
            modifier = merged_obj.modifiers.get(mod_name)
            if not modifier:
                lattice_obj = get_set_lattice(context, garment, name=name, align_objs=align_objs)
                modifier = merged_obj.modifiers.new(name=mod_name, type='LATTICE')
                modifier.object = lattice_obj
                modifier.vertex_group = name
            update_mod_position(modifier)

        # NOTE: Always setup global bend modifier (geo nodes)
        modifier = merged_obj.modifiers.get(bend_mod_name)
        if not modifier:
            modifier = merged_obj.modifiers.new(name=bend_mod_name, type='NODES')
            bend_ng = get_node_group(bend_deform_node_group)     # or import
            modifier.node_group = bend_ng

            modifier = merged_obj.modifiers.get(bend_mod_name)
            modifier['Input_4_attribute_name'] = 'GT_Origin' # NOTE: we should have this attribute by now?
            modifier['Input_3_attribute_name'] = 'GT_Y_Dir' # NOTE: we should have this attribute by now?
            modifier['Input_6_attribute_name'] = 'GT_X_Dir' # NOTE: we should have this attribute by now?
            modifier['Input_7_attribute_name'] = 'GT_Bend_Factor'
            # XXX: damm - sometimes it works here but sometimes not

            if bpy.app.version < (4, 0, 0):
                bpy.ops.object.geometry_nodes_input_attribute_toggle(prop_path="[\"Input_4_use_attribute\"]", modifier_name=bend_mod_name)
                bpy.ops.object.geometry_nodes_input_attribute_toggle(prop_path="[\"Input_3_use_attribute\"]", modifier_name=bend_mod_name)
                bpy.ops.object.geometry_nodes_input_attribute_toggle(prop_path="[\"Input_6_use_attribute\"]", modifier_name=bend_mod_name)
                bpy.ops.object.geometry_nodes_input_attribute_toggle(prop_path="[\"Input_7_use_attribute\"]", modifier_name=bend_mod_name)
            else:
                bpy.ops.object.geometry_nodes_input_attribute_toggle(input_name="Input_4", modifier_name=bend_mod_name)
                bpy.ops.object.geometry_nodes_input_attribute_toggle(input_name="Input_3", modifier_name=bend_mod_name)
                bpy.ops.object.geometry_nodes_input_attribute_toggle(input_name="Input_6", modifier_name=bend_mod_name)
                bpy.ops.object.geometry_nodes_input_attribute_toggle(input_name="Input_7", modifier_name=bend_mod_name)

        update_mod_position(modifier)


        for sewing_pattern in garment.sewing_patterns:
            if sewing_pattern.is_enabled:
                check_remove_mod('Bend_' + sewing_pattern.pattern_obj.name) # remove old bend deform modifier  (replaced by global GNodes bend )
                if sewing_pattern.add_lattice:
                    set_mod(sewing_pattern.pattern_obj.name, align_objs=[sewing_pattern.pattern_obj])
                else:
                    check_remove_mod('Lattice_' + sewing_pattern.pattern_obj.name)

        for vert_group in garment.generated_vgroups:
            vg_patt_names = [re.sub('\.[0-9]+$', '', vg_patt_name) for vg_patt_name in vert_group.vgroups_patterns]  # remove extension .xx
            if len(vert_group.vgroups_patterns):
                check_remove_mod('Bend_'+vert_group.name) # remove old bend deform modifier  (replaced by global GNodes bend )
                if vert_group.add_lattice:
                    set_mod(vert_group.name,  align_objs=[bpy.data.objects[name] for name in vg_patt_names])
                else:
                    check_remove_mod('Lattice_'+vert_group.name)
        return

    def create_pin_sewing(self, garment):
        ''' For Pins between Mesh obj (eg body, sphere) and sewing pattern
         Return [(is_mesh_pinned, (co1, co2)), (is_mesh_pinned, .)...]'''
        sewings_co_pairs = []
        if garment.enable_pins:
            for pin in garment.pins:
                if pin.source_obj.name in self.enabled_patterns and (pin.target_obj.name in self.enabled_patterns or pin.target_obj.type == 'EMPTY'):
                    sewings_co_pairs.append((pin.source_obj.matrix_world @ Vector(pin.source_co), pin.target_obj.matrix_world @ Vector(pin.target_co)))
                    sewings_co_pairs[-1] = (pin.target_obj.type == 'EMPTY', sewings_co_pairs[-1])
        return sewings_co_pairs

    def create_sewings_list(self, garment):
        sewings_co_pairs = []
        for sewing in garment.garment_sewings:
            if sewing.source_obj.name in self.enabled_patterns and sewing.target_obj.name in self.enabled_patterns:
                obj1 = bpy.data.objects[self.generated_objs_names[sewing.source_obj.name]]
                obj2 = bpy.data.objects[self.generated_objs_names[sewing.target_obj.name]]
                sewing1_verts_ids = obj1['segments_verts'][sewing.from_spline_idx][sewing.from_segment_idx]
                sewing2_verts_ids = obj2['segments_verts'][sewing.to_spline_idx][sewing.to_segment_idx]

                vert_stich_01 = [obj1.matrix_world@obj1.data.vertices[sewing_vert].co for sewing_vert in sewing1_verts_ids]
                vert_stich_02 = [obj2.matrix_world@obj2.data.vertices[sewing_vert].co for sewing_vert in sewing2_verts_ids]
                if sewing.flip:
                    vert_stich_02 = list(reversed(vert_stich_02))
                sewings_co_pairs.extend(list(zip(vert_stich_01, vert_stich_02)))
        return sewings_co_pairs

    def create_pocket_sewing(self, garment):
        sewings_co_pairs = []
        if garment.enable_pockets:
            for pocket in garment.pockets:
                if pocket.target_pattern.name in self.enabled_patterns:
                    generated_pocket_mesh = bpy.data.objects[self.generated_objs_names[pocket.pattern_obj.name]]
                    target_pattern = pocket.target_pattern  # cos we don need 'segments_verts' from it
                    sewing_ids = [int(i) for i in pocket.pocket_sewing]
                    previous_sewing = sewing_ids[-1]
                    for sewing_id in sewing_ids:
                        sewing_count = len(generated_pocket_mesh['segments_verts'][0])
                        # if current sewing and prevous are neighbors then skip first vert\
                        pocket_verts_ids = generated_pocket_mesh['segments_verts'][0][sewing_id][1:] if \
                            (sewing_id-previous_sewing == 1) or (sewing_id-previous_sewing == -1*sewing_count+1) else \
                            generated_pocket_mesh['segments_verts'][0][sewing_id][:]
                        previous_sewing = sewing_id

                        target_mat_inv = target_pattern.matrix_world.inverted()
                        pocket_mat_inv = generated_pocket_mesh.matrix_world.inverted()

                        pocket_verts_boundary_co = [generated_pocket_mesh.matrix_world@generated_pocket_mesh.data.vertices[sewing_vert].co for sewing_vert in pocket_verts_ids]
                        target_verts_co = [target_mat_inv@generated_pocket_mesh.matrix_world@generated_pocket_mesh.data.vertices[sewing_vert].co for sewing_vert in pocket_verts_ids]
                        target_verts_co_projected = [Vector((p.x, p.y, 0)) for p in target_verts_co]
                        target_verts_co_transformed = [target_pattern.matrix_world @ p for p in target_verts_co_projected]

                        sewings_co_pairs.extend(list(zip(pocket_verts_boundary_co, target_verts_co_transformed)))

        return sewings_co_pairs

    def merge_objs(self, context, merged_objs, cloth_smooth, garment):
        sewings_co_pairs = []
        sewings_co_pairs.extend(self.create_sewings_list(garment))
        sewings_co_pairs.extend(self.create_pocket_sewing(garment))
        # sewings_co_pairs.extend(self.create_pin_sewing(garment))
        pin_co_pairs = self.create_pin_sewing(garment)  # [(is_mesh_pinned, (co1, co2)), ... ]
        sewings_co_pairs.extend([co_pair for is_mesh_pinned, co_pair in pin_co_pairs if not is_mesh_pinned]) #not mesh li
        mesh_pinned_co_pairs = [co_pair for is_mesh_pinned, co_pair in pin_co_pairs if is_mesh_pinned]

        output_collection = bpy.data.collections[garment.name]
        bpy.ops.object.select_all(action='DESELECT')
        for obj_name in merged_objs:
            obj = bpy.data.objects[obj_name]
            obj.select_set(True)
            context.view_layer.objects.active = obj
        new_cloth_obj = context.view_layer.objects.active
        new_cloth_obj.select_set(True)

        obj_list = [bpy.data.objects[name] for name in merged_objs]
        override = {'active_object': new_cloth_obj, 'selected_editable_objects': obj_list}
        with bpy.context.temp_override(**override):
            bpy.ops.object.join()
        if new_cloth_obj.data.users > 1:
            bpy.ops.object.make_single_user(object=True, obdata=True, material=False, animation=False)#DONE: make it work on instance objs
        if new_cloth_obj.matrix_world != Matrix.Identity(4):
            # bpy.ops.object.transform_apply(location=True, rotation=True, scale=True)   #XXX: gives bad result (mat world not updated..)
            new_cloth_obj.data.transform(new_cloth_obj.matrix_world)
            new_cloth_obj.matrix_world = Matrix.Identity(4)
        context.view_layer.objects.active = new_cloth_obj

        mesh = new_cloth_obj.data
        bm = bmesh.new()
        bm.from_mesh(mesh)

        bm.verts.ensure_lookup_table()
        bm.edges.ensure_lookup_table()

        kd_bm_verts = mathutils.kdtree.KDTree(len(bm.verts))  # for marging adding
        for point in bm.verts:
            kd_bm_verts.insert(point.co, point.index)
        kd_bm_verts.balance()

        edges_to_create = set()
        for sewings_co_pair in sewings_co_pairs:
            co, index1, dist = kd_bm_verts.find(sewings_co_pair[0])
            co, index2, dist = kd_bm_verts.find(sewings_co_pair[1])
            if (index2, index1) not in edges_to_create:
                edges_to_create.add((index1, index2))  # DONE: maybe check if idx2,idx1  is in edges.

        for (index1, index2) in edges_to_create:
            try:
                bm.edges.new([bm.verts[index1], bm.verts[index2]])
            except:
                print(f'Error when adding edge from verts:{index1} TO: {index2}')

        #* create verts and edges for pins, that link to edges (not sewign patterns curve objs)
        last_vert_idx = len(bm.verts)
        verts_to_pin_mesh = []
        for sewings_co_pair in mesh_pinned_co_pairs:
            mesh_liked_vert = bm.verts.new(sewings_co_pair[1])
            verts_to_pin_mesh.append(mesh_liked_vert)
            mesh_liked_vert.index = last_vert_idx
            last_vert_idx += 1

        bm.verts.ensure_lookup_table()
        for mesh_liked_vert, sewings_co_pair in zip(verts_to_pin_mesh, mesh_pinned_co_pairs):
            co, v1_id, dist = kd_bm_verts.find(sewings_co_pair[0])
            bm.edges.new([bm.verts[v1_id], mesh_liked_vert])


        #* finalize by adding vgroups etc
        bm.verts.index_update()
        boundary_verts = [vert.index for vert in bm.verts if vert.is_boundary or vert.is_wire]
        non_boundary_verts = [vert.index for vert in bm.verts if not (vert.is_boundary or vert.is_wire)]
        verts_to_pin_mesh = [v.index for v in verts_to_pin_mesh]

        bmesh.ops.recalc_face_normals(bm, faces=bm.faces)
        bm.to_mesh(mesh)
        mesh.update()
        bm.free()

        # if garment.add_border_mask:
        vg = new_cloth_obj.vertex_groups.new(name="border_mask")
        vg.add(boundary_verts, 1, "REPLACE")

        # if garment.add_non_border_mask:
        vg = new_cloth_obj.vertex_groups.new(name="non_border_mask")
        vg.add(non_boundary_verts, 1, "REPLACE")

        if verts_to_pin_mesh:
            backup_mass_vg_name = self.backup_cloth_vg.get('vertex_group_mass')
            if backup_mass_vg_name in new_cloth_obj.vertex_groups.keys():
                vg = new_cloth_obj.vertex_groups[backup_mass_vg_name]
            else:
                vg = new_cloth_obj.vertex_groups.new(name="PIN")
                self.backup_cloth_vg['vertex_group_mass'] = 'PIN'
            vg.add(verts_to_pin_mesh, 1, "ADD")


        cloth_mod = new_cloth_obj.modifiers.new('Cloth', type='CLOTH') #? do we always generate it???
        cloth_mod.settings.use_sewing_springs = True
        # cloth_mod.settings.vertex_group_shrink = "Group"
        cloth_mod.settings.sewing_force_max = 2
        cloth_mod.collision_settings.distance_min = 0.005
        if cloth_mod.settings.quality <= 10:
            cloth_mod.settings.quality = 18

        if verts_to_pin_mesh:
            cloth_mod.settings.vertex_group_mass = "PIN"

        if cloth_smooth > 0:
            override = {'active_object': new_cloth_obj, 'selected_objects': [new_cloth_obj], 'selected_editable_objects': [new_cloth_obj]}
            with bpy.context.temp_override(**override):
                bpy.ops.object.cloth_smooth(iteration_amount=cloth_smooth)
        if (garment.target_cloth_object and garment.target_cloth_object.type == 'MESH') or \
           (bpy.data.objects.get(garment.name) and bpy.data.objects[garment.name].type == 'MESH'):
            if garment.target_cloth_object:  # copy data from new generated garment mesh to previously generated garment mesh
                previous_obj = garment.target_cloth_object
            elif garment.name in bpy.data.objects.keys():
                previous_obj = bpy.data.objects[garment.name]
                garment.target_cloth_object = previous_obj
            previous_obj.vertex_groups.clear()
            for mat in previous_obj.data.materials:
                new_cloth_obj.data.materials.append(mat)
            old_mesh_name = previous_obj.data.name[:]
            previous_obj.data.name += '_old'  # previous mesh name _old
            # vgroups would are be transfered here
            # new_cloth_obj.data.transform(previous_obj.matrix_world) # we will clear prev_obj mat, for apply it to mesh or else target mesh will move
            previous_obj.data = new_cloth_obj.data
            previous_obj.matrix_world = Matrix.Identity(4)
            previous_obj.data.name = old_mesh_name

            # transfer ver groups, cos above wont do it.
            if previous_obj.name not in context.scene.objects.keys():
                output_collection.objects.link(previous_obj)
            if bpy.app.version < (3, 0):
                override = {'active_object': previous_obj, 'selected_objects': [new_cloth_obj], 'selected_editable_objects': [new_cloth_obj]}
                with bpy.context.temp_override(**override):
                    bpy.ops.object.vertex_group_copy_to_linked()
            #else done automatically in blender 3.0


            # bpy.ops.object.delete(override, use_global=False)
            bpy.data.objects.remove(new_cloth_obj)
            # enable all modes for previous_obj
            for mod in previous_obj.modifiers:
                mod.show_viewport = True
            merged_obj = previous_obj
        else:  # use merged obj as output
            new_cloth_obj.name = garment.name
            garment.target_cloth_object = new_cloth_obj
            merged_obj = new_cloth_obj
            merged_obj.name = garment.name  # cos rename above leaves .001 sometimes

        if merged_obj.name not in output_collection.objects.keys():
            output_collection.objects.link(merged_obj)
        if merged_obj.name in context.view_layer.objects.keys():
            merged_obj.select_set(True)
        context.view_layer.objects.active = merged_obj
        return merged_obj

    def get_extra_pocket_verts(self, garment, pattern_curve_obj):
        mat_scale = get_scale_mat(pattern_curve_obj.matrix_world.to_scale())  # cos we have to scale up generated mesh objs
        pocket_verts_co_projected = []
        pockets_using_pattern = [pocket for pocket in garment.pockets if pocket.target_pattern == pattern_curve_obj]
        target_mat_inv = pattern_curve_obj.matrix_world.inverted()
        for pocket in pockets_using_pattern:
            pocket_sewings_seg_ids = sorted([int(i) for i in pocket.pocket_sewing])
            pocket_resampled_segs = self.curve_resample(pocket.pattern_obj)[0]  #! pocket is using only first curve[0]; multiplied by sca_mat
            sewing_count = len(pocket_resampled_segs)  # cos  pocket.pocket_sewing - is EnumProp with only selected items
            p_mat_loc, p_mat_rot, p_mat_sca = matrix_decompose(pocket.pattern_obj.matrix_world)
            p_loc_rot_mat = p_mat_loc @ p_mat_rot  # @ scale - done in self.curve_resample
            if sewing_count > 0:
                prev_seg_idx = pocket_sewings_seg_ids[-1]
                for sewing_id in pocket_sewings_seg_ids: # if current sewing and prevous are neibors then skip first vert\
                    pocket_seg_co = pocket_resampled_segs[sewing_id][1:] if (sewing_id-prev_seg_idx == 1) or (sewing_id-prev_seg_idx == -1*sewing_count+1) else pocket_resampled_segs[sewing_id][:]
                    target_verts_co = [mat_scale @ target_mat_inv @ p_loc_rot_mat @ s_point_co for s_point_co in pocket_seg_co]
                    pocket_verts_co_projected.extend([Vector((p.x, p.y, 0)) for p in target_verts_co])
                    prev_seg_idx = sewing_id
        return pocket_verts_co_projected


    def create_mesh_cloth(self, context, pattern_curve_obj):
        spl_segments_points_flat = []  # spl_segments_points_flat[spline_id][outline_vert_id] = vert_co
        cache_segments_vert_ids = []  # store verts.idx for: sewing.verts_ids_from: cache_segments_vert_ids[spline_ic][seg_id][vert_id]
        vert_ids_low_bound = 0
        is_cyclic_spline = [b_spline.cyclic for b_spline in self.bsplines[pattern_curve_obj.name]]
        resampled_splines_segments = self.curve_resample(pattern_curve_obj)
        for spline_idx, resampled_spline_segments in enumerate(resampled_splines_segments):
            spl_segments_points_flat.append([])
            cache_segments_vert_ids.append([])
            for i, segment in enumerate(resampled_spline_segments):  # add cloth outline
                seg_len = len(segment)
                spl_segments_points_flat[spline_idx].extend(segment[:-1])
                cache_segments_vert_ids[spline_idx].append([x for x in range(vert_ids_low_bound, vert_ids_low_bound+seg_len)]) #ids
                vert_ids_low_bound += (seg_len - 1) #next_segment.pt[0].idx == prev_segment.pt[-1].idx, cos last point vert.idx is in prev_seg and next_seg

            if is_cyclic_spline[spline_idx]:  # basically last_segment, last_vert.idx == 0 (first seg, first_vert.idx ==0)
                cache_segments_vert_ids[spline_idx][-1][-1] = cache_segments_vert_ids[spline_idx][0][0]  # loop and close last vert for spline
            else:  # add last non cyclic vert, for last segment
                spl_segments_points_flat[spline_idx].append(segment[-1])
                cache_segments_vert_ids[spline_idx][-1].append(vert_ids_low_bound)  # [last_spline][last_segment][vert_ids]  - append last vert
                vert_ids_low_bound += 1

        # NOTE: fills kd tree with outline verts, also create boundary bm.edges
        bm = bmesh.new()   # create an empty BMesh
        kd_outline = mathutils.kdtree.KDTree(2*vert_ids_low_bound)  # for marging, 2x resolution of verts
        for spl_id, segments_points_flat in enumerate(spl_segments_points_flat):
            prev_vert = None
            for i, point_co in enumerate(segments_points_flat):
                new_vert = bm.verts.new(point_co)
                kd_outline.insert(point_co, i)
                if i == 0:
                    first_vert = new_vert
                if prev_vert:
                    kd_outline.insert((prev_vert.co+point_co)/2, i*2) #kd ids are not importatn here, increase border res 2x
                    bm.edges.new([prev_vert, new_vert])
                prev_vert = new_vert
            if is_cyclic_spline[spl_id]:
                kd_outline.insert((first_vert.co+new_vert.co)/2, i*2)  # kd ids are not importatn here
                bm.edges.new([first_vert, new_vert])  # close the boundary loop connecting last and firt verlF)t
        kd_outline.balance()

        garment = self.garment
        # filled 2d curve
        if pattern_curve_obj.data.dimensions == '2D' and pattern_curve_obj.data.fill_mode != 'NONE': #
            bmesh.ops.triangle_fill(bm, use_beauty=False, use_dissolve=False, edges=bm.edges)
            # sourceTri_BVHT = BVHTree.FromPolygons(spl_segments_points_flat, [tuple(i for i in range(len(spl_segments_points_flat)))], all_triangles=False)  # [0,1,2] - polygon == vert indices list
            outline_bvh = BVHTree.FromBMesh(bm)  # [0,1,2] - polygon == vert indices list
            bm.free()

            seg_len = self.get_segments_spacing(context, spl_segments_points_flat[0])  # ? make it for all splines? Or leave and loop over splines...
            seg_len /= context.scene.unit_settings.scale_length

            # generate point cloud in shape of resampled curve outline
            merged_outline_pts = []
            [merged_outline_pts.extend(p) for p in spl_segments_points_flat]
            outline_as_np = np.array(merged_outline_pts)  # !assume first spline is the outer one...
            # outline_as_np = np.einsum('ij,aj->ai', mat, outline_as_np))  # mat times point list
            max_x = np.max(outline_as_np[:, 0])
            min_x = np.min(outline_as_np[:, 0])
            max_y = np.max(outline_as_np[:, 1])
            min_y = np.min(outline_as_np[:, 1])
            point_cloud = []
            i = 0
            # seg_len /= obj_scale
            if garment.output_topology == 'TRIS':
                border_marging = 0.5 * seg_len
                for y in frange(min_y, max_y, seg_len):
                    i += 1
                    for x in frange(min_x, max_x, seg_len):  # add evenly spaced out grid of verts
                        x_mod = x + seg_len / 2 if i % 2 == 0 else x
                        point = Vector((x_mod, y, 0))
                        location, normal, index, dist = outline_bvh.ray_cast(Vector((x_mod, y, 10)), Vector((0, 0, -10)), 100)  # from position1, in vec2 direction
                        if index:
                            location2, index2, dist2 = kd_outline.find(point)  # skip points to close to border / give margin
                            if dist2 < border_marging:  # point to close to border
                                continue
                            point_cloud.append(point)
            if garment.output_topology == 'QUAD':
                y_cnt = floor((max_y-min_y)/seg_len)+1
                x_cnt = floor((max_x-min_x)/seg_len)+1
                grid_hit = []
                grid_pts = []
                border_marging = 0.35 * seg_len
                for idy, y in enumerate(frange(min_y, max_y, seg_len)):
                    grid_hit.append([False]*x_cnt)
                    grid_pts.append([0]*x_cnt)
                    for idx, x in enumerate(frange(min_x, max_x, seg_len)):  # add evenly spaced out grid of verts
                        point = Vector((x, y, 0))
                        location, normal, index, dist = outline_bvh.ray_cast(Vector((x, y, 10)), Vector((0, 0, -10)), 100)  # from position1, in vec2 direction
                        if index:
                            location2, index2, dist2 = kd_outline.find(point)  # skip points to close to border / give margin
                            if dist2 > border_marging:  # give margin to border
                                grid_pts[idy][idx] = point
                                grid_hit[idy][idx] = True

                for y in range(y_cnt):
                    y_down = grid_hit[y-1] if y >= 0 else [False]*x_cnt #return prev row, or falses on border
                    y_up = grid_hit[y+1] if y < y_cnt-1 else [False]*x_cnt  # return next row, or falses on border
                    for x in range(x_cnt):
                        x_left = grid_hit[y][x-1] if x >= 0 else False
                        x_right = grid_hit[y][x+1] if x < x_cnt-1 else False
                        if grid_hit[y][x] and (x_left or x_right) and (y_down[x] or y_up[x]):  # at least one neibor in +-x, and +-y
                            point_cloud.append(grid_pts[y][x])

                    # returns face as [(x,y,z) ...] list verre x,y,z are verrt ids.
            # NOTE: get projected verts, from pocket linked to pattern_curve_obj - could be passed to do_tris() in past
            # if garment.enable_pockets:
            #     pocket_verts_co_projected = self.get_extra_pocket_verts(garment, pattern_curve_obj)
            new_mesh = do_tris_from_pointcloud(spl_segments_points_flat, point_cloud, seg_len, garment.output_topology, outline_bvh, is_cyclic_spline)
        else:  # NOTE: 3d curve outline, no fill. Helper for shaping patterns
            new_mesh = bpy.data.meshes.new(name="Cloth")
            bm.to_mesh(new_mesh)
            bm.free()

        mat_scale = get_scale_mat(pattern_curve_obj.matrix_world.to_scale())  # cos we have to scale up generated mesh objs

        # NOTE: get projected verts, from pocket linked to pattern_curve_obj
        pocket_verts_co_projected = self.get_extra_pocket_verts(garment, pattern_curve_obj) if garment.enable_pockets else []

        pin_verts_co = []
        if garment.enable_pins:
            for pin in garment.pins:  # assume pins co are in obj local space...
                if pin.source_obj == pattern_curve_obj:
                    # pin_verts_co.append(mat_scale.inverted() @ pin.source_obj.matrix_world @ Vector(pin.source_co))
                    pin_verts_co.append(mat_scale @ Vector(pin.source_co))
                elif pin.target_obj == pattern_curve_obj:
                    # pin_verts_co.append(mat_scale.inverted() @ pin.target_obj.matrix_world @ Vector(pin.target_co))
                    pin_verts_co.append(mat_scale  @ Vector(pin.target_co))

        poke_new_verts(new_mesh, pocket_verts_co_projected+pin_verts_co, seg_len) # insert new verts into pattern

        new_mesh.transform(mat_scale.inverted())
        mesh_pattern_obj = bpy.data.objects.new(pattern_curve_obj.name+"_pattern", new_mesh)  # add a new object using the mesh
        mesh_pattern_obj['segments_verts'] = cache_segments_vert_ids # store vert.idx for search segment

        set_weights_uniform(mesh_pattern_obj, pattern_curve_obj.name, 1)
        for new_groups in garment.generated_vgroups:
            vg_patterns = [re.sub('\.[0-9]+$', '', vg_patt_name) for vg_patt_name in new_groups.vgroups_patterns] #remove extension .xx
            if pattern_curve_obj.name in vg_patterns:
                set_weights_uniform(mesh_pattern_obj, new_groups.name, 1)


        bpy.data.collections[garment.name].objects.link(mesh_pattern_obj)  # put the object into the scene (link)
        mesh_pattern_obj.select_set(True)  # select object
        pattern_curve_obj.select_set(False)  # select object
        context.view_layer.objects.active = mesh_pattern_obj  # set as the active object in the scene
        mesh_pattern_obj.matrix_world = pattern_curve_obj.matrix_world

        self.set_pattern_attribs(mesh_pattern_obj, pattern_curve_obj)

        return mesh_pattern_obj

