'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
from math import sqrt

import bpy
import mathutils
from mathutils import Vector
from .helper_functions import *


def get_predicted_vert_count(context, garment):
    total_point_count = 0
    units_scale = context.scene.unit_settings.scale_length
    density = units_scale * units_scale * garment.cloth_res * garment.cloth_res
    for pattern in garment.sewing_patterns:
        if pattern.pattern_obj and pattern.is_enabled:
            # print(pattern.pattern_obj.dimensions)
            area = pattern.pattern_obj.dimensions.x * pattern.pattern_obj.dimensions.y #? sometimes blender bug - dimensions ignore mw. Force scn update, or obj update?
            total_point_count += area * density
    total_point_count = int(total_point_count * 0.85) #for empty space use 85%
    print(f'Predicted output vert count: {total_point_count}')
    return total_point_count


def get_avg_pos_deviation(points):
    ''' takes points list '''
    p_count = len(points)
    avg_co = Vector((0, 0, 0))
    for point in points:
        avg_co += point.co
    avg_co = avg_co / p_count
    std_dev = 0
    for point in points:
        std_dev += (point.co - avg_co).length_squared
    std_dev = sqrt(std_dev/(p_count-1))
    return (avg_co, std_dev)


def get_segments_len(segments_points):
    '''Return segments  points len, from input pts '''
    out_segments_lens = []
    for points in segments_points:
        previous_point = points[0]
        segment_len = 0
        for point in points:
            point_diff_vec = point - previous_point
            segment_len += point_diff_vec.length
            previous_point = point
        out_segments_lens.append(segment_len)
    return out_segments_lens


def distance_point_line(line1, line2, p):
    # two points define the line
    p = Vector((0.0, 0.2, 0.5))
    n = (line2 - line1).normalized()

    ap = p - line1
    # t = ap.dot(n)
    # x = line1 + t * n  # x is a point on line
    # print("distance from p:", (p - x).length)

    d = ap.cross(n).length
    return d


def spline_bbox(spline):
    max_x = -1000
    min_x = 1000
    max_y = -1000
    min_y = 1000
    for p in spline.bezier_points:
        max_x = max(p.co.x, max_x)
        min_x = min(p.co.x, min_x)
        max_y = max(p.co.y, max_y)
        min_y = min(p.co.y, min_y)
    return (min_x, max_x, min_y, max_y)

def spline_remove_point(obj, spline, points_ids):
    ''' remov points froom obj while mantaining selection state'''
    backup_mode = obj.mode

    #to prevent removing points from both curves in edit mode
    for ob in bpy.context.selected_objects:
        ob.select_set(False)
    bpy.context.view_layer.objects.active = obj
    obj.select_set(True)

    # remove points from points_ids by selecting them and dissolve
    for spl in obj.data.splines:
        for p in spl.bezier_points:
            p.select_control_point = False
            p.select_left_handle = False
            p.select_right_handle = False
    for point_id in points_ids:
        spline.bezier_points[point_id].select_control_point = True
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.curve.delete(type='VERT')  # context override dosent work no matter what
    bpy.ops.object.mode_set(mode=backup_mode)  # we loose spline after going in and out of edit mode

    spline.bezier_points.update()
    #? why it was here? Hack
    #? spline = obj.data.splines[0]


def find_curve_pairs(splines):
    ''' Takes blender b_splines '''
    spl_pairs = {}
    splines_stats = []
    kd = mathutils.kdtree.KDTree(len(splines))
    for spl_id, spl in enumerate(splines):
        splines_stats.append(get_avg_pos_deviation(spl.bezier_points))  # (avg_co, std_dev)
        kd.insert(splines_stats[-1][0], spl_id)
    kd.balance()

    for spl_id, (avg_co, std_dev) in enumerate(splines_stats):
        spl_pairs[spl_id] = None
        if spl_id == 0:
            continue
        mirrored_co = avg_co
        mirrored_co.x = -1 * mirrored_co.x
        for (co, index, dist) in kd.find_n(mirrored_co, 2):
            if index and index != spl_id and dist < std_dev/2:  # use 1/20 of std_dev to find doubles
                spl_pairs[spl_id] = index
    return spl_pairs
