'''
Copyright (C) 2018 CG Cookie
https://github.com/CGCookie/retopoflow
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import inspect

import bpy
import bmesh
import gpu
import mathutils
from bpy_extras import view3d_utils
from gpu_extras.batch import batch_for_shader
from mathutils import Matrix, Vector
from mathutils.bvhtree import BVHTree
from mathutils.geometry import intersect_line_sphere
from .debug import debugger
from .curve_helpers import get_segments_len
from ..spline_tools.spline_base import *

class CookieCutter_FSM:
    class FSM_State:
        @staticmethod
        def get_state(state, substate):
            return '%s__%s' % (str(state), str(substate))

        def __init__(self, state, substate='main'):
            self.state = state
            self.substate = substate

        def __call__(self, fn):
            self.fn = fn
            self.fnname = fn.__name__

            def run(*args, **kwargs):
                try:
                    return fn(*args, **kwargs)
                except Exception as e:
                    print(f'Caught exception in function "{self.fnname}" ("{self.state}", "{self.substate}")')
                    # self.end_cancel()
                    debugger.print_exception()
                    args[0].end_cancel()  # self.end_cancel()
                    return
            run.fnname = self.fnname
            run.fsmstate = CookieCutter_FSM.FSM_State.get_state(self.state, self.substate)
            return run

    def fsm_init(self):
        self._state_next = 'main'
        self._state = None
        self._fsm_states = {}
        self._fsm_change_callbacks = []
        for (m, fn) in self.find_fns('fsmstate'):
            assert m not in self._fsm_states, 'Duplicate states registered!'
            self._fsm_states[m] = fn

    def _fsm_call(self, state, substate='main', fail_if_not_exist=False):
        s = CookieCutter_FSM.FSM_State.get_state(state, substate)
        if s not in self._fsm_states:
            assert not fail_if_not_exist
            return
        try:
            return self._fsm_states[s](self)
        except Exception as e:
            print('Caught exception in state ("%s")' % (s))
            debugger.print_exception()
            return

    def fsm_change(self, state):
        self._state_next = None
        if self._state == state:
            return True
        if self._fsm_call(self._state, substate='can exit') == False:
            print('Cannot exit %s' % str(self._state))
            return False
        if self._fsm_call(state, substate='can enter') == False:
            print('Cannot enter %s' % str(state))
            return False
        print('%s -> %s' % (str(self._state), str(state)))
        self._fsm_call(self._state, substate='exit')
        self._state = state
        self._fsm_call(self._state, substate='enter')
        for fn in self._fsm_change_callbacks:
            fn()
        return True

    def fsm_change_callback(self, callback):
        self._fsm_change_callbacks += [callback]

    def fsm_update(self):
        if self._state_next is not None and self._state_next != self._state:
            if self._fsm_call(self._state, substate='can exit') == False:
                # print('Cannot exit %s' % str(self._state))
                self._state_next = None
                return
            if self._fsm_call(self._state_next, substate='can enter') == False:
                # print('Cannot enter %s' % str(self._state_next))
                self._state_next = None
                return
            # print('%s -> %s' % (str(self._state), str(self._state_next)))
            self._fsm_call(self._state, substate='exit')
            self._state = self._state_next
            self._fsm_call(self._state, substate='enter')
        self._state_next = self._fsm_call(self._state, fail_if_not_exist=True)


class CookieCutter_Utils:
    def find_fns(self, key):
        c = type(self)
        objs = [getattr(c, k) for k in dir(c)]
        fns = [fn for fn in objs if inspect.isfunction(fn)]
        return [(getattr(fn, key), fn) for fn in fns if hasattr(fn, key)]


class ModalCommon(CookieCutter_Utils):
    """Common Operator methods for garment editing are defined here"""

    _handle_draw_verts = None
    _handle_draw_text = None

    patter_segments_cache = {}
    garment_index = 0
    event = None
    context = None
    allow_nav = False
    done = False

    @staticmethod
    def draw_text(self, context):
        '''Cos self is passed by handler'''
        pass

    @staticmethod
    def draw_px(self, context):
        '''Cos self is passed by handler'''
        pass

    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
    shader_2d = gpu.shader.from_builtin('UNIFORM_COLOR')

    @staticmethod
    def visible_curves(context):
        """Loop over (object, matrix) pairs (mesh only)"""
        for obj in context.visible_objects:
            if obj.type == 'CURVE' and len(obj.data.splines) and obj.data.splines[0].type == 'BEZIER':
                yield (obj)

    @staticmethod
    def scene_objects_and_duplis(context):
        """Loop over (object, matrix) pairs (mesh only)"""
        for obj in context.scene.objects:
            if obj.type == 'CURVE' and len(obj.data.splines) and obj.data.splines[0].type == 'BEZIER':
                yield (obj)

    def garment_index_from_obj(self, context, obj):
        garment_index = -1
        if obj.type == 'CURVE' and len(obj.data.splines) and obj.data.splines[0].type == 'BEZIER':
            for idx, garment in enumerate(context.scene.cloth_garment_data):
                for pattern in garment.sewing_patterns[:] + garment.pockets[:]:
                    if pattern.pattern_obj == obj:
                        garment_index = idx
                        break
        elif obj.type == 'MESH': # check if it is garment.target_cloth_object (output mesh obj)
            for idx, garment in enumerate(context.scene.cloth_garment_data):
                if garment.target_cloth_object == obj:
                    garment_index = idx
                    break
        return garment_index

    def obj_ray_cast(self, obj, ray_origin, ray_target):
        """Wrapper for ray casting that moves the ray into object space"""
        # get the ray relative to the object
        ray_direction = ray_target - ray_origin
        sourceTri_BVHT = self.construct_pattern_BVHTree(obj)  # [0,1,2] - polygon == vert indices list
        # location, normal, index, dist =
        return sourceTri_BVHT.ray_cast(ray_origin, ray_direction, 600)

    def curve_ray_cast(self, obj, ray_origin, ray_target):
        """Wrapper for ray casting that moves the ray into object space"""
        # get the ray relative to the object
        ray_direction = ray_target - ray_origin
        sourceTri_BVHT = self.construct_pattern_BVHTree(obj)
        # location, normal, index, dist =
        return sourceTri_BVHT.ray_cast(ray_origin, ray_direction, 9999)

    def construct_pattern_BVHTree(self, obj):
        # give world space points ( multiplied by obj.mat_world)
        resampled_splines_segments = self.patter_segments_cache[obj.name]
        spl_segments_points_flat = []  # spl_segments_points_flat[spline_id][outline_vert_id] = vert_co
        for spline_idx, resampled_spline_segments in enumerate(resampled_splines_segments):
            spl_segments_points_flat.append([])
            for segment in resampled_spline_segments:  # add cloth outline
                spl_segments_points_flat[spline_idx].extend(segment[:-1])

        bm = bmesh.new()   # create an empty BMesh
        for segments_points_flat in spl_segments_points_flat:
            border_edges = []
            prev_vert = None
            for i, point_co in enumerate(segments_points_flat):
                new_vert = bm.verts.new(point_co)
                if i == 0:
                    first_vert = new_vert
                if prev_vert:
                    border_edges.append(bm.edges.new([prev_vert, new_vert]))
                prev_vert = new_vert
            border_edges.append(bm.edges.new([first_vert, new_vert]))  # close the boundary loop connecting last and firt vert
            bmesh.ops.triangle_fill(bm, use_beauty=True, use_dissolve=False, edges=border_edges)
            # bmesh.ops.contextual_create(bm, geom=bm.verts)

        # for segments_points_flat in spl_segments_points_flat:
        #     for point_co in segments_points_flat:
        #         bm.verts.new(point_co)
        silhouette_bvh = BVHTree.FromBMesh(bm)
        # new=bpy.data.meshes.new('temp')
        # bm.to_mesh(new)
        # new_obj = bpy.data.objects.new('temp', new)
        # bpy.context.collection.objects.link(new_obj)
        bm.free()
        return silhouette_bvh

    @staticmethod
    def find_mesh_obj_under_cursor(context, event, depsgraph):
        ''' using scn.ray_cast '''
        region = context.region
        rv3d = context.region_data
        coord = event.mouse_region_x, event.mouse_region_y
        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
        if bpy.app.version >= (2, 91, 0):
            hit, loc, norm, idx, obj, mat = context.scene.ray_cast(depsgraph, ray_origin, view_vector, distance=1.0e+8)
        else:
            hit, loc, norm, idx, obj, mat = context.scene.ray_cast(context.view_layer, ray_origin, view_vector, distance=1.0e+8)
        return hit, loc, obj

    def find_curve_under_cursor(self, event, curves_2d = True, curves_3d = False, depth=None):
        """Run this function on left mouse, execute the ray cast
        Get only 2d curves, or 3d curves or both
        Use falback depth, for raycasting when nothing was hit
        """
        context = bpy.context
        # get the context arguments
        region = context.region
        rv3d = context.region_data
        coord = event.mouse_region_x, event.mouse_region_y

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

        ray_target = ray_origin + view_vector

        # cast rays and find the closest object
        best_length_squared = 9000.0
        best_2d_dist = 10 # 10 px max for picking
        hit, loc_world, best_curve_obj = False, None, None
        for obj in self.visible_curves(context):
            if curves_2d and obj.data.dimensions=='2D':
                hit_w, normal, face_index, dist = self.curve_ray_cast(obj, ray_origin, ray_target)
                if hit_w is not None:
                    length_squared = (hit_w - ray_origin).length_squared
                    if best_curve_obj is None or length_squared < best_length_squared:
                        best_length_squared = length_squared
                        hit, loc_world, best_curve_obj = True, hit_w, obj
            if curves_3d and obj.data.dimensions=='3D': #returns 2d distance... in screen space
                spline_3d = BezierSplines(obj)
                closest_spl_id, closest_spl_point_l, closest_tn, dist_2d = spline_3d.closest_spl_point_view_space(context, event, max_dist=best_2d_dist, world_space=True)
                if closest_spl_id is not None: #cos can be spl_id ==0 for first spline
                    if best_curve_obj is None or dist_2d < best_2d_dist:
                        best_2d_dist = dist_2d
                        hit, loc_world, best_curve_obj = True, obj.matrix_world @ closest_spl_point_l, obj

        if not hit and depth:  # from last hit co
            loc_world = view3d_utils.region_2d_to_location_3d(region, rv3d, coord, depth_location=depth)
        # now we have the object under the mouse cursor,
        # return best_curve_obj
        return hit, loc_world, best_curve_obj

    def deselect_all(self):
        for obj in self.context.view_layer.objects:
            obj.select_set(False)

    def ray_cast_curve_local(self, context, hit_obj):
        ''' Return index of pin closest to mouse'''
        """Run this function on left mouse, execute the ray cast"""
        region = context.region
        rv3d = context.region_data
        coord = self.event.mouse_region_x, self.event.mouse_region_y

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

        ray_target = ray_origin + view_vector
        # cast rays and find the closest object
        hit_world, normal, face_index, dist = self.curve_ray_cast(hit_obj, ray_origin, ray_target)
        hit_local = hit_obj.matrix_world.inverted() @ hit_world if hit_world else None
        if hit_local is not None:
            hit_local.z = 0  # cos it is flat
        return hit_local

    def find_closest_segment(self, curve_name):
        ''' Return index of sewing closest to mouse'''
        region = self.context.region
        rv3d = self.context.region_data
        splines_segments = self.patter_segments_cache[curve_name]
        best_dist = 1000
        ret_idx = 0
        ret_spline = 0
        for spl_idx, segments in enumerate(splines_segments):
            sum_segments_points_count = 0
            for segment in segments:
                sum_segments_points_count += len(segment)

            kd = mathutils.kdtree.KDTree(sum_segments_points_count)
            for seg_id, segment in enumerate(segments):
                for point in segment:
                    point_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, point, default=None)
                    kd.insert(Vector((point_2d[0], point_2d[1], 0)), seg_id)
            kd.balance()
            # Find the closest point to the center
            co, index, dist = kd.find(Vector((self.event.mouse_region_x, self.event.mouse_region_y, 0)))
            if dist < best_dist:
                best_dist = dist
                ret_idx = index
                ret_spline = spl_idx
        return (ret_spline, ret_idx)

    def assign_target_obj_to_garment(self,  target_obj):
        for pattern in self.garment.sewing_patterns:
            if pattern.pattern_obj == target_obj:
                return
        self.garment.sewing_patterns.add()
        self.garment.sewing_patterns[-1].pattern_obj = target_obj

    def get_pattern_from_curve(self,  curve):
        for p_id, pattern in enumerate(self.garment.sewing_patterns):
            if pattern.pattern_obj == curve:
                return p_id
        return -1

    # @CookieCutter_FSM.FSM_State('main')
    # def modal_main(self):
    #     pass

    def end_cancel(self):
        self.disable_handlers()
        bpy.ops.ed.undo()
        bpy.ops.garment.cleanup(garment_index=self.garment_index)
        self.done = 'CANCELLED'

    def end_commit(self):
        self.disable_handlers()
        bpy.ops.garment.cleanup(garment_index=self.garment_index)
        self.done = 'FINISHED'

    def disable_handlers(self):
        if self._handle_draw_verts:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_draw_verts, 'WINDOW')
            self._handle_draw_verts = None
        if self._handle_draw_text:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_draw_text, 'WINDOW')
            self._handle_draw_text = None

    def enable_draw_handlers(self, context):
        args = (self, context)
        if not self._handle_draw_verts:
            self._handle_draw_verts = bpy.types.SpaceView3D.draw_handler_add(self.draw_px, args, 'WINDOW', 'POST_VIEW')
        if not self._handle_draw_text:
            self._handle_draw_text = bpy.types.SpaceView3D.draw_handler_add(self.draw_text, args, 'WINDOW', 'POST_PIXEL')

    def invoke_basic(self, context, event):
        bpy.ops.ed.undo_push()
        self.event = event
        self.context = context
        self.allow_nav = False
        self.done = False

        self.enable_draw_handlers(context)

    def calc_pattern_segments_cache(self, context):
        self.patter_segments_cache.clear()

        #maybe do this on pocket picking? But we need it for hover before
        for obj in self.visible_curves(context):
            if obj.type == 'CURVE' and obj.data.splines[0].type == 'BEZIER':  # use_cyclic_u
                bspline = BezierSplines(obj)
                resampled_splines_segments = bspline.generate_splines_pts_cloud(8, world_space=True)  # for fast bgl drawing
                self.patter_segments_cache[obj.name] = resampled_splines_segments  # pattern_segments_cache[obj.name][spline_idx][segment_idx]
                self.patter_segments_cache[obj.name+'_len'] = [get_segments_len(resampled_segments) for resampled_segments in resampled_splines_segments]   # patter_segments_cache[obj.name][spline_idx][segment_idx]



    def invoke_garment(self, context, event):
        bpy.ops.garment.cleanup(garment_index=self.garment_index)
        self.invoke_basic(context, event)
        self.deselect_all()

        self.garment = context.scene.cloth_garment_data[self.garment_index]
        self.calc_pattern_segments_cache(context)






