'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import gpu
import gpu
from mathutils import Matrix, Vector
from gpu_extras.batch import batch_for_shader
from bpy_extras import view3d_utils
import blf
import math
import numpy as np
from mathutils.bvhtree import BVHTree

def obj_ray_cast(obj, ray_origin, ray_target, depsgraph = None):
    # get the ray relative to the object
    ray_direction = ray_target - ray_origin
    sourceTri_BVHT = get_obj_mesh_bvht(obj, depsgraph = None)  # [0,1,2] - polygon == vert indices list
    # location, normal, index, dist =
    return sourceTri_BVHT.ray_cast(ray_origin, ray_direction, 600)


def view_obj_raycast(context, obj, event, applyModifiers=True, depsgraph=None, world_space=True):
    ''' Return index of pin closest to mouse'''
    """Run this function on left mouse, execute the ray cast"""
    region = context.region
    rv3d = context.region_data
    coord = event.mouse_region_x, event.mouse_region_y
    # coord = mouse_coordinates

    # get the ray from the viewport and mouse
    view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
    ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

    sourceTri_BVHT = get_obj_mesh_bvht(obj, applyModifiers=applyModifiers, depsgraph=depsgraph, world_space=False)  # do not transform worldspace for performance

    matrix_inv = obj.matrix_world.inverted()  # faster than transforming whole mesh
    ray_origin_obj = matrix_inv @ ray_origin
    view_vector_obj = matrix_inv @ Vector(view_vector.xyz[:]+(0,))
    hit_local, normal, index, dist = sourceTri_BVHT.ray_cast(ray_origin_obj, view_vector_obj.xyz)
    # get the ray relative to the object
    if hit_local and world_space: #cos we always hit in local space, for performance
        location = obj.matrix_world @ hit_local
    else:
        location = hit_local

    return location, normal, index, dist

def raycast_bvh(obj_bvh, context, event):
    region = context.region
    rv3d = context.region_data
    coord = event.mouse_region_x, event.mouse_region_y
    # get the ray from the viewport and mouse
    view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
    ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

    # if rv3d.view_perspective == 'ORTHO':
    ray_target = ray_origin + view_vector

    obj = context.active_object
    """Wrapper for ray casting that moves the ray into object space"""
    # get the ray relative to the object
    mat_inv = obj.matrix_world.inverted()
    ray_origin_obj = mat_inv @ ray_origin  # to bbox space
    ray_target_obj = mat_inv @ ray_target  # to bbox space
    view_vector_obj = ray_target_obj - ray_origin_obj

    face_hit_loc, normal, index, distance = obj_bvh.ray_cast(ray_origin_obj, view_vector_obj)
    return face_hit_loc, normal, index, distance

def get_weights(ob, vgroup):
    group_index = vgroup.index
    return [[g.weight for g in vert.groups if g.group == group_index] for vert in ob.data.vertices]


def set_weights(obj, weights, vgroup):
    g_id = vgroup.index
    if not weights:
        obj.vertex_groups[g_id].remove([i for i in range(len(obj.data.vertices))])

    for v_id, weight in enumerate(weights):
        if weight:
            obj.vertex_groups[g_id].add([v_id], weight[0], "REPLACE")
        else:
            obj.vertex_groups[g_id].remove([v_id])

def set_weights_uniform(obj, vg_name, weight, v_ids = None):
    ''' Set weights to all vertices in vg_name to weight '''
    vg = obj.vertex_groups.get(vg_name)
    if not vg:
        vg = obj.vertex_groups.new(name=vg_name)
    v_ids = v_ids or list(range(len(obj.data.vertices)))
    vg.add(v_ids, weight, "REPLACE") # "ADD", "REPLACE", "SUBTRACT"

def set_vert_vec_attrib_uniform(obj, attrib_name, data):
    # XXX: make sure prop vector is in correct space. World (for after merge) or local?
    origin_attrib = obj.data.attributes.get(attrib_name)
    if not origin_attrib:
        origin_attrib = obj.data.attributes.new(attrib_name, 'FLOAT_VECTOR','POINT')
    np_array = np.array(data.to_3d().to_tuple()*len(obj.data.vertices))
    # end_uv_attr.data[v_idx].vector = uv_end
    origin_attrib.data.foreach_set("vector",np_array)


def set_vert_float_attrib_uniform(obj, attrib_name, value):
    # XXX: make sure prop vector is in correct space. World (for after merge) or local?
    origin_attrib = obj.data.attributes.get(attrib_name)
    if not origin_attrib:
        origin_attrib = obj.data.attributes.new(attrib_name, 'FLOAT','POINT')
    np_array = np.full(len(obj.data.vertices), value)
    # end_uv_attr.data[v_idx].vector = uv_end
    origin_attrib.data.foreach_set("value",np_array)


def get_obj_mesh_bvht(obj, depsgraph=None, applyModifiers=True, world_space=True):
    if applyModifiers:
        if world_space:
            # faster 4x  than:
            # obj.update_tag()
            # bpy.context.view_layer.update()
            depsgraph.objects[obj.name].data.transform(obj.matrix_world)
            bvh = BVHTree.FromObject(obj, depsgraph)
            depsgraph.objects[obj.name].data.transform(obj.matrix_world.inverted())

            return bvh
        else:
            return BVHTree.FromObject(obj, depsgraph)
    else:
        if world_space:
            # 4 times slower than data.transform
            #bvh1 =  BVHTree.FromPolygons([obj.matrix_world @ v.co for v in obj.data.vertices], [p.vertices for p in obj.data.polygons])
            obj.data.transform(obj.matrix_world)
            bvh = BVHTree.FromPolygons([v.co for v in obj.data.vertices], [p.vertices for p in obj.data.polygons])
            obj.data.transform(obj.matrix_world.inverted())
            return bvh
        else:
            return BVHTree.FromPolygons([v.co for v in obj.data.vertices], [p.vertices for p in obj.data.polygons])


def assign_material(obj, material_name, clear_materials=True):
    '''Add material to obj. If clear_material is True - reomve all slots except first.'''
    if material_name not in bpy.data.materials.keys():
        print('Material %s dosen\'t exist!' % (material_name))
        return
    mat = bpy.data.materials[material_name]
    if clear_materials is True:
        while len(obj.material_slots) > 1:
            obj.data.materials.pop()
    if len(obj.material_slots) == 0:  # make sure first slot is assigned
        obj.data.materials.append(mat)
    else:
        obj.material_slots[0].material = mat


def unlink_from_scene(obj):
    for col in obj.users_collection:
        col.objects.unlink(obj)
    if obj.name in bpy.context.scene.collection.objects.keys():
        bpy.context.scene.collection.objects.unlink(obj)
    obj.use_fake_user = True


def link_child_to_collection(parent_obj, clone):
    ''' Link clone to collection where parent_obj is located '''
    if parent_obj.users_collection:
        if clone.name not in parent_obj.users_collection[0].objects.keys():
            parent_obj.users_collection[0].objects.link(clone)
        else:
            return
    else:
        if clone.name not in bpy.context.scene.collection.objects.keys():
            bpy.context.scene.collection.objects.link(clone)
        else:
            return

# function to get all node groups in a node tree recursively
def get_nested_node_groups(node_tree):
    node_groups = []
    for node in node_tree.nodes:
        if node.bl_idname == 'GeometryNodeGroup':
            node_groups.append(node.node_tree)
            node_groups.extend(get_nested_node_groups(node.node_tree))
    return set(node_groups) # withotu duplicates


def remap_node_groups(new_ngroups, original_ngroups, data_type='node_groups'):
    for orig_ng_name, old_node_gr in original_ngroups.items(): # old name: Twist
        for new_ng in new_ngroups: # imported name: Twist.001
            if new_ng.name.startswith(orig_ng_name):   # 'Twist.001'.startswith('Twist')
                # new_data = data_to.node_groups[all_lib_node_groups_names.index(orig_node_gr_name)] # we acces by index
                print(f'remapping: {orig_ng_name} to new: {new_ng.name}')
                old_node_gr.user_remap(new_ng)
                new_ng.name = orig_ng_name
                data_handle = getattr(bpy.data, data_type)
                data_handle.remove(old_node_gr)


data_types = ['node_groups']  # 'materials', 'textures' etc from gpro

def import_node_group(node_name, lib_file="\bend_deform_gt.blend", force_update=False):
    ''' import node gr. from ht_nodes.blend '''
    #TODO: we should also udpate inner group nodes....
    old_node_gr = bpy.data.node_groups.get(node_name)

    # store name now before it gets overridden
    original_node_groups = {node_group.name: node_group for node_group in bpy.data.node_groups}

    if not old_node_gr or force_update:
        old_node_for_remap = bpy.data.node_groups.get(node_name)
        with bpy.data.libraries.load(lib_file) as (data_from, data_to):
            data_to.node_groups = [node_n for node_n in data_from.node_groups if node_n == node_name]
            imported_gr_names = [node_n for node_n in data_from.node_groups if node_n == node_name]  # stores names
            print(f"{imported_gr_names=}")
            all_lib_node_groups_names = data_from.node_groups[:] # part gets imported

        if old_node_for_remap: #remap main node
            old_node_for_remap.user_remap(data_to.node_groups[0])
            data_to.node_groups[0].name = node_name
            bpy.data.node_groups.remove(old_node_for_remap)

        # imported node_gr - may contain nested node groups - scan for them and remap old ngroups to new ones
        imported_node_groups = get_nested_node_groups(data_to.node_groups[0])
        remap_node_groups(imported_node_groups, original_node_groups)

        return data_to.node_groups[0]
    else:
        return old_node_gr


def add_driver( source, target, prop, dataPath, index=-1, negative=False, func='' ):
    ''' Add driver to source prop (at index), driven by target dataPath '''

    if index != -1:
        d = source.driver_add(prop, index).driver
    else:
        d = source.driver_add(prop).driver

    v = d.variables.new()
    v.name = prop
    v.targets[0].id = target
    v.targets[0].data_path = dataPath

    d.expression = func + "(" + v.name + ")" if func else v.name
    d.expression = d.expression if not negative else "-1 * " + d.expression


def angle_signed(vA, vB, vN):
    '''angle betwen a - b, is vN space '''
    a = vA.normalized()
    b = vB.normalized()
    adotb = a.dot(b)  # not sure why but cos(x) goes above 1, and below -1   if a= -b
    if a.dot(b) > 1:
        adotb = 1
    elif a.dot(b) < -1:
        adotb = -1
    angle = math.acos(adotb)
    cross = a.cross(b)
    if vN.dot(cross) < 0:  # // Or > 0
        angle = -angle
    return angle


def sgn(x):
    return -1 if (x < 0.0) else 1

def frange(x, y, jump):
    while x < y:
        yield x
        x += jump


def get_scale_mat(vec_scale):
    mat_scale = Matrix.Identity(4)
    mat_scale[0][0] = vec_scale.x
    mat_scale[1][1] = vec_scale.y
    mat_scale[2][2] = vec_scale.z
    return mat_scale


LEFT = ('l', 'L','left','Left','LEFT')
RIGHT = ('r', 'R','right','Right','RIGHT')

def is_name_r_l(name):
    return name.endswith(LEFT) or name.endswith(RIGHT)

def get_mirrored_name(name):
    for i, l_suffx in enumerate(LEFT):
        if name.endswith('.'+l_suffx) or name.endswith('_'+l_suffx):
            return name[:-1*len(l_suffx)] + RIGHT[i]
    for i, r_suffx in enumerate(RIGHT):
        if name.endswith('.'+r_suffx) or name.endswith('_'+r_suffx):
            return name[:-1*len(r_suffx)] + LEFT[i]
    return ''

LEFT_MARGIN = 80
TOP_POSITION = 200
BOX_WIDTH = 395

def draw_text_line(lines):
    line_height_h1 = 35
    line_height_p = 28
    font_id = 0
    current_line_pos = TOP_POSITION
    blf.color(font_id, 1.0, 1.0, 1.0, 1.0)
    for line in lines:
        text, line_type = line
        font_size = 24 if line_type == 'H1' else 20
        if bpy.app.version < (4, 0, 0):
            blf.size(font_id, font_size, 60)
        else: # for blender 4 4 and up
            blf.size(font_id, font_size)
        blf.position(font_id, LEFT_MARGIN, current_line_pos, 0)
        blf.draw(font_id, text)
        if line_type == 'H1':
            current_line_pos -= line_height_h1
        else:
            current_line_pos -= line_height_p
    draw_background_box(width=BOX_WIDTH, height=TOP_POSITION +
                        line_height_h1, x0=LEFT_MARGIN-20, y0=current_line_pos)


shader2d = gpu.shader.from_builtin('UNIFORM_COLOR')
def draw_background_box(width, height, x0, y0):
    # draw box
    x1 = width
    y1 = height
    positions = [[x0, y0], [x0, y1], [x1, y1], [x1, y0]]

    vertices = (
        (x0, y0), (width, y0),
        (x0, height), (width, height))
    indices = ( (0, 1, 2), (2, 1, 3))
    gpu.state.blend_set("ALPHA")
    batch = batch_for_shader(shader2d, 'TRIS', {"pos": vertices}, indices=indices)

    shader2d.bind()
    shader2d.uniform_float("color", (0., 0., 0., 0.4))
    batch.draw(shader2d)
    gpu.state.blend_set("NONE")


def get_addon_name():
    return __package__.split(".")[0]


def addon_name_lowercase():
    return get_addon_name().lower()


def get_addon_preferences():
    return bpy.context.preferences.addons[get_addon_name()].preferences

def get_mat_world_obj(obj):
    obj.matrix_basis == (Matrix.Translation(obj.location) @ obj.rotation_quaternion.to_matrix().to_4x4() @ get_scale_mat(obj.scale))
    # replace rotation_quaternion with rotation_euler if you are using Euler mode
    matrix_local = obj.matrix_parent_inverse @ obj.matrix_basis
    matrix_world = matrix_local if obj.parent is None else obj.parent.matrix_word @ matrix_local
    return matrix_world


def matrix_decompose(matrix_world):
    ''' returns active_obj_mat_loc, active_obj_mat_rot, active_obj_mat_sca
        reconstruct by loc @ rotQuat @ scale '''

    loc, rotQuat, scale = matrix_world.decompose()

    active_obj_mat_loc = Matrix.Translation(loc)
    active_obj_mat_rot = rotQuat.to_matrix().to_4x4()
    active_obj_mat_sca = Matrix()
    for i in range(3):
        active_obj_mat_sca[i][i] = scale[i]

    return active_obj_mat_loc, active_obj_mat_rot, active_obj_mat_sca


def get_new_mat_parent_inv(old_parent_m_w, new_parent_m_w, old_obj_m_p_inv):
    ''' cos: old_m_w == new_m_w'''
    ''' so:  old_parent_m_w @ obj.m_p_inv1 @ obj.m_basis  ==  new_parent_m_w @ obj.m_p_inv2 @ obj.m_basis'''
    ''' so:  new_parent_m_w.inv() @ old_parent_m_w @ obj.m_p_inv1   ==   obj.m_p_inv2'''
    return new_parent_m_w.inverted() @ old_parent_m_w @ old_obj_m_p_inv

def create_obj_copy(obj, instance = False):
    'create object copy, non instance'
    clone = obj.copy()  # copy object
    clone.data = obj.data.copy()  # and its data
    clone.matrix_world = obj.matrix_world
    link_to_same_collection(obj, clone)
    if instance:
        clone.name = obj.name+'.l'
        obj.name = obj.name+'.r'
    return clone


def create_empty(name):
    'create object copy, non instance'
    prefs = get_addon_preferences()
    new_empty = bpy.data.objects.new(name, None)
    new_empty.empty_display_size = prefs.pin_empties_size
    new_empty.empty_display_type = 'SPHERE'
    return new_empty
    # link_to_same_collection(obj, clone)


def get_new_mat_parent_inv(old_parent_m_w, new_parent_m_w, old_obj_m_p_inv):
    ''' cos: old_m_w == new_m_w'''
    ''' so:  old_parent_m_w @ obj.m_p_inv1 @ obj.m_basis  ==  new_parent_m_w @ obj.m_p_inv2 @ obj.m_basis'''
    ''' so:  old_parent_m_w @ obj.m_p_inv1 @ new_parent_m_w.inv()   ==   obj.m_p_inv2'''
    return old_parent_m_w @ old_obj_m_p_inv @ new_parent_m_w.inverted()

def clear_parent(obj):
    parent = obj.parent
    if parent:
        obj.matrix_basis = parent.matrix_world @ obj.matrix_parent_inverse @ obj.matrix_basis
        obj.parent = None


def set_parent(obj, new_parent):
    if obj.parent:
        old_parent_m_w = obj.parent.matrix_world.copy()  # backup cos we will change obj.parent below, so backup
        backup_obj_mpi = obj.matrix_parent_inverse.copy()  # bug in blender: obj.parent = new_parent clears obj.matrix_parent_inverse so backup it
        obj.parent = new_parent
        new_m_p_inv = new_parent.matrix_world.inverted() @ old_parent_m_w @ backup_obj_mpi
        obj.matrix_parent_inverse = new_m_p_inv
    else:
        obj.parent = new_parent
        obj.matrix_parent_inverse = new_parent.matrix_world.inverted()


def link_to_same_collection(col_src_obj, clone):
    ''' Link clone to collection where col_src_obj is located '''
    if col_src_obj.users_collection:
        for obj_coll in col_src_obj.users_collection:
            if clone.name not in obj_coll.objects.keys():
                obj_coll.objects.link(clone)
    else:
        if clone.name not in bpy.context.scene.collection.objects.keys():
            bpy.context.scene.collection.objects.link(clone)


def get_all_subcollections(current_coll):
    '''Get all sub collections (including nested) for searched collection'''
    child_colls = []
    for coll in current_coll.children:
        child_colls.extend(get_all_subcollections(coll))
    return child_colls + current_coll.children[:]


def find_view_coll(p_coll, search_name):
    if search_name == p_coll.name:
        return p_coll
    for child_col in p_coll.children:
        found_child = find_view_coll(child_col, search_name)
        if found_child:
            return found_child
    return None

def find_create_collection(context, parent_col, searched_col_name, enable_in_vlayer = True):  # TODO: set - C.view_layer.layer_collection.children['Collection'].exclude = False !
    ''' Try to find searched collection name in parent childrens. If not found create it and link it to master'''
    collection = find_view_coll(parent_col, searched_col_name)
    if not collection: #try finding it anyway and link to parent coll.
        if searched_col_name in bpy.data.collections.keys():
            collection = bpy.data.collections[searched_col_name]
        else:
            collection = bpy.data.collections.new(searched_col_name)
        parent_col.children.link(collection)
    if enable_in_vlayer:
        vcol = find_view_coll(context.view_layer.layer_collection, searched_col_name)  # find in view_layer cols (! same as bpy.data.colls...)
        vcol.exclude = False
    return collection

# find all parent collections to current coll
def get_ancestors_collections(context, coll):
    '''Returns [searched_col, .., context.scene.collection ]'''
    def get_child_recursive(coll, target):
        if coll == target:
            return [coll]
        for child in coll.children:
            result = get_child_recursive(child, target)
            if result:
                result.append(coll)
                return result
        return []
    return get_child_recursive(context.scene.collection, coll)

def get_all_children(col):
    yield col
    for child in col.children:
        yield from get_all_children(child)

# view_layer = bpy.context.view_layer
# for layer_col in reversed(list(get_all_children(view_layer.layer_collection))):
#     layer_col.exclude = True

def update_keyframe_prop(obj, data_path, frame, max_val, min_val = 0):
    ''' Basically only for 'shrink_max'  and 'sewing_force_max' '''
    if not obj.animation_data:
        obj.animation_data_create()
    if not obj.animation_data.action:
        action = bpy.data.actions.new(name='ht_cloth_anim')
        obj.animation_data.action = action
    action = obj.animation_data.action
    fc = action.fcurves.find(data_path) #   find(data_path, index=0)
    if not fc:
        fc = action.fcurves.new(data_path, index=0)
        fc.keyframe_points.add(2)
        fc.keyframe_points[0].co = 0, min_val
        fc.keyframe_points[0].handle_left_type = 'AUTO'
        fc.keyframe_points[0].handle_right_type = 'AUTO'
        fc.keyframe_points[0].interpolation = 'BEZIER' # 'LINEAR'
        fc.keyframe_points[1].handle_left_type = 'AUTO'
        fc.keyframe_points[1].handle_right_type = 'AUTO'
        fc.keyframe_points[1].co = frame, max_val
        fc.keyframe_points[1].interpolation = 'BEZIER' # 'LINEAR'
        fc.update()
    else:
        # fix second keyframe value
        # old_val = fc.keyframe_points[1].co[1]     # returns:  Vec((#frame, value))
        fc.keyframe_points[0].co = (0,min_val)
        fc.keyframe_points[1].handle_left_type = 'AUTO'
        fc.keyframe_points[1].handle_right_type = 'AUTO'
        fc.keyframe_points[1].co[1] = max_val
        fc.update() # forces handles to update

    # data_path = garment_obj.animation_data.action.fcurves[0].data_path     # data_path :=  'modifiers["Cloth"].settings.shrink_max'
    # garment_obj.path_resolve(data_path)  # gives value of current prop



def remap_sewing(context, sewing_obj, old_new_map):
    ''' When we edit sewing pattern segments, this will help remap changed segments'''
    #old_map corresponds to new_map_id sewings
    if not old_new_map:
        return
    instances = [obj for obj in bpy.data.objects if obj.data == sewing_obj.data]  # for finding patterns that were using splited pattern geo

    #DONE: add back rem sewing, if spl_id > len(obj.splines) or p_id > len(spline.bpoints) - Done in garment.cleanup
    for i, garment in enumerate(context.scene.cloth_garment_data):
        sewing_obj_is_used = False
        for idx, sewing in enumerate(garment.garment_sewings):
            if sewing.source_obj in instances and (sewing.from_spline_idx, sewing.from_segment_idx) in old_new_map:
                new_sewing = old_new_map[(sewing.from_spline_idx, sewing.from_segment_idx)]
                sewing.from_spline_idx = new_sewing[0]
                sewing.from_segment_idx = new_sewing[1]
                flip_sew = new_sewing[2]
                if flip_sew:
                    sewing.flip = not sewing.flip
                sewing_obj_is_used = True
            if sewing.target_obj in instances and (sewing.to_spline_idx, sewing.to_segment_idx) in old_new_map:
                new_sewing = old_new_map[(sewing.to_spline_idx, sewing.to_segment_idx)]
                sewing.to_spline_idx = new_sewing[0]
                sewing.to_segment_idx = new_sewing[1]
                flip_sew = new_sewing[2]
                if flip_sew:
                    sewing.flip = not sewing.flip
                sewing_obj_is_used = True
        patterns = [p.pattern_obj for p in garment.sewing_patterns]
        if sewing_obj_is_used and sewing_obj not in patterns:  # sewing pattern is listed in garment sewings. Make sure it is also in garmetn patterns list
            new_pattern = garment.sewing_patterns.add()
            new_pattern['pattern_obj'] = sewing_obj #to avoid pattern update func


def sewing_rem_spline_id(context, sewing_obj, rem_spl_id, remap_removed_spline = True):
    ''' used when we removed spline from sewing_obj'''
    #old_map corresponds to new_map_id sewings
    instances = [obj for obj in bpy.data.objects if obj.data == sewing_obj.data]  # for finding patterns that were using splited pattern geo
    garments = context.scene.cloth_garment_data
    for i, garment in enumerate(garments):
        sewing_for_removal = []
        for idx, sewing in enumerate(garment.garment_sewings):
            if sewing.source_obj in instances:
                if sewing.from_spline_idx == rem_spl_id and remap_removed_spline:
                    sewing_for_removal.append(idx)
                    continue
                elif sewing.from_spline_idx > rem_spl_id:
                    sewing.from_spline_idx -= 1
            if sewing.target_obj in instances:
                if sewing.to_spline_idx == rem_spl_id and remap_removed_spline:
                    sewing_for_removal.append(idx)
                    continue
                elif sewing.to_spline_idx > rem_spl_id:
                    sewing.to_spline_idx -= 1

        for sewing_id in reversed(sewing_for_removal):
            garment.garment_sewings.remove(sewing_id)

def remove_sewings_by_sewid(context, sewing_obj, rem_sewings_ids):
    ''' used when we removed points from splines from sewing_obj'''
    #old_map corresponds to new_map_id sewings
    if not rem_sewings_ids:
        return
    instances = [obj for obj in bpy.data.objects if obj.data == sewing_obj.data]  # for finding patterns that were using splited pattern geo
    garments = context.scene.cloth_garment_data
    for i, garment in enumerate(garments):
        sewing_for_removal = []
        for idx, sewing in enumerate(garment.garment_sewings):
            if sewing.source_obj in instances and (sewing.from_spline_idx, sewing.from_segment_idx) in rem_sewings_ids:
                sewing_for_removal.append(idx)
                continue
            if sewing.target_obj in instances and (sewing.to_spline_idx, sewing.to_segment_idx) in rem_sewings_ids:
                sewing_for_removal.append(idx)
                continue

        for sewing_id in set(sewing_for_removal):
            garment.garment_sewings.remove(sewing_id)

def remap_sewing_to_new_obj(context, sewing_obj, old_new_map, new_obj):
    ''' When we eg. split pattern obj, this will move sewings from original sewing_obj to new_obj'''
    if not old_new_map:
        return
    #old_map corresponds to new_map_id sewings
    #ind patterns that were using splited pattern geo (exept new_obj, that may be just created)
    instances = [obj for obj in bpy.data.objects if obj.data == sewing_obj.data and obj != new_obj]

    for i, garment in enumerate(context.scene.cloth_garment_data):
        pattern_objs = [p.pattern_obj for p in garment.sewing_patterns]
        for idx, sewing in enumerate(garment.garment_sewings):
            if sewing.source_obj in instances and (sewing.from_spline_idx, sewing.from_segment_idx) in old_new_map:
                new_sewing = old_new_map[(sewing.from_spline_idx, sewing.from_segment_idx)]
                sewing.from_spline_idx = new_sewing[0]
                sewing.from_segment_idx = new_sewing[1]
                sewing.source_obj = new_obj
            if sewing.target_obj in instances and (sewing.to_spline_idx, sewing.to_segment_idx) in old_new_map:
                new_sewing = old_new_map[(sewing.to_spline_idx, sewing.to_segment_idx)]
                sewing.to_spline_idx = new_sewing[0]
                sewing.to_segment_idx = new_sewing[1]
                sewing.target_obj = new_obj
            #if we created new obj, add it to patterns
        if new_obj not in pattern_objs:  # old obj was in garment sewings. So add new object to sewings too
            new_pattern = garment.sewing_patterns.add()
            new_pattern['pattern_obj'] = new_obj  # to avoid pattern update func


def backup_sewings(context):
    ''' required since blender operator undo does not reset custom properties set in operator. '''
    backup_sewings = []  # backup_sewings[spl_id][sewing_id] =
    for i, garment in enumerate(context.scene.cloth_garment_data):
        backup_sewings.append([])
        for idx, sewing in enumerate(garment.garment_sewings):
            backup_sewings[i].append([(sewing.from_spline_idx, sewing.from_segment_idx), (sewing.to_spline_idx, sewing.to_segment_idx)])
    return backup_sewings


def load_backup_sewings(context, backup_sewings):
    ''' required - see > backup_sewings()'''
    for i, garment in enumerate(context.scene.cloth_garment_data):
        for s_idx, sewing in enumerate(garment.garment_sewings):
            (sewing.from_spline_idx, sewing.from_segment_idx) = (backup_sewings[i][s_idx][0][0], backup_sewings[i][s_idx][0][1])
            (sewing.to_spline_idx, sewing.to_segment_idx) = (backup_sewings[i][s_idx][1][0], backup_sewings[i][s_idx][1][1])


def get_create_pin_group(garment):
    '''check if target obj exist with ver_mass_group: if so check is garment has this vg, use it. Else Create one with same name
    If no target_obj, or no mass_vg: just create PIN garment.vg_groups
    '''
    cloth_mod_pin_group_name = None
    #searc target obj cloth mod first
    if garment.target_cloth_object:
        sim_obj = garment.target_cloth_object
        cloth_mod = sim_obj.modifiers.get('Cloth')
        if cloth_mod and cloth_mod.settings.vertex_group_mass:
            cloth_mod_pin_group_name = cloth_mod.settings.vertex_group_mass

    if cloth_mod_pin_group_name:
        found_in_garment_pin_group = [vg for vg in garment.generated_vgroups if vg.name == cloth_mod_pin_group_name]
        if found_in_garment_pin_group:
            return found_in_garment_pin_group[0]
        else:
            new_vg_from_cloth_mod = garment.generated_vgroups.add()
            new_vg_from_cloth_mod.name = cloth_mod_pin_group_name
            return new_vg_from_cloth_mod

    find_generic_pin_group = [vg for vg in garment.generated_vgroups if vg.name == 'PIN']
    if find_generic_pin_group:
        return find_generic_pin_group[0]
    else:
        new_generic_garment_vg = garment.generated_vgroups.add()
        new_generic_garment_vg.name = "PIN"
        return new_generic_garment_vg


def fix_garment_vgroups(garment, new_element_id = None, removed_id = None):
    if new_element_id:  # when added obj
        remaining_pat_ids = [i for i, p in enumerate(garment.sewing_patterns) if p.pattern_obj is not None] #when pat is none == it was removed.
        pat_obj = garment.sewing_patterns[new_element_id]
        remaining_pat_ids.insert(new_element_id, None)
    else:
        # only remove element i, if it has.patter_obj, but was removed
        # removed only if: i == removed_id and p.pattern_obj #must have containe obj
        remaining_pat_ids = [i for i, p in enumerate(garment.sewing_patterns) if i != removed_id]  # remaining element in here

    map_old_new = {old_ids: i for i, old_ids in enumerate(remaining_pat_ids)}
    for vg in garment.generated_vgroups:
        vg_enabled_patterns = [int(vg_pattern_name.split('.')[-1]) for vg_pattern_name in vg.vgroups_patterns] #getting ids, from 'patt_name.01' - suffix
        remapped_vgroups_patterns = set()
        for vg_pt_id in vg_enabled_patterns:
            if vg_pt_id in map_old_new:
                if garment.sewing_patterns[map_old_new[vg_pt_id]].pattern_obj:
                    remapped_vgroups_patterns.add(f"{garment.sewing_patterns[map_old_new[vg_pt_id]].pattern_obj.name}.{map_old_new[vg_pt_id]}")
                else: #we will remap to empty pattern slot: None.xyz
                    remapped_vgroups_patterns.add(f"None.{map_old_new[vg_pt_id]}")

        vg.vgroups_patterns=remapped_vgroups_patterns

def get_garment(context):
    garment_index = context.scene.garment_index
    if garment_index < 0 or garment_index >= len(context.scene.cloth_garment_data):
        return None
    garment = context.scene.cloth_garment_data[garment_index]
    return garment


def fix_garment_colliding_names(new_garment, context):
    while True:
        found_collision = False
        for garment in context.scene.cloth_garment_data:
            if garment.name == new_garment.name and garment != new_garment:
                found_collision = True
                # add 00x suffix. Check if name has suffix already
                has_suffix = new_garment.name[-3:].isdigit() and new_garment.name[-4] == '.'
                if has_suffix:
                    new_garment['name'] = new_garment.name[:-3] + str(int(new_garment.name[-3:])+1).zfill(3)
                else:
                    new_garment['name'] = new_garment.name + '.001'
        if not found_collision:
            break



class GTOOL_OT_GarmentCleanup(bpy.types.Operator):
    bl_idname = "garment.cleanup"
    bl_label = "Cleanup garments"
    bl_description = 'Remove unused sewings/garment meshes that are no longer on scene. Resets broken sewings ids to 0'
    bl_options = {'REGISTER', 'UNDO'}

    garment_index: bpy.props.IntProperty()

    def execute(self, context):
        if self.garment_index > len(context.scene.cloth_garment_data)-1:
            return {'CANCELLED'}
        garment = context.scene.cloth_garment_data[self.garment_index]
        #check view layer, cos obj might have been removed, but is it sitll in blend file, cos of pattern_obj pointer
        meshes_to_remove = [idx for idx, garment_mesh in enumerate(garment.sewing_patterns) \
                            if garment_mesh.pattern_obj is None
                            or context.scene.objects.get(garment_mesh.pattern_obj.name) is None]  # pattern_obj.data.dimensions != '2D' ?
        for idx in reversed(meshes_to_remove):
            fix_garment_vgroups(garment, removed_id=idx)
            garment.sewing_patterns.remove(idx)
            # self.report({'INFO'}, 'Removed empty pattern: ')

        pattern_objs = {obj.pattern_obj.name for obj in garment.sewing_patterns}

        #remove sewing that are not attached to pattern objs
        sewing_to_remove = []
        for idx, sewing in enumerate(garment.garment_sewings):
            if sewing.source_obj is None or sewing.source_obj.name not in pattern_objs \
                    or sewing.source_obj.type != 'CURVE' \
                    or sewing.from_spline_idx >= len(sewing.source_obj.data.splines) \
                    or sewing.from_segment_idx >= len(sewing.source_obj.data.splines[sewing.from_spline_idx].bezier_points):
                sewing_to_remove.append(idx)
                continue
            if sewing.target_obj is None or sewing.target_obj.name not in pattern_objs \
                    or sewing.target_obj.type != 'CURVE' \
                    or sewing.to_spline_idx >= len(sewing.target_obj.data.splines) \
                    or sewing.to_segment_idx >= len(sewing.target_obj.data.splines[sewing.to_spline_idx].bezier_points):
                sewing_to_remove.append(idx)
                continue
            if sewing.target_obj == sewing.source_obj and sewing.from_spline_idx == sewing.to_spline_idx and sewing.from_segment_idx == sewing.to_segment_idx:
                sewing_to_remove.append(idx)
                continue

        for idx in reversed(sewing_to_remove):
            self.report({'INFO'}, 'Removed broken sewing: ' + str(idx))
            garment.garment_sewings.remove(idx)
        sewing_to_remove.clear()

        #DONE: remove duplicate sewing
        sewing_count = len(garment.garment_sewings)
        for idx, sewing in enumerate(garment.garment_sewings):
            for idx2 in range(idx+1, sewing_count):
                if sewing.source_obj == garment.garment_sewings[idx2].source_obj \
                        and sewing.target_obj == garment.garment_sewings[idx2].target_obj  \
                        and sewing.from_spline_idx == garment.garment_sewings[idx2].from_spline_idx \
                        and sewing.from_segment_idx == garment.garment_sewings[idx2].from_segment_idx \
                        and sewing.to_spline_idx == garment.garment_sewings[idx2].to_spline_idx \
                        and sewing.to_segment_idx == garment.garment_sewings[idx2].to_segment_idx:
                    sewing_to_remove.append(idx)
                    continue
                if sewing.source_obj == garment.garment_sewings[idx2].target_obj \
                        and sewing.target_obj == garment.garment_sewings[idx2].source_obj \
                        and sewing.from_spline_idx == garment.garment_sewings[idx2].to_spline_idx \
                        and sewing.from_segment_idx == garment.garment_sewings[idx2].to_segment_idx \
                        and sewing.to_spline_idx == garment.garment_sewings[idx2].from_spline_idx \
                        and sewing.to_segment_idx == garment.garment_sewings[idx2].from_segment_idx:
                    sewing_to_remove.append(idx)
                    continue
        for idx in reversed(sewing_to_remove):
            garment.garment_sewings.remove(idx)

        #fix unused pockets. - pattern_obj is empty, or not in scene, target_pattern is empty, or not in scene, or not in PatternsList
        pockets_ids_to_remove = [idx for idx, pocket in enumerate( garment.pockets) if \
                                not pocket.pattern_obj or pocket.pattern_obj.name not in context.scene.objects.keys() or
                                not pocket.target_pattern or pocket.target_pattern.name not in pattern_objs or
                                len(pocket.pocket_sewing) == 0]
        for p_id in reversed(pockets_ids_to_remove):
            garment.pockets.remove(p_id)

        #* Remove broken pins
        pins_to_remove = [idx for idx, pin in enumerate(garment.pins)  \
                          if (pin.source_obj is None or pin.target_obj is None) \
                          or (pin.source_obj.type == 'CURVE' and pin.source_obj.name not in pattern_objs)
                          or (pin.target_obj.type == 'CURVE' and pin.target_obj.name not in pattern_objs)
                          or (pin.source_obj.type == 'EMPTY' and context.scene.objects.get(pin.source_obj.name) is None)
                          or (pin.target_obj.type == 'EMPTY' and context.scene.objects.get(pin.target_obj.name) is None)
                          or (pin.target_obj.type == 'EMPTY' and pin.source_obj.type == 'EMPTY')] #no two empties
        for idx in reversed(pins_to_remove):
            pin = garment.pins[idx]
            if pin.source_obj and pin.source_obj.type == 'EMPTY':
                bpy.data.objects.remove(pin.source_obj)
            if pin.target_obj and pin.target_obj.type == 'EMPTY':
                bpy.data.objects.remove(pin.target_obj)
            garment.pins.remove(idx)

        #for pin.source_obj is always curve first.
        for pin in garment.pins:
            if pin.source_obj.type == 'EMPTY' and pin.target_obj.type == 'CURVE': #fix
                back_src_obj = pin.source_obj
                back_src_co = pin.source_co[:]

                pin.source_obj = pin.target_obj
                pin.source_co = pin.target_co[:]

                pin.target_obj =  back_src_obj
                pin.target_co = back_src_co[:]

        return {'FINISHED'}
